/*
 * StatsCapsule.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *
 */
package JANOS_ANTSR_STATS_APP_PKG;

import JANOS_ANTSR_CORE_PKG.Protocol;
import JANOS_ANTSR_CORE_PKG.Capsule;
import JANOS_ANTSR_CORE_PKG.Xdr;
import JANOS_ANTSR_CORE_PKG.Node;
import JANOS_ANTSR_CORE_PKG.ByteArray;
import JANOS_ANTSR_CORE_PKG.NodeAddress;
import JANOS_ANTSR_CORE_PKG.NodeCache;

import JANOS_ANTSR_CORE_PKG.RouteTable;
import JANOS_ROUTE_PKG.RouteEntry;

import java.util.Enumeration;
import java.util.Vector;

/**
 * Stats request capsule.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public class StatsCapsule
	extends Capsule
{
	final private static byte[] MID = findMID(JANOS_ANTSR_STATS_APP_PKG_S + ".StatsCapsule");
	final private static byte[] PID = findPID(JANOS_ANTSR_STATS_APP_PKG_S + ".StatsCapsule");

	public int rootaddress;
	public long keyid;
	public short dpt;
	public byte depth;
	public long now;
	private final static int cachetime=180;

	public Vector branches;

	protected byte[] mid()
	{
		return MID;
	}

	protected byte[] pid()
	{
		return PID;
	}

	public int length()
	{
		return super.length()+Xdr.INT+Xdr.LONG+Xdr.SHORT+Xdr.BYTE+Xdr.LONG;
	}

	/**
	 * Stats request capsules look like:<br>
	 * 4	Address of the root of the tree<br>
	 * 4	Unique key id of this request tree<br>
	 * 2	Application port<br>
	 * 1	Depth<br>
	 * 4	Time capsule was sent
	 */
	public Xdr encode()
	{
		Xdr xdr = super.encode();
		xdr.PUT(rootaddress);
		xdr.PUT(keyid);
		xdr.PUT(dpt);
		xdr.PUT((byte)(depth-1));
		xdr.PUT(now);
		return xdr;
	}

	public Xdr decode()
	{
		Xdr xdr = super.decode();
		rootaddress=xdr.INT();
		keyid=xdr.LONG();
		dpt=xdr.SHORT();
		depth=xdr.BYTE();
		now=xdr.LONG();
		return xdr;
	}

	/**
	 * Does the actual process of branching the request packet
	 * out to all neighboring nodes except those the have already
	 * been visited.
	 * <p>
	 * Uses soft-state to persistantly keep track of which nodes
	 * have already been visited and to keep from sending back to
	 * root or the source of this request capsule
	 * <p>
	 * @return The number of requests branched out
	 */
	private int branch(Node n)
	{
		int numbranches;
		RouteTable routetable;
		Enumeration routes;
		RouteEntry route;
		int target;
		StatsCapsuleCacheKey skey;
		StatsCapsuleCacheValue sval;
		NodeCache cache;

		branches=new Vector();
		routetable=n.getRouteTable();
		routes=routetable.getNeighbors();
		cache=n.getCache();
		// Add the source to the soft-state
		skey=new StatsCapsuleCacheKey(keyid,rootaddress,getPrevious());
		sval=new StatsCapsuleCacheValue();
		sval.time=n.time();
		cache.put(skey,sval,cachetime);
		// Add the root to the soft-state
		skey=new StatsCapsuleCacheKey(keyid,rootaddress,rootaddress);
		sval=new StatsCapsuleCacheValue();
		sval.time=n.time();
		cache.put(skey,sval,cachetime);
		numbranches=0;
		while (routes.hasMoreElements())
		{
			route=(RouteEntry)routes.nextElement();
			target=route.getLogicalDest();
			synchronized (cache)
			{
				n.log(1,"*** Checking "+NodeAddress.toString(target));
				skey=new StatsCapsuleCacheKey(keyid,rootaddress,target);
				sval=(StatsCapsuleCacheValue)cache.get(skey);
				if (sval!=null)
					continue;	// Already been visited
				else
				{
					sval=new StatsCapsuleCacheValue();
					sval.time=n.time();
					cache.put(skey,sval,cachetime);
				}
			}
			n.log(1,"*** Sending stats request to "+
				NodeAddress.toString(target));
			StatsCapsule c=new StatsCapsule(target);
			c.rootaddress=rootaddress;
			c.keyid=keyid;
			c.dpt=dpt;
			c.depth=(byte)depth;
			c.now=n.time();
			c.prime(this);
			n.routeForNode(c,target);
			numbranches++;
			branches.addElement(new Integer(target));
		}
		return numbranches;
	}

	/**
	 * Encode information about this node to be returned to the root
	 * of the stats
	 * <p>
	 * To change the statistics returned by the tree, change this function
	 */
	private ByteArray encodeNodeInfo(Node n)
	{
		ByteArray buf = new ByteArray(Xdr.INT+
			Xdr.STRING("Hostname")+Xdr.STRING(n.hostname())+
			Xdr.STRING("Physical address spec")+Xdr.STRING(n.getPhysicalAddressSpec())+
			Xdr.STRING("Physical protocol spec")+Xdr.STRING(n.getPhysicalProtocolSpec()));
		Xdr xdr = new Xdr(buf,0);
		xdr.PUT(3);		// Indicates three items of info
		xdr.PUT("Hostname");
		xdr.PUT(n.hostname());
		xdr.PUT("Physical address spec");
		xdr.PUT(n.getPhysicalAddressSpec());
		xdr.PUT("Physical protocol spec");
		xdr.PUT(n.getPhysicalProtocolSpec());
		return buf;
	}

	public boolean evaluate(Node n)
	{
		int numbranches;
		if (n.getAddress()!=getSrc())
		{
			StatsReturnCapsule retCap = new StatsReturnCapsule(dpt,dpt,getSrc(),encodeNodeInfo(n));
			n.log(1,"*** StatsCapsule came from "+NodeAddress.toString(getPrevious()));
			if (depth>0)
			{
				// Branch out to neighboring nodes
				numbranches=branch(n);
			} else {
				numbranches=0;
			}
			// Even if depth has expired, still need to notify
			// root that I am here
			n.log(1,"*** Sending return stats capsule to root "+
				NodeAddress.toString(getSrc()));
			// Return stats
			retCap.infosource=n.getAddress();
			retCap.parentnode=getPrevious();
			retCap.latency=0;
			retCap.capTime=n.time();
			retCap.nodessentto=numbranches;
			retCap.branches=branches;
			retCap.prime(this);
			if (!n.routeForNode(retCap,getSrc()))
				return false;
			if (depth<=0)
			{
				n.log(1,"I have been to the edge; and I'm not coming back (depth expired)");
			}
			return true;
		} else if (n.getAddress()!=getDst()) {
			n.log(1,"*** Routing root stats request out to "+
				NodeAddress.toString(getDst())+
				" from root "+
				NodeAddress.toString(rootaddress)+
				" (key "+keyid+") at "+now);
			return n.routeForNode(this,getDst());
		}
		return false;
	}

	public StatsCapsule()
	{
	}

	public StatsCapsule(int d)
	{
		setDst(d);
	}
}
