/*
 * StatsApplication.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *
 */
package JANOS_ANTSR_STATSFUSION_APP_PKG;

import JANOS_ANTSR_CORE_PKG.Application;
import JANOS_ANTSR_CORE_PKG.Capsule;
import JANOS_ANTSR_CORE_PKG.Protocol;
import JANOS_ANTSR_CORE_PKG.Node;
import JANOS_ANTSR_CORE_PKG.NodeAddress;
import JANOS_ANTSR_CORE_PKG.Xdr;
import JANOS_ANTSR_CORE_PKG.ByteArray;

import JANOS_ANTSR_UTIL_PKG.KeyArgs;
import JANOS_ANTSR_UTIL_PKG.InvalidKeyArgsException;

import JANOS_ANTSR_CORE_PKG.RouteTable;
import JANOS_ROUTE_PKG.RouteEntry;		// XXX

import java.util.Enumeration;			// XXX
import java.util.Vector;			// XXX

/**
 * This application transverses the network, starting with the node
 * it is run from and returns statistics from all reachable nodes
 * within a specifiable depth
 * <p>
 * Parameters:
 * <ul>
 *  <li> -depth <int>: The maximum depth to go to in the tree [15].
 *  <li> -maxwait <int>: The number of seconds to wait for responses before
 *   timing out [30].
 *  <li> -senddelay <int>: The number of milliseconds to wait between sending
 *   root request capsules.  Helps decrease network strain [500].
 *  <li> -cachetime <int>: The number of seconds for a return capsule to wait
 *   in the cache and for a request capsule to keep its tree information [180].
 *  <li> -returnwait <int>: The maximum number of milliseconds to wait for
 *   return capsules before sending what has already been accumulated [10000].
 *  <li> -returntimeout <int>: The maximum number of milliseconds to wait for
 *   unreturned capsules before timing out [60000].
 * </ul>
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public class StatsApplication
	extends Application
	implements Runnable
{
	private final static String[] defaults = { "-depth", "15",
						   "-maxwait", "40",
						   "-senddelay", "500",
						   "-cachetime", "180",
						   "-returnwait", "10000",
						   "-returntimeout", "60000"};

	/**
	 * The logging level messages issued by this class
	 * are issued at.  Gross hack to hide the grosser
	 * hack that is `L'.
	 */
	private final int LOG_LEVEL = L[6];

	/**
	 * The maximum depth that this request will go into
	 * the network before cancelling itself (similar to TTL)
	 */
	private byte depth;

/**
	 * Seconds to wait for activity before the run is stoped
	 */
	private int maxwait;

	/**
	 * Msec to wait between sending root request packets
	 * (lessens network strain)
	 */
	private int rootsenddelay;

	/**
	 * Keeps track of the number of returned packets
	 */
	private int retcount;

	/**
	 * Parameters passed on to the capsule
	 */
	private int cachetime, returnwait, returntimeout;

	private class SrcDestPair
	{
		public int src,dest;

		public SrcDestPair(int src, int dest)
		{
			this.src=src;
			this.dest=dest;
		}

		public boolean equals(Object o)
		{
			return (((SrcDestPair)o).src==src&&
				((SrcDestPair)o).dest==dest);
		}
	}

	/**
	 * Keeps track of the nodes that have sent requests that
	 * have not been replied to
	 */
	private Vector waitingon;

	/**
	 * Keeps track of the parents of the nodes that have returned,
	 * but that the application was not waiting on.
	 * <p>
	 * Sometimes return capsules come from the child before the
	 * parent, this way they are not lost from the tracking system
	 */
	private Vector unmatchedrets;

	/**
	 * Handle the incoming capsule and tell it to print its information
	 */
	synchronized public void receive(Capsule unknownCap)
	{
		int i,j;
		SrcDestPair sdp;
		StatsReturnCapsule cap2;

		super.receive(unknownCap);

		if (!(unknownCap instanceof StatsReturnCapsule))
		{
			thisNode().log(LOG_LEVEL, "Non-stats capsule delivered!?!"+unknownCap);
			thisNode().log(LOG_LEVEL, "StatsCap.classLoader= "+StatsCapsule.class.getClassLoader()+"; other="+unknownCap.getClass().getClassLoader());
			thisNode().log(LOG_LEVEL, "Current flow= "+JANOS_NODEOS_PKG.Flow.currentFlow());
			return;
		}

		StatsReturnCapsule cap = (StatsReturnCapsule)unknownCap;

		cap.showInformation(null,thisNode());
		retcount++;
		// Keep track of what is expected to be returned
		// First add next expectations to the list
		// Need to iterate across all fused capsules as well
		for (j=-1;j<cap.fusedcapsules.size();j++)
		{
			if (j==-1)
				cap2=cap;
			else
				cap2=(StatsReturnCapsule)cap.fusedcapsules.elementAt(j);
			for (i=0;i<cap2.nodessentto;i++)
			{
				// Have I already received this one?
				sdp=new SrcDestPair(cap2.infosource,
					((Integer)cap2.branches.elementAt(i)).intValue());
				if (unmatchedrets.contains(sdp))
				{
					// Its not unmatched now
					unmatchedrets.removeElement(sdp);
				} else {
					waitingon.addElement(sdp);
				}
			}
		}
		// Remove what just came from the expectations
		for (j=-1;j<cap.fusedcapsules.size();j++)
		{
			if (j==-1)
				cap2=cap;
			else
				cap2=(StatsReturnCapsule)cap.fusedcapsules.elementAt(j);
			sdp=new SrcDestPair(cap2.parentnode,cap2.infosource);
			if (waitingon.contains(sdp))
			{
				waitingon.removeElement(sdp);
			} else {
				unmatchedrets.addElement(sdp);
			}
		}
		synchronized (waitingon)
		{
			waitingon.notify();
		}
	}

	/**
	 * Sends out the root request capsules to all
	 * neighboring nodes then sleeps forever
	 */
	public void run()
	{
		int lastcount;
		long lasttime;
		boolean done;
		int i;

		waitingon=new Vector();
		unmatchedrets=new Vector();
		thisNode().log(LOG_LEVEL, "*** Sending root stats request capsules");
		sendRootCapsules();

		try
		{
			lastcount=0;
			lasttime=thisNode().time();
			done=false;
			while (!done&&
				((waitingon.size()!=0||
				unmatchedrets.size()!=0)||
				retcount==0))
			{
				lastcount=retcount;
				synchronized (waitingon)
				{
					waitingon.wait(maxwait*1000);
				}
				if (lastcount==retcount&&thisNode().time()-lasttime>=maxwait*1000)
					done=true;
			}
			if (waitingon.size()==0&&unmatchedrets.size()==0)
			{
				thisNode().log(LOG_LEVEL,"*** Stats collection completed successfully; have a nice day.");
			} else if (done) {
				thisNode().log(LOG_LEVEL,"*** Timeout: Stats run complete");
				for (i=0;i<waitingon.size();i++)
				{
					thisNode().log(LOG_LEVEL,"*** Stats: "+NodeAddress.toString(((SrcDestPair)waitingon.elementAt(i)).dest)+" off of "+NodeAddress.toString(((SrcDestPair)waitingon.elementAt(i)).src)+" did not respond");
				}
			}
		} catch (InterruptedException _) {
		}
		thisNode().shutdown();
	}

	/**
	 * Sends the root request capsule to all neighboring nodes
	 */
	private void sendRootCapsules()
	{
		RouteTable routetable;
		Enumeration routes;
		RouteEntry route;
		int target;
		long keyid;

		routetable=thisNode().getRouteTable();
		routes=routetable.getNeighbors();
		keyid=thisNode().time();
		while (routes.hasMoreElements())
		{
			route=(RouteEntry)routes.nextElement();
			target=route.getLogicalDest();
			thisNode().log(LOG_LEVEL, 
				"*** Sending root stats request (key "+
				keyid+") to "+NodeAddress.toString(target));
			StatsCapsule c = new StatsCapsule(target);
			c.rootaddress=thisNode().getAddress();
			c.keyid=keyid;
			c.dpt=getPort();
			c.depth=(byte)(depth+1);	// Don't count the root node
			c.cachetime=cachetime;
			c.returnwaittime=returnwait;
			c.maxwaittime=returntimeout;
			c.now=thisNode().time();
			send(c);
			// Add that destination to the expected returns
			waitingon.addElement(new SrcDestPair(
				thisNode().getAddress(),target));
			try
			{
				// Lessen network strain
				thisNode().sleep(rootsenddelay);
			} catch (InterruptedException _) {
				// Ignore it
			}
		}
	}

	/**
	 * Parse command line arguments.
	 * <p>
	 * See the class description for a list of the arguments and
	 * their meaning.
	 */
	public void setArgs(KeyArgs k)
		throws InvalidKeyArgsException
	{
		int i;
		k.merge(defaults);

		try
		{
			for (i=0;i<k.length();i++)
			{
				if (k.key(i).equals("-depth"))
				{
					depth=Byte.parseByte(k.arg(i));
					k.strike(i);
				}
				if (k.key(i).equals("-maxwait"))
				{
					maxwait=Integer.parseInt(k.arg(i));
					k.strike(i);
				}
				if (k.key(i).equals("-senddelay"))
				{
					rootsenddelay=Integer.parseInt(k.arg(i));
					k.strike(i);
				}
				if (k.key(i).equals("-cachetime"))
				{
					cachetime=Integer.parseInt(k.arg(i));
					k.strike(i);
				}
				if (k.key(i).equals("-returnwait"))
				{
					returnwait=Integer.parseInt(k.arg(i));
					k.strike(i);
				}
				if (k.key(i).equals("-returntimeout"))
				{
					returntimeout=Integer.parseInt(k.arg(i));
					k.strike(i);
				}
			}
		} catch (Exception e) {
			throw new InvalidKeyArgsException(e.getMessage());
		}
		super.setArgs(k);
	}

	/**
	 * Register the protocol and start the thread
	 */
	public void start()
		throws Exception
	{
		thisNode().register(new StatsProtocol());
		thisNode().threadStart(this);
	}

	/**
	 * Construct a non-GUI stats app.
	 * <p>
	 * Doesn't do anything.
	 */
	public StatsApplication()
		throws Exception
	{
	}
}

