/*
 * AppBooster.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_NODEOS_PKG.Flow;
import JANOS_NODEOS_PKG.CommSpaceHandle;
import JANOS_NODEOS_PKG.CommQueue;
import JANOS_NODEOS_PKG.CommSpaceElement;
import JANOS_NODEOS_PKG.InfoHook;
import JANOS_NODEOS_PKG.Status;

import JANOS_IO_PKG.Log;

import JANOS_ANTSR_UTIL_PKG.InvalidKeyArgsException;

import java.util.Vector;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * Cannot be made public.
 * XXX: well, at least it can't be instantiated by the unwashed masses.
 *
 * CommSpaceHandle should point to a valid CSH for
 * the Booster class, plus:
 *   [2] this class will write the App port (integer) to this offset in the CSH.
 */
public class AppBooster
	extends Booster
{
	/**
	 * Must have the appropriate key to create an AppBooster.
	 * Only the core can actually create Keys, so only trusted
	 * code can create AppBoosters...
	 *
	 * @exception SecurityException if the key is inappropriate
	 * @exception NullPointerException if the key is null
	 */
	public AppBooster(JANOS_NODEOS_PKG.ClassInstantiationKey key)
		throws SecurityException
	{
		key.checkEquals(AppBooster.class);
	}

	protected void run(final Flow flow,
			   String[] args,
			   CommSpaceHandle commSpaceHandle)
	{
		final Application app;
		final Class appClass; 
		final Node localNode;

		require(flow);
		require(args);
		require(args.length >= 1);
		require(args[0]);

		DEBUGONLY(flow.setName("Application: " +args[0]));
		TRACE(Trace.ANTSR_DYNLOAD, "Starting app flow=" +flow);

		// setupLocalNode was inherited from Booster
		localNode = setupLocalNode(flow, commSpaceHandle);

		try
		{
			// Load the class and create initial instance
			appClass = flow.loadClass(args[0]);
			app = (Application)(appClass.newInstance()); 

			// Configure the app
			app.setNode(localNode);
			app.setName(args[0]);
			app.setArgs(args, 1);

			// Install the app in the global application table
			localNode.attachApplication(app);
			TRACE(Trace.ANTSR_APP, "App " +args[0]+ " attached at port " +app.getPort());


			// Start the handler for incoming packets
			flow.threadStart(new Runnable() {
				public void run() {
				/* XXX create a separate class AppPacketProcessor */
					processPacket(flow,
						      app,
						      localNode);
				}
			},
				"Packet Processor for " +args[0]);

			// Start the app
			app.start();

			// Add the app to the application information
			AppsInfo.add(args,app.getPort());
			
		}
		catch (InvalidKeyArgsException ikae)
		{
			localNode.log(Log.ERR, "Bad arguments passed to app: " +ikae.getMessage());
			localNode.log(Log.ERR, ikae);
			localNode.shutdown();
		}
		catch (AttachException ae)
		{
			localNode.log(Log.ERR, "Cannot attach app: " +ae.getMessage());
			localNode.log(Log.ERR, ae);
			localNode.shutdown();
		}
		catch (Exception e)
		{
			localNode.log(Log.ERR, "Exception starting app: " +e.getMessage());
			localNode.logStackTrace(Log.ERR, e);
			localNode.shutdown();
		}
	}

	/**
	 * Delivers packets from a protocol flow to an application.
	 */
	private void processPacket(Flow flow, Application app, Node node)
	{
		CommQueue queue = app.getPacketQueue();

		/*
		 * Loop processing packets from protocol flows.
		 * <p>
		 * XXX Looping on whether the flow was destroyed or not
		 * doesn't really work since we'll be stuck elsewhere and
		 * won't notice unless we get a packet...  Oh well, this
		 * should be taken care of by other stuff.
		 */
		while(!flow.destroyed())
		{
			/* Grab the next capsule from our CommQueue */
			Capsule cap = Capsule.dequeueCapsule(node, queue);

			/* Pass to the app for processing */
			app.receive(cap);
		}
	}

	/**
	 * InfoHook class that reports on the currently running
	 * applications.
	 */
	private static class AppsInfo
		extends InfoHook
	{
		static Vector apps=new Vector();
		private AppsInfo()
		{
			super("Apps");
		}

		public String[] getInfo()
		{
			String[] retval=new String[apps.size()];
			int i;
			for (i=0;i<apps.size();i++)
				retval[i]=(String)apps.elementAt(i);
			return retval;
		}

		public static void add(String[] args, int port)
		{
			String appinfo=new String();
			int i;
			for (i=0;i<args.length;i++)
				appinfo+=args[i]+" ";
			appinfo+="[port "+port+"]";
			apps.addElement(appinfo);
		}
	}

	static
	{
		Status.addInfoHook(new AppsInfo());
	}
}
