/* 
 *  $Id: ConfigurationManager.Java,v 1.8 2000/08/28 22:21:20 tullmann Exp $
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_ANTSR_UTIL_PKG.KeyArgs;
import JANOS_ANTSR_UTIL_PKG.InvalidKeyArgsException;
import JANOS_ANTSR_UTIL_PKG.LineListTokenizer;
import JANOS_ANTSR_UTIL_PKG.LineListException;

import JANOS_NODEOS_PKG.Address;
import JANOS_NODEOS_PKG.AddressFormatException;
import JANOS_NODEOS_PKG.BufferHandle;
import JANOS_NODEOS_PKG.CommSpaceHandle;
import JANOS_NODEOS_PKG.Flow;
import JANOS_NODEOS_PKG.FlowRun;
import JANOS_NODEOS_PKG.NoSuchNodeAddress;
import JANOS_IO_PKG.Log;

import java.io.InputStreamReader;
import java.io.StringReader;
import java.io.FileNotFoundException;
import java.io.IOException;

import java.util.Enumeration;
import java.util.Vector;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * A configuration manager that creates a node interface,
 * channel and app configuration.  Gets the configuration
 * from a control file.
 * <p>
 * Really, these functions could go on the Node class.
 * <p>
 * XXX this shouldn't be public, or arbitrary Capsules
 * can call it up...
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 * @author David Wetherall
 */
public class ConfigurationManager
	extends FlowRun
{
	private Flow flow;

	/**
	 * Not private so that the inner Runnable() created on the
	 * channel command can see it and use it for the thread it
	 * starts up on the main message loop.
	 */
	PACKAGE static int port = 4242;

	/**
	 * Parse a single command.
	 * 
	 * @param node the node to attach created entities to
	 * @param command array of command and its argument words
	 * @exception AddressFormatException if an address parameter is bad
	 * @exception ArrayIndexOutOfBoundsException if insufficent paramters are given for some commands.
	 * @exception AttachException if attaching Channel/Extension/App fails
	 * @exception ClassNotFoundException if any dynamically loaded class isn't found.
	 * @exception IllegalAccessException if newInstance() fails
	 * @exception InstantiationException if newInstance() fails
	 * @exception InvalidKeyArgsException if any optional args are improperly formatted
	 */
	private final void parseCommandLine(Vector appVector, String[] command)
		throws 
		  AddressFormatException,
		  ArrayIndexOutOfBoundsException,
		  AttachException,
		  ClassNotFoundException,
		  IllegalAccessException,
		  InstantiationException,
		  InvalidKeyArgsException
	{
		TRACE(Trace.ANTSR_STARTUP, "Parsing command " +command[0]);

		/*
		 * Format of all command lines:
		 * 	<command> <address> [<args> ...]
		 * The first two elements are required.  The args are dependent
		 * upon the command.
		 */
		if (command[0].equals("node")) {
			// FORMAT: node <node addr> [<optional arg> ...]

			// XXX node.setName("Node " + command[1]);

			// Set address before calling setArgs so that routeTable will exist.
			PrimordialNode.setAddress(Address.fromString(command[1]));
			PrimordialNode.setArgs(new KeyArgs(command, 2));

		} else if (command[0].equals("channel")) {
			// FORMAT: channel <node addr> <physical addr> [<optional args> ... ]

			if (command.length < 3)
				throw new ArrayIndexOutOfBoundsException("insufficent "
									 + "arguments for "
									 + "channel command");


			// XXX Parsing physAddr is now all redundant.  Remove this code/checks.
			String physAddr = command[2];
			String host = "localhost";
			String transport = "udp";
			int sep = physAddr.indexOf(':');
			if (sep != 0)
			{
				int tsep = physAddr.indexOf(':', sep + 1);
				try
				{
					host = physAddr.substring(0,sep);
					assert(host.equals("localhost")||host.equals("127.0.0.1")); // XXX
					port = Integer.parseInt(physAddr.substring(sep+1,
										   ((tsep > 0) ?
										    tsep : 
										    physAddr.length())));
					if (tsep > 0)
						transport = physAddr.
							substring(tsep+1,
								  physAddr.length());
								
				// above might throw NumberFormatException
				}
				catch (Exception e)
				{
					panic("Problem parsing channel address: " +e.getMessage());
				}
			}

			// XXX exactly one channel per node (only one physical address per node!)
			PrimordialNode.setPhysicalAddress("if/ipv4/" +transport,
							  "*:" +port+ ":" +host);
			
		} else if (command[0].equals("application")) {
			// FORMAT: application <node addr> <app class> [<optional args> ...]
			
			if (command.length < 3)
				throw new ArrayIndexOutOfBoundsException("insufficient arguments for application");
			
			/*
			 * Store the name of the app, it will be forced into a
			 * new Flow and started up in Node.start()
			 */
			appVector.addElement(new AppConfig(command[2], command, 3));

		} else if (command[0].equals("extension")) {
			PrimordialNode.message("extension command ignored by ANTSR.");  // XXX
			// IGNORE.  Not used in Janos ANTS
			// FORMAT: extension <node addr> <class name> [<optional args> ... ]

		} else if (command[0].equals("manager")) {
			PrimordialNode.message("manager command ignored by ANTSR.");  // XXX
			// IGNORE.  Not used in Janos ANTS
			
		} else if (command[0].equals("connect")) {
			// IGNORE.  Its a legit command that is handled by makeroutes...
			
		} else {
			PrimordialNode.message("Ignored bad command " +command[0]+ "."); // XXX
		}
	}
	

	/**
	 */
	private final void parseCommands(Vector apps, String[] params)
	{
		// two required args
		String filename = params[0];
		String nodename = params[1];

		// Remainder go in a KeyArgs structure
		try
		{
			KeyArgs k = new KeyArgs(params, 2);
		}
		catch (InvalidKeyArgsException ikae)
		{
			FATALEXCEPTION(ikae, "ConfigurationManager args are bogus.");
		}
		
		/*
		 * Process the configuration file, creating entities
		 * as needed for each command.
		 */

		LineListTokenizer lt = null;
		try
		{
			lt = new LineListTokenizer(filename);
		}
		catch (FileNotFoundException fnfe)
		{
			panic(fnfe.getMessage());
		}
    
		String[] args;
		do
		{
			try
			{
				args = lt.list();
			}
			catch (LineListException lle)
			{
				FATALEXCEPTION(lle, "malformed line in " +filename+ ": " +lle.getMessage());
			}
			
			if (args == null)
				continue;

			if (args.length < 2)
			{
				// XXX do real i/o
				PrimordialNode.message(Log.ERR,
						       "Bad command/line "
						       + args[0]
						       + ".");
				continue;
			}
			
			// Ignore commands that aren't for this node.
			if (! args[1].equals(nodename))
				continue;

			try
			{
				parseCommandLine(apps, args);
			}
			catch (Exception e)
			{
				FATALEXCEPTION(e, "When parsing command " +args[0]+ ".");
			}
		}
		while (args != null);
	}
  
	/**
	 * OLD way to start an ANTS node.  Does not apply in ANTSR.
	 *
	 * @see #run()
	 */
	public static void main(String[] args)
		throws Exception
	{
		PrimordialNode.message(Log.ERR,
				       "ConfigurationManager must be started as a Janos Flow.");
		System.exit(11);
	}

	/**
	 * Part of the FlowRun interface
	 *
	 * @see core.FlowRun
	 * @deprecated Just use BufferHandle.create()
	 */
	protected BufferHandle newBufferHandle()
	{
		return BufferHandle.create();
	}

	/**
	 * Part of the FlowRun interface
	 *
	 * @see core.FlowRun
	 * @deprecated Just use BufferHandle.reset()
	 */
	protected void recycleBufferHandle(BufferHandle bh)
	{
		bh.reset();
	}

	/**
	 * Start a new configuration manager.
	 * Prints a usage message of insufficient arguments are given.
	 *
	 * @param args the arguments to pass to the ConfigurationManager constructor.
	 * @see #ConfigurationManager(java.lang.String[])
	 */
	protected void run(Flow flow,
			   String[] args,
			   CommSpaceHandle commSpaceHandle)
	{
		DEBUGONLY(JANOS_NODEOS_PKG.Flow.currentFlow().setName("Primoridal ANTSR Flow"));

		this.flow = flow; // for newBufferHandle();
		
		if ((args == null)
		    || (args.length < 2)) {
			PrimordialNode.message(Log.INFO,
					       "usage: " 
					       +ConfigurationManager.class.getName()+
					       " <file> <node> [keyargs]");
			System.exit(1);
		}
		
		Vector apps = new Vector();
		Vector exts = new Vector();
		parseCommands(apps, args);

		try
		{
			PrimordialNode.start(apps, exts);
		}
		catch (Exception e)
		{
			PrimordialNode.message("Exception in PrimordialNode startup: " +e.getMessage());
			PrimordialNode.dumpStackTrace(e);
			PrimordialNode.shutdown();
		}
	}
}
