/*
 * ConsoleCommand.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

package JANOS_ANTSR_CORE_PKG;

import java.util.Hashtable;
import java.util.Enumeration;

/**
 * A base class for any commands available in the console.
 */
PACKAGE abstract class ConsoleCommand
{
	/**
	 * The name of this command.  Set when the command
	 * is added to the the table.
	 *
	 * @see #addCommand(String,ConsoleCommand)
	 */
	private String name; 
	
	PACKAGE static class Alias
	{
		private final String name;
		private final ConsoleCommand cc;
		
		PACKAGE Alias(String name, ConsoleCommand cc)
		{
			this.name = name;
			this.cc = cc;
		}

		PACKAGE String getName()
		{
			return name;
		}

		PACKAGE ConsoleCommand getCommand()
		{
			return this.cc;
		}
	}
	
	/**
	 * The table of available commands
	 */
	PACKAGE static Hashtable commands = new Hashtable();

	/**
	 * Return a description of the command.
	 */
	PACKAGE abstract String getDescription();

	/**
	 * Return the primary name of this command.
	 */
	PACKAGE String getName()
	{
		return name;
	}

	/**
	 * Returns additional help based on the given arguments.
	 * <p>
	 * args[0] - The help command name
	 * args[1] - This command name
	 * args[2+] - Arguments that should be described in the context of
	 * this command
	 */
	protected void getAdditionalHelp(ConsoleWorker cw, String args[])
	{
		cw.getOutputStream().
			println("No additional help for this command.");
	}

	/**
	 * Handle the command event with the given arguments
	 * <p>
	 * args[0] - This command name
	 * args[1+] - Arguments to the command
	 */
	protected abstract boolean handle(ConsoleWorker cw, String args[]);

	/**
	 * Add a command indexed by the given name to the table of available
	 * commands.
	 */
	PACKAGE static void addCommand(String name, ConsoleCommand cc)
	{
		cc.name = name;
		ConsoleCommand.commands.put(cc.name, cc);
	}

	/**
	 * Add an alias for an existing console command
	 */
	PACKAGE static void addCommandAlias(String name, ConsoleCommand cc)
	{
		ConsoleCommand.commands.put(name, new Alias(name, cc));
	}
	
	/**
	 * Return the collection of all available commands
	 *
	 * XXX returns a java.util class.
	 */
	PACKAGE static Enumeration getCommands() 
	{
		return ConsoleCommand.commands.elements();
	}

	/**
	 * Find the ConsoleCommand object corresponding to the given name
	 */
	PACKAGE static ConsoleCommand findCommand(String name)
	{
		ConsoleCommand retval = null;
		Object obj;
		
		obj = ConsoleCommand.commands.get(name);
		if( obj instanceof Alias )
			retval = ((Alias)obj).getCommand();
		else
			retval = (ConsoleCommand)obj;
		return retval;
	}

	/**
	 * Handle the command line specified by args.
	 * <p>
	 * args[0] - The name of the command
	 * args[1+] - Additional arguments to the command
	 */
	PACKAGE static boolean handleCommand(ConsoleWorker cw, String args[])
	{
		ConsoleCommand cc;
		boolean retval;

		cc = ConsoleCommand.findCommand(args[0]);
		if( cc != null )
		{
			retval = cc.handle(cw, args);
		}
		else
		{
			cw.getOutputStream().println("Command "
						     + args[0]
						     + " is not recognized.");
			cw.getOutputStream().println("Type `help' for a list "
						     + "of valid commands");
			retval = false;
		}
		return retval;
	}

	public String toString()
	{
		return "ConsoleCommand[name="
			+ this.getName()
			+ "; description="
			+ this.getDescription()
			+ "]";
	}
}
