/*
 * ConsoleWorker.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_NODEOS_PKG.Status;
import JANOS_NODEOS_PKG.Address;
import JANOS_NODEOS_PKG.Credentials; // XXX for app start
import JANOS_NODEOS_PKG.ThreadPoolSpec; // XXX for app start
import JANOS_NODEOS_PKG.MemSpec; // XXX for app start
import JANOS_NODEOS_PKG.Flow;

import JANOS_IO_PKG.Log;

import JANOS_ANTSR_UTIL_PKG.LineListTokenizer;
import JANOS_ANTSR_UTIL_PKG.LineListException;

import java.net.ServerSocket;
import java.net.Socket;

import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.StringReader;
import java.io.PrintWriter;

import java.util.Date;
import java.util.Enumeration;

import java.text.DecimalFormat;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * A helper class for Console, it takes care of individual connections so that
 * more than one person can be connected at a time.
 */
PACKAGE class ConsoleWorker
	implements Runnable // XXX java.lang dependency
{
	/**
	 * The node we're executing on
	 */
	private final Node node;

	/**
	 * The client socket object
	 */
	private final Socket client;

	/**
	 * Output stream
	 */
	PACKAGE PrintWriter out;

	/**
	 * Input stream
	 */
	private LineNumberReader in;

	/**
	 * The `exit' command, just cause the console to quit out.
	 */
	private static final class ExitCommand
		extends ConsoleCommand
	{
		PACKAGE String getDescription()
		{
			return "Close the console connection.";
		}
		
		protected boolean handle(ConsoleWorker cw, String args[])
		{
			return true;
		}
	}

	/**
	 * The `debug' command, it lets the user set the debugging level at
	 * runtime.
	 */
	private static final class DebugCommand
		extends ConsoleCommand
	{
		PACKAGE String getDescription()
		{
			return "Set the debugging level for the node";
		}

		protected boolean handle(ConsoleWorker cw, String args[])
		{
#ifdef DEBUG
			if( args.length > 1 )
			{
				if( args[1].equals("list") )
				{
					cw.dumpStrings(Trace.getNameList());
				}
				else
				{
					int i;

					for( i = 1; i < args.length; i++ )
					{
						Trace.setFlags(args[i]);
					}
				}
			}
			else
			{
				cw.println("Insufficient arguments to set "
					       + "debugging flags.  Use `list' "
					       + "to get a list of flags");
			}
#else
			cw.println("'debug' command is not available in an optimized NodeOS.");
#endif
			return false;
		}
	}

	/**
	 * The `shutdown' command, just shuts down the node
	 */
	private static final class ShutdownCommand
		extends ConsoleCommand
	{
		PACKAGE String getDescription()
		{
			return "Shutdown this active node";
		}
		
		protected boolean handle(ConsoleWorker cw, String args[])
		{
			cw.log(Log.WARNING,
				    "Shutdown issued from " +cw.toString());
			cw.shutdown();
			return true;
		}
	}

	/**
	 * The `gc' command, causes the garbage collector to be run
	 */
	private static final class GarbageCollectCommand
		extends ConsoleCommand
	{
		PACKAGE String getDescription()
		{
			return "Run the garbage collector.";
		}
		
		protected boolean handle(ConsoleWorker cw, String args[])
		{
			java.lang.System.gc();
			return false;
		}
	}

	/**
	 * The `info' command, returns information on the specified objects.
	 */
	private static final class InfoCommand
		extends ConsoleCommand
	{
		/**
		 * Status object for examining the state of the nodeos
		 */
		private static final Status status = Status.getStatusObject();
	
		PACKAGE String getDescription()
		{
			return "Get info on name, or use `list' to get a "
				+ "listing of possible names.";
		}
		
		protected boolean handle(ConsoleWorker cw, String args[])
		{
			if ( (args.length > 1)
			     && (args[1] != null) )
			{
				if( args[1].equals("list") )
				{
					String ilist[] = status.getInfoList();
					
					cw.println("List of available "
						       + "info modules:");
					cw.dumpStrings(ilist);
				}
				else
				{
					String info[];

					info = status.getInfo(args[1]);
					if( info != null )
					{
						cw.println("Info for "
							       + args[1]
							       + ":");
						cw.dumpStrings(info);
					}
					else
					{
						cw.println("Info module `"
							       + args[1]
							       + "' wasn't "
							       + "found");
					}
				}
			}
			else
			{
				cw.println("The `info' command needs an "
					       + "argument.\nUse `list' to "
					       + "get a list of available "
					       + "modules.");
			}
			return false;
		}
	}

	/**
	 * The `app' command, starts up an application on the node.
	 */
	private static final class AppCommand
		extends ConsoleCommand
	{
		PACKAGE String getDescription()
		{
			return "Start an application.";
		}
		
		/**
		 * Start an application.  args[0] is the name of the app (i.e.,
		 * the class to load.  App is started in its own flow.
		 */
		protected boolean handle(ConsoleWorker cw, String args[])
		{
			AppConfig appconf;

			if (args.length < 2)
			{
				cw.println("Insufficient arguments to start an application.");
				cw.println("Usage: app <class> [app args]");
				return false;
			}

			appconf = new AppConfig(args[1], args, 2);
		
			try
			{
				PrimordialNode.fireUpNewFlow(new Credentials("remote app"), // XXX
							     new ThreadPoolSpec(8, 0, null), // XXX
							     new MemSpec(), // XXX
							     null,
							     null,
							     appconf.className(),
							     appconf.args());
			}
			catch (InterruptedException ie)
			{
				cw.println("Interrupted during app start: " +ie.getMessage());
				cw.printStackTrace(ie);
				cw.println("Try again.");
			}
			catch (java.io.IOException ioe)
			{
				cw.println("IOException during app start: " +ioe.getMessage());
				cw.printStackTrace(ioe);
				cw.println("Try again.");
			}
			
			return false;
		}
	}

	/**
	 * The `uptime' command, prints out how long the node has been running.
	 */
	private static final class UptimeCommand
		extends ConsoleCommand
	{
		PACKAGE String getDescription()
		{
			return "Print out the uptime for this active node.";
		}
		
		protected boolean handle(ConsoleWorker cw, String args[])
		{
			long uptime = System.currentTimeMillis() -
				PrimordialNode.getStartupTime();
			
			final long millisInSecond = 1000;
			final long millisInMinute = millisInSecond * 60;
			final long millisInHour = millisInMinute * 60;
			final long millisInDay = millisInHour * 24;

			final long days = uptime / millisInDay;
			uptime %= millisInDay;
			final long hours = uptime / millisInHour;
			uptime %= millisInHour;
			final long minutes = uptime / millisInMinute;
			uptime %= millisInMinute;
			final long seconds = uptime / millisInSecond;

			final DecimalFormat df = new DecimalFormat( "#00" );
			
			cw.println(" "
				       + new Date()
				       + "  up "
				       + days
				       + " days,  "
				       + hours
				       + ":"
				       + df.format( minutes )
				       + ":"
				       + df.format( seconds ));
			return false;
		}
	}

	/**
	 * The `help' command, just prints out the available commands
	 */
	private static final class HelpCommand
		extends ConsoleCommand
	{
		PACKAGE String getDescription()
		{
			return "This message.";
		}

		protected boolean handle(ConsoleWorker cw, String args[])
		{
			if( args.length > 1 )
			{
				ConsoleCommand cc;

				cc = ConsoleCommand.findCommand(args[1]);
				if( cc != null )
				{
					cc.getAdditionalHelp(cw, args);
				}
				else
				{
					cw.println(
						"Command "
						+ args[1]
						+ " is not recognized.");
					cw.println("Type `help' alone to "
						       + "get a list of valid "
						       + "commands.");
				}
			}
			else
			{
				Enumeration e;
				int i;

				e = ConsoleCommand.getCommands();
				cw.println("Janos ANTSR Console help:");
				while (e.hasMoreElements())
				{
					Object obj;

					obj = e.nextElement();
					if( obj instanceof ConsoleCommand )
					{
						ConsoleCommand cc;

						cc = (ConsoleCommand)obj;
						cw.println(
							"  "
							+ cc.getName()
							+ " -- "
							+ cc.getDescription());
					}
					else if ( obj instanceof Alias )
					{
						Alias cca;

						cca = (Alias)obj;
						cw.println("  "
							   + cca.getName()
							   + " is an alias for "
							   + cca.getCommand().getName());
					}
				}
			}
			return false;
		}
	}

	static
	{
		/* Add our commands to the console command table */
		
		ExitCommand ec = new ExitCommand();
		HelpCommand hc = new HelpCommand();
		AppCommand ac = new AppCommand();
		
		ConsoleCommand.addCommand("exit", ec);
		ConsoleCommand.addCommandAlias("quit", ec);
		ConsoleCommand.addCommand("debug", new DebugCommand());
		ConsoleCommand.addCommand("shutdown", new ShutdownCommand());
		ConsoleCommand.addCommand("gc", new GarbageCollectCommand());
		ConsoleCommand.addCommand("uptime", new UptimeCommand());
		ConsoleCommand.addCommand("help", hc);
		ConsoleCommand.addCommandAlias("?", hc);
		ConsoleCommand.addCommand("app", ac);
		ConsoleCommand.addCommandAlias("start", ac);
		ConsoleCommand.addCommand("info", new InfoCommand());
	}
	
	/**
	 * Setup the input and output stream objects for the given client
	 */
	public ConsoleWorker(Node node, Socket sock)
		throws java.io.IOException
	{
		this.node = node;
		this.client = sock;
		this.out = new PrintWriter(sock.getOutputStream());
		this.in = new LineNumberReader(
			new InputStreamReader(sock.getInputStream()));
	}

	/**
	 * Returns the output stream for this console object, only really
	 * useful for external console commands.
	 */
	PACKAGE PrintWriter getOutputStream()
	{
		return this.out;
	}

	/**
	 * Returns the input stream for this console object, only really
	 * useful for external console commands.
	 */
	PACKAGE LineNumberReader getInputStream()
	{
		return this.in;
	}
	
	public String toString()
	{
		return "ConsoleWorker["
			+ client.toString()
			+ "]";
	}

	/**
	 * Set the event handler in motion
	 */
	public void run()
	{
		try
		{
			this.handleClient();
			this.in.close();
			this.out.close();
			this.client.close();
		}
		catch(Throwable th)
		{
			log(Log.ERR,
				 "There was an error in the console handler");
			logStackTrace(Log.ERR, th);
		}
	}

	/**
	 * Event handler for the console, just prints out a simple prompt and
	 * then accepts some simple commands from the user
	 */
	private void handleClient()
		throws java.io.IOException
	{
		boolean done = false;

		banner();
		
		while( !done )
		{
			String[] args;
			
			this.prompt();
			args = this.getCommand();
			
			if( args != null )
			{
				try
				{
					done = runCommand(args);
					out.flush();
				}
				catch( Throwable th )
				{
					log(Log.WARNING,
						 "Error executing - "
						 + args[0]
						 + ":  "
						 + th.getMessage());
					logStackTrace(Log.WARNING, th);
				}
			}
		}
		TRACE(Trace.ANTSR_CONSOLE, "Session complete.");
	}

	/**
	 * Print out a welcome message for the user
	 */
	private void banner()
	{
		println("Welcome to Janos ANTSR " +node);
		println("Type `exit' to exit.");
		out.flush();
	}

	/**
	 * Print out a meaningful prompt for the user
	 */
	private void prompt()
	{
		this.out.print(Address.toString(PrimordialNode.getAddress())
			       + "> ");
		this.out.flush();
	}

	/**
	 * Parse the command line entered by the user
	 *
	 * @return An array of strings corresponding to the arguments entered
	 */
	private String[] getCommand()
		throws java.io.IOException
	{
		String line = in.readLine();
		String[] cmd;

		TRACE(Trace.ANTSR_CONSOLE, "Got input line: " +line);

		line = line + "\n";
		LineListTokenizer lt =
			new LineListTokenizer(new StringReader(line));

		try
		{
			cmd = lt.list();
		}
		catch (LineListException lle)
		{
			cmd = null;
			println("Error parsing command: "
				    + lle.getMessage());
		}

		// System.out.println("Parsed line into " +cmd.length+ " words.");
		// System.out.println("cmd[0] = " +((cmd!=null)?cmd[0]:"<cmd is null>"));

		return cmd;
	}

	/**
	 * Print out an array of strings
	 */
	PACKAGE final void dumpStrings(String[] s)
	{
		for (int i = 0; i < s.length; i++)
		{
			if( s[i] != null )
				out.println("  " + s[i]);
		}
	}

	/**
	 * Execute whatever command the user entered.  Return true if
	 * connection should close (i.e., exit was run).
	 *
	 * @param args the command line (broken into one token per String)
	 * @return true if connection should close
	 */
	private boolean runCommand(String[] args)
	{
		if ((args == null) || (args[0] == null))
			return false;

		return ConsoleCommand.handleCommand(this, args);
	}

	PACKAGE void println(String msg)
	{
		out.println(msg);
	}

	PACKAGE void printStackTrace(Throwable t)
	{
		t.printStackTrace(out);
	}

	PACKAGE void log(int level, String msg)
	{
		node.log(level, msg);
	}

	PACKAGE void logStackTrace(int level, Throwable th)
	{
		node.logStackTrace(level, th);
	}

	PACKAGE void shutdown()
	{
		node.shutdown();
	}
}
