/* 
 *  $Id: DLResponseCapsule.Java,v 1.2 2000/01/12 21:26:57 stack Exp $
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTSR_CORE_PKG;

import java.util.Vector;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * Demand Load Response capsule.  DLResponseCapsules are only created
 * by a DLRequestCapsule that has been delivered to a node that can
 * satisfy the request.  DLResponseCapsules contain a single "chunk"
 * of a CodeGroup.
 * <p>
 * A DLResponseCapsule packet contains the following fields (on top of the
 * base Capsule fields):
 * <ul>
 *  <li> payloadID: the TypeID of the CapsuleClass for which this
 *  DLResponseCapsule holds code.
 *  <li> groupID: the TypeID of the CodeGroup to which this
 *  DLResponseCapsule's CapsuleClass should be added.
 *  <li> array len: the length of the array (next entry) (short).
 *  <li> Array of groupIDs: the TypeIDs of the other CodeGroup's in
 *  the Protocol to which this DLResponseCapsule will be added.
 *  <li> chunkMask: the chunk in the CodeGroup's chunk list which this
 *  DLResponseCapsule provides.
 *  <li> clen: the length of this chunk (int, could be short) (could
 *  be nuked, implicit in packet length)
 *  <li> coff: the offset of this chunk in the CapsuleClass's code
 *  array.  (int, could be short).
 *  <li> name: A string name of the CapsuleClass this
 *  DLResponseCapsule represents.
 *  <li> code: the chunk itself.
 * </li>
 * 
 * @see DLProtocol
 * @see DLRequestCapsule
 * @see CodeGroup
 * @author David Wetherall
 */
public class DLResponseCapsule
	extends Capsule 
{
	// WARNING: this is a builtin capsule, whose version number
	// must be manually kept in synch with  its functionality
	final public static String VERSION = "JANOS DLResponseCapsule July 17, 1999";

	final PACKAGE static byte[] MID = findMID(JANOS_ANTSR_CORE_PKG_S + ".DLResponseCapsule");
	final private static byte[] PID = findPID(JANOS_ANTSR_CORE_PKG_S + ".DLResponseCapsule");

	protected byte[] mid()
	{
		return MID;
	}

	protected byte[] pid()
	{
		return PID;
	}

	/**
	 * The MID of the CapsuleClass owning the chunk this capsule
	 * contains.
	 */
	private TypeID payloadID;

	/**
	 * The GID of the CodeGroup this chunk belongs to.
	 */
	private TypeID groupID;

	/**
	 * The GID of every other CodeGroup in the Protocol this CodeGroup
	 * is part of.  This is the only way a CodeGroup can compute
	 * its ProtocolID.
	 * <p>
	 * XXX seems like it would be more efficient to just send
	 * the ProtocolID across the wire and double check it after
	 * the CodeGroup is loaded.  XXX Probably a security concern.
	 */
	private TypeID[] groups;

	/**
	 * The mask containing only the index of the chunk (in the
	 * CodeGroup's chunkmask) contained in this Capsule.
	 */
	private int chunkMask;

	/**
	 * Length of the full bytecode for the Class.
	 */
	private int clen;

	/**
	 * Offset of this chunk in the CapsuleClass's code block.
	 */
	private int coff;

	/**
	 * Name of the CapsuleClass this chunk represents.
	 * <p>
	 * XXX Only needs to be in the first chunk for a class.
	 */
	private String name;

	/**
	 * The chunk.
	 */
	private ByteArray code;

	public int length() {
		int s = Xdr.TYPEID + // groupID
			Xdr.TYPEID + // payloadID
			Xdr.SHORT + // number of groups in groupID array
			groups.length*Xdr.TYPEID + // array of groupIDs
			Xdr.INT + // chunkMask
			Xdr.INT + // clen
			Xdr.INT + // coff
			Xdr.STRING(name) + // CapsuleClass name
			Xdr.BYTEARRAY(code); // the code block
		return super.length() + s;
	}
  
	public Xdr encode() {
		Xdr xdr = super.encode();
		xdr.PUT(groupID);
		xdr.PUT(payloadID);
		xdr.PUT((short)groups.length);
		for (int n = 0; n < groups.length; n++)
			xdr.PUT(groups[n]);
		xdr.PUT(chunkMask);
		xdr.PUT(clen);
		xdr.PUT(coff);
		xdr.PUT(name);
		xdr.PUT(code);

		return xdr;
	}

	public Xdr decode() {
		Xdr xdr = super.decode();
		groupID = xdr.TYPEID();
		payloadID = xdr.TYPEID();
		int l = xdr.SHORT();
		groups = new TypeID[l];
		for (int n = 0; n < l; n++) {
			groups[n] = xdr.TYPEID();
		}
		chunkMask = xdr.INT();
		clen = xdr.INT();
		coff = xdr.INT();
		name = xdr.STRING();
		code = xdr.BYTEARRAY();
    
		return xdr;
	}
  
	/**
	 * Evalute a DLResponseCapsule at this node.
	 * Looks up the appropriate CapsuleClass and CodeGroup info,
	 * creating new ones if necessary.
	 * <p>
	 * XXX should deal with unanticipated DLResponseCapsules better.
	 */
	public boolean evaluate(Node node)
	{
		TRACE(Trace.ANTSR_DYNLOAD, "evaluate DLResponse mid=" +payloadID+ "; coff=" +coff);
		if( node.getAddress() != getDst() )
		{
			return node.routeForNode(this, getDst());
		}
		
		// The packet contains all the GIDs but not the PID.
		// We could add the PID to each Node
		// object, or we can just recompute it here.  It seems silly but
		// here we go...
		TypeID protocolID = TypeID.computeFromIDs(groups);

		// Look up the group the CapsuleClass belongs in, create CodeGroup if necessary.
		CodeGroup g;
		synchronized(node)
		{
			g = node.lookupCodeGroup(groupID);
			if (g == null) {
				g = new CodeGroup(groupID, groups, protocolID, false);
			}
		}

		synchronized(g)
		{
			// Look up the CapsuleClass, and create one if it doesn't exist yet
			CapsuleClass cc = node.lookupCapsuleClass(payloadID);
			TRACE(Trace.ANTSR_DYNLOAD, "cc for mid " +payloadID+ " -> " +cc);

			if (cc == null) 
			{
				try
				{
					TRACE(Trace.ANTSR_DYNLOAD, "DLResponse: allocating " +clen+ " code bytes for " +name);

					// XXX should be able to construct this outside the synchronized block...
					// May throw DuplicateCapsuleClassException
					cc = new CapsuleClass(g, payloadID, name, new byte[clen]);
					g.add(cc);
				}
				catch(Exception e)
				{
					TRACE(Trace.ANTSR_DYNLOAD, "dropping DLResponse: " +e);
					return false;
				}
			}

			/*
			 * If the chunk is one we want and don't already have,
			 * allocate the CapsuleClass's code block if necessary
			 */
			if ((g.availChunkMask | chunkMask) != g.availChunkMask)
			{
				PendingRequest pr;
				
				TRACE(Trace.ANTSR_DYNLOAD, "DLResponse: adding " +name+ " to " +g+ " (off=" +coff+ "; chunkMask=" +Integer.toHexString(chunkMask)+ "; len=" +code.length()+ ")");
        
				code.getBytes(0, cc.code(), coff, code.length());
				g.availChunkMask |= chunkMask;
        
				// if we're done loading stuff for this code group
				if (g.availChunkMask == 0xffffffff) 
				{
					/*
					 * We got everything so we can remove
					 * our pending request
					 */
					node.getCache().remove(groupID);
					TRACE(Trace.ANTSR_DYNLOAD, "DLResponse: completed with "+name);
  
					try
					{
						// activate new code group
						g.activate(node, true);
					}
					catch(ClassNotFoundException e)
					{
						TRACE(Trace.ANTSR_DYNLOAD, "DLResponse: dropping packet: " +e);
						return false;
					}
				}
				else if( (pr = (PendingRequest)node.getCache().
					  get(payloadID)) != null )
				{
					DLRequestCapsule req;
					
					/*
					 * We haven't got everything yet but
					 * we don't need to send another
					 * request so just refresh the
					 * expiration time
					 */
					node.getCache().refresh(payloadID, 2);
					req = (DLRequestCapsule)
						pr.getCapsule();
					req.wantedChunkMask &= ~chunkMask;
				}
			}
			else
			{
				TRACE(Trace.ANTSR_DYNLOAD, "decode: unneeded DLResponse " + payloadID + " from " + NodeAddress.toString(getSrc()) + "(mask " + Integer.toHexString(chunkMask) + ")");
			}
		}

		return true;
	}
  

	/**
	 * Construct a blank DLResponseCapsule.
	 * <p>
	 * This is only invoked when a DLResponseCapsule capsule is received
	 * at a node and the node must construct a capsule before decoding
	 * the packet into it.
	 */
	PACKAGE DLResponseCapsule()
	{
	}

	/**
	 * Construct a DLResponseCapsule that will contain the given chunk
	 * of the given method.  Only invoked by
	 * DLRequestCapsule.evaluate().
	 *
	 * @param m The method which is being sent.
	 * @param idx the chunk id
	 * @param pos the position in the method's class file of the chunk start
	 * @param len the size of the chunk (<= CodeGroup.CHUNK_SIZE)
	 * @see DLRequestCapsule#evaluate(ants.Node)
	 */
	PACKAGE DLResponseCapsule(int dst, CapsuleClass cc, int idx, int pos, int len) 
	{
		setDst(dst);
		payloadID = cc.methodID;
		groupID = cc.group.groupID;
		groups = cc.group.groupIDs;
		chunkMask = cc.group.chunkMask(idx);
		clen = cc.codeSize();
		coff = pos;
		name = cc.name();

		code = new ByteArray(cc.code(), coff, len);
	}
	
	public String toString()
	{
		StringBuffer buf = new StringBuffer(100);

		buf.append(super.toString());
		buf.append("+DLResponseCapsule[payloadID=");
		buf.append(payloadID);
		buf.append("; groupID=");
		buf.append(groupID);
		buf.append("; chunkMask=");
		buf.append(Integer.toHexString(chunkMask));
		buf.append("; clen=");
		buf.append(clen);
		buf.append("; coff=");
		buf.append(coff);
		buf.append("; name=");
		buf.append(name);
		buf.append("]");
		return buf.toString();
	}
}
