/* 
 *  $Id: Node.Java,v 1.14 2000/08/30 21:55:45 tullmann Exp $
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_NODEOS_PKG.Address;
import JANOS_NODEOS_PKG.AddressFormatException;
import JANOS_NODEOS_PKG.BufferHandle;
import JANOS_NODEOS_PKG.NoSuchNodeAddress;
import JANOS_NODEOS_PKG.Flow;
import JANOS_NODEOS_PKG.InChannel;
import JANOS_NODEOS_PKG.OutChannel;
import JANOS_NODEOS_PKG.Credentials;
import JANOS_NODEOS_PKG.Thread;
import JANOS_NODEOS_PKG.ThreadPoolSpec;
import JANOS_NODEOS_PKG.MemSpec;
import JANOS_NODEOS_PKG.CommSpaceHandle;
import JANOS_NODEOS_PKG.CommQueue;
import JANOS_NODEOS_PKG.CommSpaceElement;
import JANOS_NODEOS_PKG.ProtocolSpecFormatException;
import JANOS_NODEOS_PKG.PacketAddressFormatException;

import JANOS_IO_PKG.Log;

import java.io.IOException;

import java.util.Hashtable;
import java.util.Enumeration;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * The per-flow Node interface.
 * <p>
 * An instance of Node is created in the context of each Application
 * and protocol.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public class Node
{
	private final int address;

	private final Hashtable outChannels = new Hashtable(12, 0.5f);

	private final CommSpaceHandle neighbors;
	PACKAGE final UnknownMidPackets unknownMidPackets;

	/**
	 * Maps groupsIDs to CodeGroups.
	 */
	private final Hashtable GIDtoCodeGroup = new Hashtable();

	/**
	 * Maps MIDs to CapsuleClasses.
	 */
	private final Hashtable MIDtoCapsuleClass = new Hashtable();

	/**
	 * Maps name to CapsuleClass.  This table is used to supply
	 * a PID or MID given the name of a CapsuleClass.
	 */
	private final Hashtable nameToCapsuleClass = new Hashtable(20,0.6f);

	/**
	 * XXX this is static.  That is bad, and wrong.  It needs to
	 * be a CommSpace so that flows can look applications up...
	 * <p>
	 * XXX probably should only allocate this on demand.
	 * <p>
	 * XXX MUCH CROSS-FLOW SHARING GOING ON THROUGH THIS TABLE.
	 */
	private static final Hashtable globalAppsTable = new Hashtable(5, 0.9f);

	/**
	 * A "list" of all the applications running on this Node (i.e.,
	 * in this Flow.
	 * <p>
	 * Optimized for the current case of exactly one app per node...
	 * <p>
	 * Will be null in protocol-based flows.
	 */
	private Application app = null;

	/**
	 * The per-flow NodeCache for leaving state between capsules.
	 */
	private NodeCache cache;

	private static int idMax = 0;

	/**
	 * Primitive node identifier hack.
	 */
	private final int id;

	/**
	 * Construct a node.  A node is only constructed by the boot
	 * routine in PrimordialNode, AppBooster or ProtocolBooster.
	 */
	PACKAGE Node(Flow flow,
		     CommSpaceHandle neighbors,
		     int address,
		     UnknownMidPackets unknownMidPackets)
	{
		// Put node object where it can be looked up.
		flow.setObject(this);

		synchronized(Node.class) 
		{
			this.id = idMax++;
		}

		this.address = address;
		this.neighbors = neighbors;
		this.unknownMidPackets = unknownMidPackets;
		this.cache = new NodeCache(this, 15);
		this.cache.collect();
		
		getNeighbors();
	}

	/**
	 * Return the logical 32-bit IP-like "active" address for this Node.
	 *
	 * @return the 32-bit IP-like "active" address of this Node
	 */
	public int getAddress()
	{
		return address;
	}

	public String getPhysicalProtocolSpec()
	{
		return PrimordialNode.getPhysicalProtocolSpec();
	}

	public String getPhysicalAddressSpec()
	{
		return PrimordialNode.getPhysicalAddressSpec();
	}

	/**
	 * Return the Node interface associated with the current flow.
	 * <p>
	 * XXX This method is a hack and should go away once we get
	 * free reign to re-write the basic ANTSR interfaces.  The
	 * "current node" should be an explicit paramter to those
	 * functions that need it.
	 */
	PACKAGE static Node currentNode()
	{
		return (Node)(Flow.currentFlow().getObject());
	}

	/**
	 * Lookup the neighbor table and construct outchans to
	 * each neighbor.  
	 *
	 */
	private void getNeighbors()
	{
		int index = 0;
		int logicalAddress = neighbors.read(index++).getInt();
		while (logicalAddress != 0)
		{
			Address physicalAddress =
				neighbors.read(index++).getAddress();

			addOutChannel(logicalAddress, physicalAddress);

			// Get logical from next pair...
			logicalAddress = neighbors.read(index++).getInt();
		}
	}

	private OutChannel addOutChannel(int logicalAddress,
				   Address physicalAddress)
	{
		TRACE(Trace.ANTSR_STARTUP, this.identityString()+ ".addOutChannel maps " +Address.toString(logicalAddress)+ " to " +physicalAddress.toString());

		OutChannel newChan;
		try
		{
			newChan = OutChannel.create(Flow.currentFlow(),
						    physicalAddress.getProtocolSpec(),
						    physicalAddress.getAddressSpec());
		}
		catch (java.io.IOException ioe)
		{
			panic("OutChannel() threw " +ioe);
		}
		catch (PacketAddressFormatException e)
		{
			panic("OutChannel() threw " +e);
		}
		catch (ProtocolSpecFormatException e)
		{
			panic("OutChannel() threw " +e);
		}
		outChannels.put(new Integer(logicalAddress),
				newChan);
		return newChan;
	}

	/**
	 * Called by applications that want to instantiate a
	 * locally originiating protocol.
	 * <p>
	 * This is a public interface that we really can't change.
	 */
	public void register(Protocol p)
	{
		TRACE(Trace.ANTSR_DYNLOAD, this.identityString()+ ": Installing protocol " +p);
		try
		{
			// Install the Protocol in its own flow.
			p.activate(this, true);
		}
		catch (Exception e)
		{
			FATALEXCEPTION(e, "gak.");
		}
	}

	/**
	 * Deliver the given capsule to the Application on the
	 * given port number.
	 * <p>
	 * XXX access control?
	 *
	 * @exception NoSuchApplicationError if the port doesn't have an app listening on it.
	 */
	public boolean deliverToApp(Capsule cap, int port)
	{
		Application app = (Application) globalAppsTable.get(new Integer(port));
		if (app == null)
			throw new NoSuchApplicationError("There is no application registered on port " +port);

		// XXX THIS CROSSES FLOW BOUNDARIES.  THIS MIGHT BE BAD.
		cap.enqueueCapsule(app.getPacketQueue());
		return true;
	}

	/**
	 * Inject a given capsule into the network via the local node.  
	 * This method is only called by Application#send().
	 * Note that the capsule's evaluate() is responsible for routing...
	 */
	public boolean send(Capsule cap)
	{
		DEBUGONLY(Flow f = Flow.currentFlow());
		assert(this == f.getObject());

		return cap.evaluate(this);
	}

	public void threadStart(java.lang.Runnable r)
	{
		threadStart(r, "Another Underappreciated Thread");
	}

	public void threadStart(java.lang.Runnable r, String threadName)
	{
		require(r);
		require(Flow.currentFlow());
		require(this == Flow.currentFlow().getObject());

		Flow.currentFlow().threadStart(r, threadName);
	}

	/**
	 * Route a capsule for a particular destination.
	 * Note that there is no next hop to the current node, and it is
	 * therefore impossible to route to ourself.  Since capsules are
	 * evaluated at the local node before being sent, it seems like a bad
	 * idea to allow them to route to the same node -- this would likely
	 * produce an infinite loop.
	 *
	 * XXX should be "sendToNode()"
	 */
	public boolean routeForNode(Capsule cap, int destination)
	{
		DEBUGONLY(Flow f = Flow.currentFlow());
		assert(this == f.getObject());

		Xdr xdr;

		TRACE(Trace.ANTSR_OUTGOING, "Sending capsule " +cap+ " to " +Address.toString(destination));

		assert(destination != 0);

		// Encode the capsule.
		xdr = cap.encodeCapsuleForSend(this);
		
		int target;
		
		if( destination == this.getAddress() )
		{
			TRACE(Trace.ANTSR_OUTGOING, "  Warning: Capsule is being routed to this node");
			return false;
		}
		else
		{
			// look up nearest neighbor on the way to destination
			target = PrimordialNode.nextHopToNode(destination);
		}
		
		if (target == 0) // XXX what to do?
		{
			TRACE(Trace.ANTSR_OUTGOING, "  Capsule send failed.  No hop to " +Address.toString(destination));
			return false;
		}
		
		TRACE(Trace.ANTSR_OUTGOING, "  Capsule being forwarded to " +Address.toString(target));
		

		// push out the appropriate out-channel
		Object oChanObj = outChannels.get(new Integer(target));
		// if the out channel doesn't already exist, then create it
		if (oChanObj == null)
		{
			Address addr;
			// Need to get the physical address
			addr = PrimordialNode.nextPhysicalToNode(target);
			oChanObj = addOutChannel(target,addr);
		}
		assert(oChanObj != null);

		// XXX fix when XDR is fixed!
		BufferHandle bh = BufferHandle.create(0, xdr.length());
		bh.setBytes(0, xdr.getBytes(), 0, xdr.length());

		TRACE(Trace.ANTSR_OUTGOING, "  Pushing buf " +Debug.dbgPacketSummary(bh, 128));
		((OutChannel)oChanObj).send(bh);

		return true;
	}

	/**
	 * Publish the given CapsuleClass in both the name -> cc map and the
	 * MID -> cc map.
	 *
	 * @param cc the CapsuleClass to publish.  Must have a name and methodID defined.
	 * @exception DuplicateCapsuleClassException if a CC is already registered under the cc's name.
	 */
	PACKAGE void publishCapsuleClass(CapsuleClass cc)
		throws DuplicateCapsuleClassException
	{
		DEBUGONLY(Flow f = Flow.currentFlow());
		assert(this == f.getObject());

		Object r;
		assert(cc);
		assert(cc.name());
		assert(cc.methodID);

		TRACE(Trace.ANTSR_DYNLOAD|Trace.ANTSR_MID, this.identityString()+ ": publishing " +cc);
		synchronized(nameToCapsuleClass)
		{
			// Publish in the name -> cc map
			r = nameToCapsuleClass.put(cc.name(), cc);
			if (r != null)
			{
				Object o;
				// put the original element back
				o = nameToCapsuleClass.put(cc.name(), r);
				assert(o == cc);
				throw new DuplicateCapsuleClassException("Tried to re-register " + cc + ".");
			}
		}
		TRACE(Trace.ANTSR_MID, this.identityString()+ ": published " +cc);

		// Publish in the mid -> cc map
		r = MIDtoCapsuleClass.put(cc.methodID, cc);
		assert(r == null);
	}

	PACKAGE void publishCodeGroup(CodeGroup group)
	{
		DEBUGONLY(Flow f = Flow.currentFlow());
		assert(this == f.getObject());

		Object rc;
		
		TRACE(Trace.ANTSR_DYNLOAD, this.identityString() + ": Publishing " +group);
		rc = GIDtoCodeGroup.put(group.groupID, group);
		assert(rc == null);  // no duplicates, please.
	}

	PACKAGE TypeID findMID(String name)
	{
		DEBUGONLY(Flow f = Flow.currentFlow());
		assert(this == f.getObject());

		TRACE(Trace.ANTSR_MID, this.identityString()+ ".findMid(name=" +name+ ")...");

		CapsuleClass cc;
		cc = (CapsuleClass)nameToCapsuleClass.get(name);
		if (cc == null)
			panic("No CapsuleClass registered under " +name+ " (in findMID()).");

		assert(cc.methodID);

		TRACE(Trace.ANTSR_MID, this.identityString()+ ".findMid(name=" +name+ ") -> " +cc.methodID);
		return cc.methodID;
	}

	PACKAGE TypeID findPID(String name)
	{
		DEBUGONLY(Flow f = Flow.currentFlow());
		assert(this == f.getObject());

		CapsuleClass cc;
		cc = (CapsuleClass)nameToCapsuleClass.get(name);
		if (cc == null)
			panic("No CapsuleClass registered under " +name+ " (in findPID()).");

		assert(cc.protocolID());
		return cc.protocolID();
	}

	/**
	 * Find a CapsuleClass for the given MID.  CapsuleClasses are
	 * registered by <code>publishCapsuleClass</code>.
	 *
	 * @see #publishCapsuleClass
	 */
	PACKAGE CapsuleClass lookupCapsuleClass(TypeID mid)
	{
		DEBUGONLY(Flow f = Flow.currentFlow());
		assert(this == f.getObject());

		TRACE(Trace.ANTSR_DYNLOAD, this.identityString()+ ": cc for mid " +mid+ " -> " +MIDtoCapsuleClass.get(mid));

		return (CapsuleClass) MIDtoCapsuleClass.get(mid);
	}

	PACKAGE CodeGroup lookupCodeGroup(TypeID gid)
	{
		DEBUGONLY(Flow f = Flow.currentFlow());
		assert(this == f.getObject());

		TRACE(Trace.ANTSR_DYNLOAD, this.identityString()+ ": CodeGroup for gid " +gid+ " -> " +GIDtoCodeGroup.get(gid));
		return (CodeGroup) GIDtoCodeGroup.get(gid);
	}

	/**
	 * Attach the given "app" object at its port.
	 * An Exception (AttachException) is thrown if some app
	 * is already registered at the app's port.
	 * <p>
	 * If the app's port is zero then an available port number
	 * will be allocated for the application.
	 * <p>
	 * The app provided must be an app running on the current Node
	 * (i.e., in the current Flow).  The app will automatically be
	 * de-registered when this Node is shutdown.
	 */
	PACKAGE void attachApplication(Application app)
		throws AttachException
	{
		Integer key = new Integer(app.getPort());

		/**
		 * Find an available port number for this app.
		 */
		if (app.getPort() == 0)
		{
			int i = 128; // XXX?
			Object o = this; // XXX just something non-null

			while ((i > 0) && (o != null))
			{
				key = new Integer(i);

				// XXX Race condition if occupied
				o = globalAppsTable.put(key, app);
				if (o != null)
				{
					// oops, someone already there...
					globalAppsTable.put(key, o);
				}
				else
				{
					app.setPort(i);
				}
				i--;
			}
		}
		else
		{
			// App has a specific, hard-coded port number

			if (globalAppsTable.containsKey(key))
				throw new AttachException("Port " +app.getPort()+ " is busy.");

			globalAppsTable.put(key, app);
		}

		assert(this.app == null); // XXX can only have one app per Node (right now).
		this.app = app;
	}

	PACKAGE void detachApplication(Application app)
		throws AttachException
	{
		Integer key = new Integer(app.getPort());
		
		if (!globalAppsTable.containsKey(key))
		{
			throw new AttachException("Port " +app.getPort()+
						  " cannot be detached (no app registered).");
		}

		Object f = globalAppsTable.remove(key);
		assert(f == app); // XXX, re-insert and throw error...
	}

	public NodeCache getCache()
	{
		DEBUGONLY(Flow f = Flow.currentFlow());
		assert(this == f.getObject());

		return this.cache;
	}

	public void sleep(int millis)
		throws InterruptedException
	{
		DEBUGONLY(Flow f = Flow.currentFlow());
		assert(this == f.getObject());

		JANOS_NODEOS_PKG.Thread.sleep(millis, 0);
	}

	
	/**
	 * Shutdown this Node.  Just kills the flow.
	 * <p>
	 * Interface for shutting down the entire node
	 * isn't (yet) visible.
	 */
	public void shutdown()
	{
		Flow f = Flow.currentFlow();
		assert(this == f.getObject());

		this.cache.stopCollecting();
		// If an app is registered by this Node, remove it
		if (this.app != null)
		{
			try
			{
				detachApplication(app);
			}
			catch (AttachException ae)
			{
				log(Log.WARNING, "Could not unregister app "
				    +this.app+
				    "when shutting down Node " +this);
			}

			// If a one-trick show, shut down the whole VM
			if (PrimordialNode.oneTrick())
				PrimordialNode.shutdown();
		}

		f.destroy();
	}

	/**
	 * Return the string name of the host machine.
	 */
	public String hostname()
	{
		DEBUGONLY(Flow f = Flow.currentFlow());
		assert(this == f.getObject());

		return JANOS_NODEOS_PKG.Node.hostname();
	}
	
	/**
	 * Return the current time in milliseconds.
	 */
	public long time()
	{
		DEBUGONLY(Flow f = Flow.currentFlow());
		assert(this == f.getObject());

		return JANOS_NODEOS_PKG.Node.time();
	}

	public void log(int level, String msg)
	{
		DEBUGONLY(Flow f = Flow.currentFlow());
		assert(this == f.getObject());

		PrimordialNode.message(level, msg);
	}

	public void log(int level, Throwable exception)
	{
		logStackTrace(level, exception);
	}

	public void logStackTrace(int level, Throwable exception)
	{
		DEBUGONLY(Flow f = Flow.currentFlow());
		assert(this == f.getObject());

		PrimordialNode.dumpStackTrace(exception);
	}

	public RouteTable getRouteTable()
	{
		DEBUGONLY(Flow f = Flow.currentFlow());
		assert(this == f.getObject());
		// XXX security
		return PrimordialNode.getRouteTable();
	}

	/**
	 * Return the host-local identity of this "node".  Currently
	 * just the integer id of the instance.  Should eventually
	 * be an identity associated with the Protocol/User.
	 */
	public int identity() 
	{
		return this.id;
	}

	/**
	 * Return a String that identifies this Node.  Useful for
	 * debugging traces.
	 */
	public String identityString()
	{
		require(Flow.currentFlow() != null);
		require(Flow.currentFlow().getObject() == this);

		if (DEBUG_ON) 
		{
			return "Node[id=" +this.identity()+ "; flow=" 
				DEBUGONLY(+"\""+ Flow.currentFlow().getName() + "\"")
				+"]";
		}
		else
		{
			return "Node[id=" +this.identity()+ "; flow=" 
				+Flow.currentFlow()+ "]";
		}
	}

	public String toString()
	{
		return "Node[a=" + Address.toString(this.address) + "]";
	}
}
