/* 
 * NodeConfigure.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 * 
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_ANTSR_UTIL_PKG.KeyArgs;
import JANOS_ANTSR_UTIL_PKG.InvalidKeyArgsException;
import JANOS_ANTSR_UTIL_PKG.LineListTokenizer;
import JANOS_ANTSR_UTIL_PKG.LineListException;

import JANOS_NODEOS_PKG.Address;
import JANOS_NODEOS_PKG.AddressFormatException;
import JANOS_NODEOS_PKG.NoSuchNodeAddress;
import JANOS_NODEOS_PKG.BufferHandle;
import JANOS_NODEOS_PKG.CommSpaceHandle;
import JANOS_NODEOS_PKG.Flow;
import JANOS_NODEOS_PKG.FlowRun;

import java.io.FileNotFoundException; // XXX
import java.io.IOException; // XXX

import java.util.Vector; // XXX

IMPORT_DEBUG
IMPORT_TRACE

/**
 * Performs one-time configuration of a Node to get it on its feet.
 * <p>
 * An alternative to the MIT ANTS ConfigurationManager.
 * <p>
 * Usage:
 * <code>NodeConfigure <ActiveAddress> <PhysicalAddress> [options]</code>
 * <menu>
 *   <li> <b>ActiveAddress</b>: the active address of the Node.  Must be an IP-like dotted-quad.
 *   <li> <b>PhysicalAddress</b>: the active address of the Node.  Must be a <var>host:port</var> pair.
 * </menu>
 * Options:
 * <menu>
 *   <li> <b>-consoleport <int></b>: Start a telnet console on the given port.
 *   <li> <b>-routes <file></b>: Static routing table is in <var>file</var>.
 *   <li> <b>-updateRoutes</b>: If dynamic routing updates should be pushed out.
 *   <li> <b>-application <appfile></b>: Read <var>appfile</var> to configure & start an application.
 *   <li> <b>-extension <extfile></b>: Read <var>extfile</var> to configure and install an extension.
 *   <li> <b>-msgprefix <string></b>: Prefix all messages with <string>
 * </menu>
 * The files <var>appfile</var> and <var>extfile</var> are simple files that are unceremoniously
 * pushed into KeyArgs.  They should define the application.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public class NodeConfigure
	extends FlowRun
{
	/**
	 * XXX A hack.  Move into PrimordialNode so that ConfigurationManager and NodeConfigure
	 * will be mutually exclusive.
	 */
	private static boolean beenHere = false;

	/**
	 * Configure this Node.  Prints a usage message if
	 * insufficient arguments are given.
	 * <p>
	 * The flow that starts this ball rolling will become the basic
	 * flow in the system.
	 * <p>
	 * XXX flow argument isn't used anymore?
	 *
	 * @param args the node configuration.  See the class documentation.
	 */
	protected void run(Flow flow,
			   String[] args,
			   CommSpaceHandle commSpaceHandle)
	{
		require(flow == Flow.currentFlow());

		if (beenHere)
		{
			// XXX throw new AlreadyInitializedException() ??
			panic("Already initialized.  Don't be doing this.");
		}
		beenHere = true;
		
		DEBUGONLY(JANOS_NODEOS_PKG.Flow.currentFlow().setName("Primordial ANTSR Flow"));

		if ((args == null) || (args.length < 2))
		{
			PrimordialNode.message("usage: " 
					       +NodeConfigure.class.getName()+
					       " <activeaddr> <physaddr> [options].");
			System.exit(1);
		}
    
		/*
		 * Parse the Node configuration, find what apps to start (if any)
		 * what extensions to install (if any), and then start the 
		 * PrimordialNode.
		 */
		try
		{
			parseCommandLine(args);
		}
		catch(Exception e)
		{
			FATALEXCEPTION(e, "Bad configuration (node " +args[0]+ " " +args[1]+ ").");
		}
	}

	/**
	 * Parse the command line arguments to NodeConfigure.  Most
	 * arguments are passed into PrimordialNode in some fashion.
	 * However, -application and -extension arguments have their
	 * files parsed and put into the <code>apps</code> and
	 * <code>exts</code> return parameters.
	 */
	private void parseCommandLine(String[] args)
		throws FileNotFoundException,
		IOException,
		InvalidKeyArgsException,
		AddressFormatException,
		NoSuchNodeAddress,
		InterruptedException,
		LineListException
	{
		Vector apps = new Vector();
		Vector exts = new Vector();
		String physAddr;
		String activeAddr;
		String host;
		String transport = "udp";
		int port;
		KeyArgs primNodeArgs;
		int i;

		assert(args.length >= 2);
		assert(apps);
		assert(exts);
		
		// Grab two fixed args
		activeAddr = args[0];
		physAddr = args[1];

		primNodeArgs = new KeyArgs();
		
		i = 2;
		while (i < args.length)
		{
			if (args[i].equals("-application"))
			{
				String[] appargs;
				String fname;
				i++;
				assert(i < args.length); // XXX

				fname = args[i];

				try
				{
					appargs = grokArgFile(fname);
					apps.addElement(new AppConfig(appargs[0], appargs, 1));
				}
				catch (Exception e)
				{
					FATALEXCEPTION(e, "groking -application arg " +fname);
				}
			}
			else if (args[i].equals("-extension"))
			{
				String[] extargs;
				String fname;
				i++;
				assert(i < args.length); // XXX
				fname = args[i];
				
				try
				{
					extargs = grokArgFile(fname);
					exts.addElement(extargs); // XXX XXX XXX
				}
				catch (Exception e)
				{
					FATALEXCEPTION(e, "groking -extension arg " +fname);
				}
			}
			else if (args[i].equals("-updateRoutes"))
			{
				primNodeArgs.set(args[i], null);
			}
			else
			{
				String arg;
				String param;
				arg = args[i];
				i++;
				assert(i < args.length); // XXX
				param = args[i];

				primNodeArgs.set(arg, param);
			}
			i++;
		}

		int sep = physAddr.indexOf(':');
		if (sep > 0)
		{
			int tsep = physAddr.indexOf(':', sep + 1);
			try
			{
				host = physAddr.substring(0,sep);
				port = Integer.parseInt(physAddr.substring(sep+1,
									   (tsep > 0) ?
									   tsep :
									   physAddr.length()));
				if( tsep > 0 )
					transport = physAddr.
						substring(tsep+1,
							  physAddr.length());
				// above might throw NumberFormatException
			}
			catch (Exception e)
			{
				FATALEXCEPTION(e, "Problem parsing physical address: ");
			}
		}
		else
		{
			panic("No port (or no host) provided in physical address '" +physAddr+ "'");
		}

		TRACE(Trace.ANTSR_STARTUP, "Node Configuration:");
		TRACE(Trace.ANTSR_STARTUP, "  ANTSR address    : " +activeAddr);
		TRACE(Trace.ANTSR_STARTUP, "  physical address: " +physAddr);
		TRACE(Trace.ANTSR_STARTUP, "  node args: " +primNodeArgs);

		if (apps.size() > 0)
		{
			for(i = 0; i < apps.size(); i++)
			{
				TRACE(Trace.ANTSR_STARTUP, "  app: " +apps.elementAt(i));
			}
		}
		else
		{
			TRACE(Trace.ANTSR_STARTUP, "  [no applications]");
		}
		
		if (exts.size() > 0)
		{
			for(i = 0; i < exts.size(); i++)
			{
				TRACE(Trace.ANTSR_STARTUP, "  extension: " +exts.elementAt(i));
			}
		}
		else
		{
			TRACE(Trace.ANTSR_STARTUP, "  [no extensions]");
		}
		

		PrimordialNode.setAddress(Address.fromString(activeAddr));
		PrimordialNode.setPhysicalAddress("if/ipv4/" +transport,
						  "*:" +port+ ":" +host);
		PrimordialNode.setArgs(primNodeArgs);

		PrimordialNode.start(apps, exts);

	}

	/**
	 * Grok an arg file (for a weak definition of grok).  Just
	 * turns the contents of the file into an array of Strings,
	 * one for each token.
	 */
	private static String[] grokArgFile(String filename)
		throws FileNotFoundException,
		IOException,
		LineListException
	{
		LineListTokenizer lt;
		String[] line;
		String[] buf = new String[64];
		String[] ret;
		int start;
		int end;
		
		start = 0;
		end = buf.length;

		lt = new LineListTokenizer(filename);
		do
		{
			line = lt.list();

			if (line != null)
			{
				assert((start + line.length) < end);
#if 0
				if ((start + line.length) > end)
				{
					String nbuf = new String[(buf.length)*2];

					// copy whole array to nbuf
					System.arraycopy(buf, 0,
							 nbuf, 0, end);
					end = nbuf.length;
					buf = nbuf;
				}
#endif
				// Copy line array in
				System.arraycopy(line, 0,
						 buf, start, line.length);
				start += line.length;
			}
		}
		while (line != null);

		// Dont return all 64 arguments
		ret=new String[start];
		System.arraycopy(buf,0,ret,0,start);
		return ret;
	}


	/**
	 * Part of the FlowRun interface
	 *
	 * @see core.FlowRun
	 * @deprecated Just use BufferHandle.create()
	 */
	protected BufferHandle newBufferHandle()
	{
		return BufferHandle.create();
	}

	/**
	 * Part of the FlowRun interface
	 *
	 * @see core.FlowRun
	 * @deprecated Just use BufferHandle.reset()
	 */
	protected void recycleBufferHandle(BufferHandle bh)
	{
		bh.reset();
	}

	/**
	 * This doesn't work.  Its just here to prevent folks from getting
	 * really confused.
	 *
	 * @see #run()
	 */
	public static void main(String[] args)
		throws Exception
	{
		PrimordialNode.message("NodeConfigure must be started as a Janos Flow.");
		System.exit(11);
	}
}
