/*
 * ProtocolBooster.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_NODEOS_PKG.Address;
import JANOS_NODEOS_PKG.BufferHandle;
import JANOS_NODEOS_PKG.CommQueue;
import JANOS_NODEOS_PKG.DemultiplexKey;
import JANOS_NODEOS_PKG.Flow;
import JANOS_NODEOS_PKG.FlowRun;
import JANOS_NODEOS_PKG.InChannel;
import JANOS_NODEOS_PKG.InChannelHandle;
import JANOS_NODEOS_PKG.CommSpaceHandle;
import JANOS_NODEOS_PKG.CommSpaceElement;
import JANOS_IO_PKG.Log;

import java.io.IOException;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * This is invoked by PrimordialNode as the intial class in a new
 * flow.  This new flow is charged with taking the new (unknown
 * protocol) packet, requesting and loading the code associated with
 * that packet and then processing the packet.
 * <p>
 * CommSpaceHandle should point to a valid CSH for
 * the Booster class, plus:
 * <pre>
 *   [2] a byte[].  The PID of the protocol this Booster is working for.  XXX Not Used
 *   [3] an integer -- the logical ANTSR address of the node that sent the packet.
 *   [4] a slot to write an InChannelHandle into.
 * </pre>
 * <p>
 * XXX Should be name DynamicProtocolBooster (As opposed to an
 * app-loaded Protocol).
 */
public class ProtocolBooster
	extends Booster
{
	/**
	 * The booster communicates with the Primordial Node through a
	 * CommQueue so that it can figure out what protocols it is supposed
	 * to be handling.  The booster can either install a new mid through
	 * dynamic loading, or load up a locally defined protocol.  To
	 * distinguish between them the first element pushed onto the queue
	 * is an integer indicating what it is supposed to do, followed by
	 * any needed data.
	 */

	/**
	 * Indicates a new MID, arguments are the MID bytes and the previous
	 * node that the packet was at.
	 */
	public final static int NEW_MID = 0;

	/**
	 * Indicates a local protocol should be instantiated, the only argument
	 * is the name of the protocol class.  Once this executes the protocol
	 * message handler thread is basically useless since the whole
	 * protocol is active.
	 */
	public final static int LOCAL_PROTOCOL = 1;

	public final static int APP_CAPSULE = 2;
	
	private boolean local = false;
	
	public ProtocolBooster(JANOS_NODEOS_PKG.ClassInstantiationKey key)
		throws SecurityException
	{
		key.checkEquals(ProtocolBooster.class);
	}

	public void run(final Flow flow,
			String[] args,
			CommSpaceHandle commSpaceHandle)
	{
		final CommQueue queue;
		final Node node;

		require(flow);
		require(commSpaceHandle);
		require(args.length == 0); // XXX not expecting any

		DEBUGONLY(flow.setName("Flow for *Unknown Protocol*"));
		TRACE(Trace.ANTSR_DYNLOAD, "Starting unknown protocol. flow=" +flow);

		// Turn on the package-level access controls for this flow.
		flow.enablePackageControls(true);
		flow.allowPackage("java.lang");
		flow.allowPackage(JANOS_ANTSR_CORE_PKG_S);
		flow.allowPackage(JANOS_ANTSR_ROUTE_PKG_S);	// XXX
		flow.allowPackage(JANOS_ROUTE_PKG_S);		// XXX
		flow.allowPackage("java.util");			// XXX

		// setupLocalNode is in Booster.
		node = setupLocalNode(flow, commSpaceHandle);

		// Dedicate a thread to handling new MIDs:
		queue = commSpaceHandle.read(3).getCommQueue();
		flow.threadStart(new Runnable() {
			public void run() {
				msgLoop(flow, queue, node);
			}
		},
			"MID/Protocol Handler for "
			+ TypeID.wrap(commSpaceHandle.read(2).
				      getByteArray()).toString());
		
		TRACE(Trace.ANTSR_DYNLOAD, "Unknown protocol on its way! " +node.identityString());
	}

	/**
	 * This will wait for and process any messages coming in on the
	 * CommQueue
	 */
	private void msgLoop(Flow flow, CommQueue queue, Node node)
	{
		while(!flow.destroyed())
		{
			try
			{
				/* Dispatch any messages we get */
				int message =
					queue.waitToTake().getInt();
				this.dispatch(flow, message, queue, node);
			}
			catch(Exception e) // dispatch() throws 6 kinds
			{
				node.log(Log.ERR,
					 "Error starting new mid in "
					 + Flow.currentFlow()
					 + ": "
					 + e.getMessage());
				node.logStackTrace(Log.ERR, e);
			}
		}
	}

	/**
	 * This will actually dispatch a message in the queue
	 *
	 * @param flow The current flow
	 * @param message The integer identifier for the message
	 * @param queue The queue the message was sent on
	 * @param node The local node object
	 */
	private void dispatch(Flow flow, int message, CommQueue queue, Node node)
		throws java.lang.InterruptedException,
		java.lang.ClassNotFoundException,
		java.lang.InstantiationException,
		java.lang.IllegalAccessException,
		FullCodeGroupException,
		DuplicateCapsuleClassException
	{
		switch(message)
		{
		case NEW_MID:
			byte[] midBytes =
				queue.waitToTake().getByteArray();
			byte[] gidBytes =
				queue.waitToTake().getByteArray();
			int prev = queue.waitToTake().getInt();
			
			if( local )
				break;
			
			TypeID mid = TypeID.wrap(midBytes);
			TypeID gid = TypeID.wrap(gidBytes);
			
			TRACE(Trace.ANTSR_DYNLOAD, node.identityString()+ " got MID " +mid+ " through CommQueue.");
			startNewGid(node, gid, mid, prev);
			break;
		case LOCAL_PROTOCOL:
			String protocolName =
				new String(queue.waitToTake().getCharArray());
			Protocol pr;
			Class cl;
			String pkgName;

			if( local )
				break;
			
			DEBUGONLY(flow.setName(protocolName));

			// Allow the package to be loaded off the disk.
			pkgName = protocolName.
				substring(0, protocolName.lastIndexOf('.'));
			flow.allowPackage(pkgName);

			/* Load and start it up */
			TRACE(Trace.ANTSR_DYNLOAD, node.identityString()+ " loading LOCAL_PROTOCOL " +protocolName);
			cl = flow.loadClass(protocolName);
			pr = (Protocol)cl.newInstance();
			pr.activate(node, false);
			this.local = true;
			break;
		case APP_CAPSULE:
			Capsule cap = Capsule.dequeueCapsule(node, queue);
			CommQueue replyQueue = queue.waitToTake().
				getCommQueue();
			boolean sent;

			sent = cap.evaluate(node);
			replyQueue.enqueue(new CommSpaceElement(sent ? 1 : 0));
			break;
		default:
			node.log(Log.ERR,
				 "Invalid message sent to protocol "
				 + "message handler");
			break;
		}
	}

	// It's possible for a MID to get started twice.  But this
	// is probably ok, and maybe even necessary: if the first
	// installation fails halfway through, a second installation
	// has to complete it.
	private void startNewGid(Node node, TypeID gid,
				 TypeID mid, int prev)
	{
		// Setup a Key for receving DLReply packets...
		new CapsuleChannel(DLResponseCapsule.class,
			new CodeRequestDemultiplexKey(
				Flow.currentFlow(),
				DLResponseCapsule.MID,
				gid.unwrap()));

		// Construct a DLRequest packet and send it to prevNode.
		DLRequestCapsule req = new DLRequestCapsule(prev, mid);
		req.prime(node.getAddress(), 1);
		node.routeForNode(req, prev);
		/* Add a pending request for this DL request */
		node.getCache().put(gid,
				    new PendingRequest(req),
				    2);
	}
}
