/*
 * UnknownPacketHandler.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_NODEOS_PKG.BufferHandle;
import JANOS_NODEOS_PKG.DeliverFunc;
import JANOS_NODEOS_PKG.DemultiplexKey;
import JANOS_NODEOS_PKG.Flow;
import JANOS_NODEOS_PKG.InChannel;
import JANOS_NODEOS_PKG.ProtocolSpecFormatException;
import JANOS_NODEOS_PKG.PacketAddressFormatException;

import java.io.IOException;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * Handle an ANTSR packet of unknown MID
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE final class UnknownPacketHandler
	implements DeliverFunc
{
	private final Node node;
	private final DemultiplexKey classifier;
	private final java.util.Random random = new java.util.Random(); // XXX
	private final InChannel inChan;

	private static final float RANDOM_DUPMIDREQ_FRACTION = 0.05F;

	/**
	 * Create the UnknownPacketHandler.  Only need one of these
	 * per node...
	 */
	PACKAGE UnknownPacketHandler(Node node)
	{
		this(node, new ANEPDemultiplexKey(Flow.currentFlow(),
						  Capsule.ANTSRTYPE));
	}
	
	PACKAGE UnknownPacketHandler(Node node, DemultiplexKey classifier)
	{
		this.node = node;
		this.classifier = classifier;
		try
		{
			inChan = InChannel.create(
				Flow.currentFlow(),
				this.classifier,
				PrimordialNode.getPhysicalProtocolSpec(),
				PrimordialNode.getPhysicalAddressSpec(),
				this);
			inChan.addBuffers(30);
		}
		catch(PacketAddressFormatException e)
		{
			panic("Bad physical address spec");
		}
		catch(ProtocolSpecFormatException e)
		{
			panic("Bad physical protocol spec");
		}
		catch(IOException e)
		{
			panic("IOException when creating in channel " + e);
		}
	}

	/**
	 * Unknown packet handler.  For ANTSR packets that don't match
	 * an installed MID.
	 * <p>
	 * This is the only method defined by the DeliverFunc interface.
	 *
	 * @see DeliverFunc
	 */
	public final void receive(BufferHandle bufh, int payloadStart)
	{

		TRACE(Trace.ANTSR_DYNLOAD, "Received unknown ANTSR packet " +Debug.dbgPacketSummary(bufh, 8));

		if( bufh.getLength() < (payloadStart + Xdr.INT + Xdr.TYPEID) )
		{
			// XXX log this event.
			TRACE(Trace.ANTSR_DYNLOAD, "Unknown packet is bad: length < " +(payloadStart + Xdr.INT + Xdr.TYPEID)+ " bytes");

			// Recycle the bufh
			inChan.addBuffer(bufh);
			
			return;
		}
		
		/*
		 * By definition this receive() function only
		 * gets packets for an unknown ANTSR protocol.
		 */

		byte[] midbuf = new byte[Xdr.TYPEID];
		bufh.getBytes(payloadStart + Xdr.INT, midbuf, 0, 16);
		TypeID mid = TypeID.wrap(midbuf).intern();
		assert(mid.intern() == mid);

		// XXX don't know what will happen with bufh (most
		// likely will get queued up in th unknownMidPackets map)
		// So add a replacement buffer to the inChan
		inChan.addBuffer(BufferHandle.create());

		// XXX create a new BufferHandle that subarrays from payloadStart on?
		payloadStart += Xdr.INT + Xdr.TYPEID;

		TRACE(Trace.ANTSR_DYNLOAD, "handling unknown mid (" +mid+ ")");

		// XXX loses payloadstart...
		boolean midAlreadySeen = PrimordialNode.unknownMidPackets
			.addPacket(mid.unwrap(), bufh);

		if(midAlreadySeen)
		{
			/*
			 * This mid request has already been sent so we don't
			 * have to worry about it since the pending request
			 * mechanism "should" take care of it.
			 */
			return;
		}

		TRACE(Trace.ANTSR_DYNLOAD, "Sending MIDIDRequest for mid (" + mid + ")");

		/*
		 * XXX this is hackish: For convenience, pull out the
		 * address of the node that this packet came from.
		 * The new flow will need the info for firing off a
		 * DLRequest capsule...
		 *
		 * Format after mid is:
		 * src (int), dst (int), and previous (int)
		 */
		int prev = bufh.getInt(payloadStart + 4 + 4);
		
		MIDIDRequestCapsule req = new MIDIDRequestCapsule(prev, mid);
		
		/*
		 * Setup the req packet to go back to whence it
		 * came.
		 */
		req.prime(node.getAddress(), 1);

		node.routeForNode(req, prev);
		/* Add a pending request for this mid request */
		node.getCache().put(mid,
				    new PendingRequest(req),
				    2);
	}

	public String toString()
	{
		return "UnknownPacketHandler for " +node;
	}
}
