/*
 * DatagramLogListener.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_IO_PKG;

import java.io.IOException;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.UnknownHostException;

/**
 * A log listener that tracks the log level and can reply to StatusMessage
 * queries about the level.  This is only really useful for getting a quick
 * summary of the status of a node.
 */
public class DatagramLogListener
	implements LogListener
{
	/**
	 * The localhost InetAddress
	 */
	private static InetAddress localAddr;

	/**
	 * The level we reply to queries with
	 */
	private int currentLevel = 0;

	static
	{
		try
		{
			localAddr = InetAddress.getLocalHost();
		}
		catch(UnknownHostException e)
		{
		}
	}

	/**
	 * The server that replies to queries about the status of the node
	 */
	private class SocketListener
		implements Runnable
	{
		/**
		 * Listener socket
		 */
		private DatagramSocket ds;
		
		/**
		 * Reply message
		 */
		private StatusMessage outMessage = new StatusMessage();

		/**
		 * Incoming message
		 */
		private StatusMessage inMessage = new StatusMessage();

		/**
		 * Construct a listener for a certain port
		 */
		public SocketListener(int port)
			throws Exception
		{
			this.ds = new DatagramSocket(port);
		}

		/**
		 * Process an incoming packet
		 */
		public void processPacket()
		{
			if( inMessage.getDatagramPacket().getLength() <
			    inMessage.getValueOffset() )
			{
				return;
			}
			switch( inMessage.getMethodID() )
			{
			case StatusMessage.SMM_QUERY:
				/* Build the reply */
				outMessage.setMethodID(StatusMessage.
						       SMM_QUERY);
				outMessage.setFlags(StatusMessage.SMF_REPLY);
				StatusMessage.encodeInt(
					outMessage.getValue(),
					outMessage.getValueOffset(),
					currentLevel);
				outMessage.setValueSize(4);
				outMessage.setAddress(inMessage.getAddress());
				outMessage.setPort(inMessage.getPort());
				try
				{
					/* Send it back to the querier */
					this.ds.send(outMessage.
						     getDatagramPacket());
				}
				catch(IOException e)
				{
				}
				break;
			}
		}
		
		public void run()
		{
			/**
			 * Loop forever processing requests
			 */
			while( true )
			{
				try
				{
					this.ds.receive(this.inMessage.
							getDatagramPacket());
					this.processPacket();
				}
				catch(IOException ioe)
				{
				}
			}
		}
	}

	/**
	 * Start a listener on the given port
	 */
	public void start(int port)
		throws Exception
	{
		Thread th;

		th = new Thread(new SocketListener(port));
		th.start();
	}

	/**
	 * Receive the log message, we just record the level for now.
	 */
	public void logMsg(int level, String msg)
	{
		/*
		 * XXX This sucks, we should be more intelligent about the
		 * level
		 */
		this.currentLevel = level;
	}
}
