/*
 * AddressSpecPortAddr.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.net.InetAddress;
import java.net.UnknownHostException;

import java.util.StringTokenizer;
import java.util.NoSuchElementException;

/**
 * Provides a ip:port support for various protocols that support
 * 16-bit port ids coupled with a 32-bit IP-like address.
 * <p>
 * This is the internal representation of most "AddressSpec"'s.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class AddressSpecPortAddr
	implements AddressSpec
{
	PACKAGE static final InetAddress WILDCARD_ADDRESS = null;
	PACKAGE static final int         WILDCARD_PORT    = -1;

	/**
	 * Source internet address for this AddressSpec
	 */
	private InetAddress saddr = WILDCARD_ADDRESS;
	
	/**
	 * Source port this AddressSpec represents.
	 */
	private int sport = WILDCARD_PORT;

	/**
	 * Destination internet address for this AddressSpec
	 */
	private InetAddress daddr = WILDCARD_ADDRESS;
	
	/**
	 * Destination port this AddressSpec represents.
	 */
	private int dport = WILDCARD_PORT;

	PACKAGE AddressSpecPortAddr(String addr)
		throws PacketAddressFormatException
	{
		this.parseAddress(addr);
	}

	/**
	 * Parses UDP address strings.  The format is:
	 *  "<src address>:<dest port>:<dest address>:<src port>
	 * <p>
	 * XXX should make this private, require all users to come through
	 * the constructor.
	 */
	public void parseAddress(String addr)
		throws PacketAddressFormatException
	{
		StringTokenizer st = new StringTokenizer(addr, ":");
		
		try
		{
			String frag;
			frag = st.nextToken(); // srcAddr
			if (!frag.equals("*"))
			{
				// XXX Move this check to somewhere where it belongs.  (not here).
				throw new PacketAddressFormatException
					("Janos cannot filter on UDP source address (" +frag+ ").");
			}
			this.saddr = WILDCARD_ADDRESS;
			
			assert(st.hasMoreTokens()); // XXX throw legit error
			frag = st.nextToken();
			if (frag.equals("*"))
				this.dport = WILDCARD_PORT;
			else
				this.dport = Integer.parseInt(frag);
			
			assert(st.hasMoreTokens()); // XXX throw legit error
			frag = st.nextToken();
			if (frag.equals("*"))
				this.daddr = WILDCARD_ADDRESS;
			else
				this.daddr = InetAddress.getByName(frag);
			
			if (st.hasMoreTokens())
			{
				// XXX Move this check to somewhere where it belongs.  (not here).
				throw new PacketAddressFormatException
					("Janos cannot filter on UDP source port (" +st.nextToken()+ ").");
			}
			this.sport = WILDCARD_PORT;
		}
		catch(UnknownHostException e)
		{
			throw new PacketAddressFormatException
				("Cannot parse address " +addr+": " +e);
		}
		catch(NoSuchElementException e)
		{
			throw new PacketAddressFormatException
				("Address " +addr+ " formatted incorrectly.");
		}
	}

	PACKAGE InetAddress destAddr()
	{
		return daddr;
	}
	
	PACKAGE int destPort()
	{
		return dport;
	}

	PACKAGE InetAddress sourceAddr()
	{
		return saddr;
	}
	
	PACKAGE int sourcePort()
	{
		return sport;
	}

	public int hashCode()
	{
		return this.saddr.hashCode() ^ this.daddr.hashCode() ^
			(this.sport << 16) ^ this.dport; 
	}

	/**
	 * Two AddressSpecPortAddr objects are equal if the 
	 * source and destination ports and addresses are equal.
	 */
	public boolean equals(Object o)
	{
		boolean retval = false;

		if ( o instanceof AddressSpecPortAddr )
		{
			AddressSpecPortAddr other = (AddressSpecPortAddr)o;

			retval =
				(other.sport == this.sport)
				&& (other.dport == this.dport)
				&& (other.saddr == this.saddr)
				&& (other.daddr == this.daddr);
		}
		return( retval );
	}
	
	public String toString()
	{
		return "AddressSpecPortAddr[sa="
			+ this.saddr
			+ "; sp="
			+ this.sport
			+ "; da="
			+ this.daddr
			+ "; dp="
			+ this.dport
			+ "]";
	}
}
