/*
 * DemultiplexKey.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

/**
 * A key for matching incoming packets.
 *
 * @see InChannel
 * @see CutThroughChannel
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public class DemultiplexKey
	extends NATIVE_DEMULTIPLEXKEY
{
	CHECK_NATIVE_NODEOS

	/**
	 * Create a new demultiplex key.  By default, the key
	 * is empty (matches anything, or everything, I don't remember.)
	 */
	public DemultiplexKey(Flow flow)
	{
		super(flow.getNativeFlow());
		require(Flow.currentFlow() == flow);
	}

	/**
	 * Add a segment that matches the given sequence of bytes at the given
	 * offset in the packet.  The mask says which <emph>bytes</emph> of
	 * sequence are relevant.  If mask is null, all bytes are relevant.
	 * <p>
	 * XXX currently mask must be null.
	 * <p>
	 * XXX should throw a real exception for bogon parameters (vs. Error).
	 *
	 * @param offset The offset from the previous segment to start matching
	 * @param length The length of the bytes to match
	 * @param sequence The sequence of bytes to match
	 * @param mask If non-null, its a bit mask that indicates which bytes
	 * in `sequence' should be matched.
	 */
	public void match(int offset, int length,
			  byte sequence[], boolean mask[])
	{
		if ((offset < 0)
		    || (length <= 0)
		    || (sequence == null))
			throw new Error("bad match parameters"); // XXX need real exception
		
		// XXX other sanity checks?

		super.match(offset, length, sequence, mask);
	}

	/**
	 * Add a segment to match an IPv4 header.
	 * Offset is the expected offset of the IPv4 header in the packet.
	 * <p>
	 * Use this instead of the vanilla match() to get a "payloadOffset"
	 * in the deliver func that will skip IP options.  (A vanilla
	 * match() call will leave payloadOffset in the first byte after
	 * the "protocol" byte, which is not necessairly the first byte
	 * of the payload in an IP packet (e.g., one with options).
	 * <p>
	 * XXX need well-defined wildcards for addresses and protocols.
	 *
	 * @param offset the offset of the IPv4 header in the packet.
	 * @param source the 32-bit IPv4 source address to match in the header.
	 * @param dest the 32-bit IPv4 destination address to match in the header.
	 * @param protocol the 16-bit IPv4 protocol identifier.
	 */
	public void matchIPv4(int offset, int source, int dest, int protocol)
	{
		if (offset < 0)
			throw new Error("bad match parameters"); // XXX need real exception
		
		// XXX other sanity checks?

		super.matchIPv4(offset, source, dest, protocol);
	}

	/**
	 * Add a segment to match a UDP header, needs to be preceded by a
	 * call to <code>matchIPv4</code>.
	 * <p>
	 * XXX the exact semantics of 'offset' are not worked out yet.
	 */
	public void matchUDP(int offset, int srcPort, int destPort)
	{
		if ((offset < 0)
		    || (srcPort < -1)
		    || (srcPort > (64 * 1024))
		    || (destPort < -1)
		    || (destPort > (64 * 1024)))
			throw new Error("bad matchUDP() parameters"); // XXX need real exception
		
		// XXX other sanity checks?
		
		super.matchUDP(offset, srcPort, destPort);
	}

	/**
	 * Add a segment to match a TCP header, needs to be preceded by a
	 * call to <code>matchIPv4</code>
	 * <p>
	 * XXX the exact semantics of 'offset' are not worked out yet.
	 */
	public void matchTCP(int offset, int srcPort, int destPort)
	{
		if ((offset < 0)
		    || (srcPort < -1)
		    || (srcPort > (64 * 1024))
		    || (destPort < -1)
		    || (destPort > (64 * 1024)))
			throw new Error("bad matchUDP() parameters"); // XXX need real exception
		
		// XXX other sanity checks?

		super.matchTCP(offset, srcPort, destPort);
	}
	
	/**
	 * Add a segment that matches an ANEP header
	 *
	 * @param offset The offset from the previous segment
	 * @param version The ANEP version number to match
	 * @param allVersions Match any version of ANEP
	 * @param protocolID The protocolID to match
	 */
	public void matchANEP(int offset, byte version,
			      boolean allVersions, short protocolID)
	{
		if ((offset < 0)
		    || (version < 0))
			throw new Error("bad matchUDP() parameters"); // XXX need real exception
		
		// XXX other sanity checks?
		super.matchANEP(offset, version, allVersions, protocolID);
	}

	/**
	 * Test a packet to see if this key matches.
	 * <p>
	 * XXX payloadLength has odd semantics.  Either explain why, or fix
	 * the callers of this method.
	 *
	 * @param payload The packet data
	 * @param payloadStart The place to start matching in `payload'
	 * @param payloadLength The length of data in payload, including bytes
	 * up to payloadStart.
	 *
	 * @return 0 if there was an error otherwise the index of the first byte after the
	 * end of the key.
	 */
	public final int matches(byte payload[],
				 int payloadStart,
				 int payloadLength)
	{
		if ((payloadStart < 0)
		    || (payloadStart > payload.length)
		    || (payloadLength > payload.length))
			throw new Error("Bad payload for matches()"); // XXX need real exception
		
		return super.matches(payload, payloadStart, payloadLength);
	}

	PACKAGE NATIVE_DEMULTIPLEXKEY getNativeKey()
	{
		return this;
	}

	public String toString()
	{
		return super.toString();
	}
}
