/*
 * Flow.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector; // XXX

/**
 * Java wrapper for a flow.
 * <p>
 * If you know BeOS or V, this is like a "Team".  If you prefer Mach
 * (or know what Fluke is), this is similar to a "Task".  If all you
 * know is Unix, then think of this as a "process".  You might also
 * call it a "resource container" or possibly a "protection domain."
 * 
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public final class Flow
	extends NATIVE_FLOW
{
	CHECK_NATIVE_NODEOS

	/**
	 * The Flow ClassLoader for this flow.
	 * <p>
	 * This object is allocated and initialized by the flow's 
	 * FlowInitializer.  That is the only reason it isn't final.
	 * <p>
	 * The loader also acts as a flag for a destroyed flow.  If
	 * loader is null, the flow is destroyed.  Note that loader is
	 * null until the first thread in the new flow is started, so
	 * don't rely on destroyed() until the flow actually running.
	 *
	 * @see FlowInitializer#run()
	 */
	PACKAGE FlowClassLoader loader;

	/**
	 * Create a new Flow, which is a child of the current Flow.
	 * Starts a new thread executing within the new flow.
	 * <p>
	 * This is the only interface for creating a Flow.
	 * <p>
	 * The termHandler will be invoked by a thread in the parent
	 * Flow's if the child Flow is terminated.  This handler has a
	 * short limit for execution.
	 * <p>
	 * If termHandler is null, then no termination handler is called.
	 * <p>
	 * XXX Should be called "create()" to sync back with NodeOS spec.
	 *
	 * @param cred The principal associated with this Flow
	 * @param tPoolSpec The specification for the new Flow's ThreadPool
	 * @param memSpec The memory limit for this flow
 	 * @param flowStartSpec Class loading and startup information
	 * @param termHandler Termination handler invoked by the NodeOS when an immediate child
	 * flow is being terminated.
	 * @return a FlowHandle onto the new child flow.
	 */
	public FlowHandle newFlow(final Credentials cred,
				  final ThreadPoolSpec tPoolSpec,
				  final MemSpec memSpec,
				  final FlowStartSpec flowStartSpec,
				  final Runnable termHandler)
	{
		require(cred);
		require(tPoolSpec);
		require(memSpec);
		require(flowStartSpec);
		// XXX require(termHandler);
		
		TRACE(Trace.NODEOS_FLOWSTART, "Flow.newFlow()");

		Flow flow = new Flow(this,
				     cred,
				     tPoolSpec,
				     memSpec,
				     termHandler);

		flow.start(flowStartSpec);

		// XXX flow has no pointer to its handle

		return new FlowHandle(flow, termHandler);
	}

	/**
	 * Initalize a new Flow.  See newFlow().  Does not "start"
	 * the flow.
	 */
	PACKAGE Flow(Flow parent,
		     Credentials cred,
		     ThreadPoolSpec tPoolSpec,
		     MemSpec memSpec,
		     Runnable termHandler)
	{
		super((parent == null) ? null : parent.getNativeFlow(),
		      cred.getNativeCreds(),
		      tPoolSpec.getNativeSpec(),
		      memSpec.getNativeSpec(),
		      termHandler);

		require((parent == null) // null for root flow only
			|| (parent == Flow.currentFlow()));

		TRACE(Trace.NODEOS_FLOWSTART, "Flow.<init>(parent=" +parent+ ")");

		FlowRegister.register(this); // XXX force a name?
	}

	/**
	 * Start a new flow going.  This must only be invoked once per flow.
	 * It is automatically invoked by the newFlow() method.
	 */
	PACKAGE void start(FlowStartSpec flowStartSpec)
	{
		/*
		 * Create a FlowInitializer to finish the Flow
		 * initialization in the context of a thread in the
		 * new Flow's ThreadPool.
		 */
		FlowInitializer fi;

		TRACE(Trace.NODEOS_FLOWSTART, "Flow.start(fss=" +flowStartSpec+ ")");

		fi = FlowInitializer.createInFlow(this,
						  flowStartSpec);

		// fi is a pointer into the new flow's heap.

		TRACE(Trace.NODEOS_FLOWSTART, "Flow.start() creating/kicking thread.");
		this.threadStart(fi,
				 "Flow init thread "
				 + flowStartSpec.startClass);
	}

	public synchronized void destroy()
	{
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);

		if (loader != null)
		{
			loader.destroy();
			loader = null;
			FlowRegister.unregister(this);
			super.destroy();
		}
	}

	public synchronized boolean destroyed()
	{
		// destroyed() is only reliable for threads inside the flow.
		assert(this == Flow.currentFlow());
		
		// XXX ???
		return (loader == null);
	}
	
	/**
	 * Create an explicit thread.  This is a new thread which is
	 * not subject to whatever concurrency limit this Flow's
	 * ThreadPool has.  But, it will be scheduled by the
	 * ThreadPool's scheduler.
	 * 
	 * @allocations: This function will allocate a new Thread object.
	 * @param r the Runnable to start the target thread on.
	 * @param name the String name of the new thread.
	 */
	public void threadStart(java.lang.Runnable r, String name)
	{
		super.threadStart(r, name);
	}

	public Object getObject()
	{
		return super.getObject();
	}

	public void setObject(Object obj)
	{
		super.setObject(obj);
	}

	public String toString()
	{
		return super.toString();
	}

	/**
	 * Add a new class to an existing flow.
	 */
	public void addClass(String name, byte[] bytecode)
	{
		assert(this == Flow.currentFlow());
		assert(name != null);
		assert(bytecode != null);
		assert(loader);
		loader.addClass(name, bytecode);
	}
	
	/**
	 * Add a package to the list of "system" packages.  System
	 * packages will be loaded by the system class loader.
	 * All other classes will be loaded by the per-flow
	 * classloader.
	 * <p>
	 * All sub-flows *share* the same definition of "system"
	 * packages for now.  This is because only one EE System
	 * is loaded on the NodeOS.  This will have to change.
	 * <p>
	 * Implicitly allows the package.
	 */
	public void addSystemPackage(String name)
	{
		assert(this == Flow.currentFlow());
		assert(name != null);
		assert(loader != null);

		loader.addSystemPackage(name);
	}
	 
	/**
	 * Allow access to a Java package.  Must be invoked in the
	 * context of this flow.
	 * <p>
	 * addLocalPackage implicitly allows those packages, so
	 * addPackage need not be called on those packages.
	 * <p>
	 * Allowed packages that are not local are either dynamically
	 * loaded over the net, or are found locally.
	 */
	public void allowPackage(String pkgPrefix)
	{
		assert(this == Flow.currentFlow());
		assert(pkgPrefix != null);
		assert(loader != null);

		loader.allowPackage(pkgPrefix);
	}
	
	public void enablePackageControls(boolean doit)
	{
		assert(this == Flow.currentFlow());
		assert(loader != null);

		loader.enablePackageControls(doit);
	}

	public Class loadClass(String name)
		throws java.lang.ClassNotFoundException
	{
		assert(name != null);
		assert(loader);
		return loader.loadClass(name);
	}

	/**
	 * Get the current flow.  The current Flow is the
	 * flow associated with the current Thread.
	 *
	 * @see Thread#getFlow()
	 */
	public final static Flow currentFlow()
	{
		NATIVE_FLOW nflow;

		nflow = NATIVE_FLOW.native_currentFlow();
		assert((nflow == null) || (nflow instanceof Flow));
		return (Flow)nflow;
	}

	PACKAGE NATIVE_FLOW getNativeFlow()
	{
		return this;
	}

#ifdef DEBUG
	// always non-null.
	private String name = "John Doe Flow";
	
	public void setName(String name)
	{
		if (name != null)
			this.name = name;
	}

	public String getName()
	{
		return name;
	}
#endif /*DEBUG*/	
}

