/*
 * FlowInitializer.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.lang.reflect.Constructor;

/**
 * Callback for initializing a new Flow (in the context of one of its
 * own threads).
 * <p>
 * This class's <code>run()</code> should be a method on a Flow
 * object, but starting a thread requires a <emph>public</emph>
 * <code>run()</code> method, which would be visible to the world on
 * Flow.
 * <p>
 * XXX overlaps with the public FlowStartSpec?
 * 
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class FlowInitializer
	implements Runnable
{
	private CommSpaceHandle commSpaceHandle;
	private String startClass;
	private String[] args;
	
	/**
	 * Create the a FlowInitializer in the heap of the given flow
	 * (i.e., not the current flow).
	 *
	 * @return a cross-heap pointer to a FlowInitializer
	 */
	PACKAGE static FlowInitializer createInFlow(Flow flow,
						    FlowStartSpec fss)
	{
		FlowInitializer fi;
		CommSpaceHandle commSpaceHandle;
		String startClass;
		String[] args;

		require(flow);
		require(fss);
		require(fss.args);
		require(fss.startClass);

		if (fss.commSpaceHandle == null) // XXX for root flow only
			commSpaceHandle = new CommSpaceHandle();
		else
			commSpaceHandle = new CommSpaceHandle(fss.commSpaceHandle);

		startClass = new String(fss.startClass);
		args = new String[fss.args.length];
		for (int i = 0; i < args.length; i++)
			args[i] = new String(fss.args[i]);

		fi = new FlowInitializer(startClass, args, commSpaceHandle);

		return fi; // fi is a cross-heap pointer!
	}

	private FlowInitializer(String startClass,
				String[] args,
				CommSpaceHandle commSpaceHandle)
	{
		require(startClass);
		require(startClass.length() > 0);
		require(args);

		this.startClass = startClass;
		this.args = args;
		this.commSpaceHandle = commSpaceHandle;
	}

	public void run()
	{
		Flow flow;

		TRACE(Trace.NODEOS_FLOWSTART, "FlowInitializer.run()");

		flow = Flow.currentFlow();

		try
		{
			/* Create the FlowClass loader for the flow. */
			initializeFlow(flow);
			
			/* Start the startclass in the flow. */
			startInitialClass(flow,
					  this.startClass,
					  this.args,
					  this.commSpaceHandle);
		}
		catch (Error e) // XXX
		{
			Node.println("Exception in FlowInitializer: " +e);
			Node.printStackTrace(e);

			flow.destroy();
		}
	}
	
	/**
	 * Initialize the flow object.
	 * Just creates the flow's class loader.
	 */
	private static void initializeFlow(Flow flow)
	{
		FlowClassLoader loader;

		require(flow);
		require(flow.loader == null);
		
		// Initialize the FlowClassLoader for this Flow.
		loader = new FlowClassLoader(flow);
		flow.loader = loader;
	}

	/**
	 * Start the flow on its way.
	 */
	private static void startInitialClass(Flow flow,
					      String startClass,
					      String[] args,
					      CommSpaceHandle csh)
		throws ExceptionInInitializerError
	{
		FlowRun flowRun;
		FlowClassLoader loader;

		require(flow);
		require(startClass);
		require(args);
		
		TRACE(Trace.NODEOS_FLOWSTART, "FlowInitializer.startInitialClass()");

		try
		{
			Class c;

			c = flow.loader.loadClass(startClass, true);
			flowRun = createStartClassInstance(c);
		}
		catch (Exception e)
		{
			throw new ExceptionInInitializerError(e);
		}
		
		TRACE(Trace.NODEOS_FLOWSTART, "FlowInitializer.startInitialClass(): running startup class.");
		flowRun.run(flow, args, csh); 
	}
	
	private static FlowRun createStartClassInstance(Class c)
		throws Error,  // XXX Error should be a Janos-specific Error
		NoSuchMethodException,
		IllegalAccessException,
		InstantiationException,
		java.lang.reflect.InvocationTargetException // XXX XXX XXX
	{
		Object instance;

		require(c);

		/*
		 * This seemed so easy:
		 *    flow.flowRun = (FlowRun) c.newInstance();
		 *
		 * But no, it's not, because we don't want to require
		 * c to have a public constructor.  See
		 * ClassInstantiationKey.
		 *
		 * Or, we need to have a VM method that will execute a
		 * private/package constructor on the class.  We trust
		 * the EE runtime to only pass safe classes.
		 */
		try
		{
			Class[] paramTypes;
			Object[] params;
			Constructor con;

			paramTypes = new Class[]{ ClassInstantiationKey.class };
			con = c.getConstructor(paramTypes);
							   
			params = new Object[] { new ClassInstantiationKey(c) };

			instance = con.newInstance(params);
		}
		catch(NoSuchMethodException e)
		{
			// Try the no-arg public constructor version(?)
			instance = c.newInstance();
		}

		if (! (instance instanceof FlowRun))
			throw new Error(c.getName() + " is not an instanceof FlowRun.");

		return (FlowRun)instance;
	}
}
