/*
 * FlowRegister.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector; // XXX

/**
 * Maintain a map of all the Flows in the system.  Currently
 * only used by the FlowInfo InfoHook.
 * <p>
 * <b>NOTE:</b> We don't use a hashtable to map flows
 * from name to flow object, becuase names are mutable and
 * are set after the flow is registered.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class FlowRegister
{
	private static final Vector flows = new Vector(30);

	/**
	 * InfoHook class that reports the current set of flows in the
	 * system via the NodeOS Status interface.
	 */
	private static class FlowInfo
		extends InfoHook
	{
		private FlowInfo()
		{
			super("Flow");
		}

		public String[] getInfo()
		{
			String retval[];

			synchronized(flows)
			{
				retval = new String[flows.size()];
				
				for (int i = 0; i < flows.size(); i++)
				{
					Flow elem = (Flow)flows.elementAt(i);
					retval[i] = elem.toString();
				}
			}
			return retval;
		}
	}

	static {
		Status.addInfoHook(new FlowInfo());
	}

	/**
	 * Register a flow.
	 */
	PACKAGE static void register(Flow flow)
	{
		synchronized(flows)
		{
			flows.addElement(flow);
		}
	}

	/**
	 * Unregister a flow.
	 * <p>
	 * NOTE: this method is invoked in a context where most of the
	 * Flow has been destroyed.  Beware.
	 */
	PACKAGE static void unregister(Flow flow)
	{
		boolean foundit = false;

		TRACE(Trace.NODEOS_DESTROY, "Unregistering flow " +flow);

		synchronized(flows)
		{
			// Check all elements to make sure it wasn't duplicated
			for (int i = 0; i < flows.size(); i++)
			{
				if (flow == flows.elementAt(i))
				{
					flows.removeElementAt(i);
					assert(foundit == false);
					foundit = true;
				}
			}
		}

		if (!foundit)
			panic("Flow " +flow+ " could not be unregistered!");
	}

#ifdef DEBUG
	/**
	 * Return the flow object corresponding to the given name
	 */
	public static Flow findFlow(String name)
	{
		Flow retval = null;
		int i;

		for( i = 0; (i < flows.size()) && (retval == null); i++ )
		{
			Flow elem = (Flow)flows.elementAt(i);
			if( elem.getName().equals(name) )
				retval = elem;
		}
		return retval;
	}
#endif
}
