/*
 * InChannel.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.io.IOException;

/**
 * Java wrapper for a NodeOS InChannel.
 *
 * @see CutThroughChannel
 * @see OutChannel
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public final class InChannel
	extends NATIVE_INCHANNEL
{
	CHECK_NATIVE_NODEOS

	/**
	 * Create a new InChannel.
	 */
	public static InChannel create(Flow flow,
				       DemultiplexKey classifier,
				       String protocolSpec,
				       String addrSpec,
				       DeliverFunc deliverFunc)
		throws IOException,
		       PacketAddressFormatException,
		       ProtocolSpecFormatException
	{
		// XXX doesn't sufficiently protect the kernel...
		if((flow == null)
		   || (classifier == null)
		   || (protocolSpec == null)
		   || (deliverFunc == null))
			throw new NullPointerException();

		return new InChannel(flow, classifier,
				     protocolSpec,
				     addrSpec,
				     deliverFunc);
	}
	

	private InChannel(Flow flow,
			  DemultiplexKey classifier,
			  String protocolSpec,
			  String addrSpec,
			  DeliverFunc deliverFunc)
		throws IOException,
		       PacketAddressFormatException,
		       ProtocolSpecFormatException
	{
		super(flow.getNativeFlow(),
		      classifier.getNativeKey(),
		      protocolSpec,
		      addrSpec,
		      deliverFunc);
	}
	
	/**
	 * Add a buffer to the InChannel
	 */
	public void addBuffer(BufferHandle bufh)
	{
		NATIVE_BUFFERHANDLE nb;

		if (bufh == null)
			throw new NullPointerException();

		bufh.systemOwns();
		
		nb = bufh.getNativeBufH();
		super.native_addBuffer(nb);
	}

	/**
	 * Remove a buffer from the queue of free buffers
	 * on this InChannel.  Returns null if there are
	 * no free buffers on this InChannel.
	 */
	public BufferHandle remBuffer()
	{
		return (BufferHandle)super.native_remBuffer();
	}
	
	/**
	 * Add `count' buffers to the in channel.
	 *
	 * @allocation Creates 'count' new BufferHandles (and buffers).
	 */
	public void addBuffers(int count)
	{
		int lpc;
		
		require(count > 0);

		for( lpc = 0; lpc < count; lpc++ )
		{
			BufferHandle bh;
			bh = BufferHandle.create();
			bh.systemOwns();
			super.native_addBuffer(bh.getNativeBufH());
		}
	}

	/**
	 * Remove `count' buffers from the channel.  Or all the buffers
	 * if there are less than `count'.
	 */
	public void remBuffers(int count)
	{
		int lpc;
		
		require(count > 0);

		for( lpc = 0; lpc < count; lpc++ )
		{
			Object o;
			o = super.native_remBuffer();
			if (o == null)
				return;
			require(o instanceof BufferHandle);
		}
	}
	
	public String toString()
	{
		return super.toString();
	}

	PACKAGE void destroy()
	{
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);
		super.destroy();
	}
	
	PACKAGE NATIVE_INCHANNEL getNativeInChan()
	{
		return this;
	}

	protected void finalize()
	{
		destroy(); // XXX ???
	}
}
