/*
 * Main.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

IMPORT_PROFILER

/**
 * Boot the Janos Java NodeOS.  Initializes the underlying
 * native NodeOS, sets up the basic system packages in the
 * ClassLoader, and creates the root flow.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public class Main
{
	/**
	 * The default termination handler for the root flow.
	 * Just shuts down the entire VM (via Node.shutdown()).
	 */
	private static class RootFlowTermHandler 
		implements java.lang.Runnable
	{
		public void run()
		{
			Node.println("Root flow term handler.  Shutting down.");
			Node.shutdown();
		}
	}

	/**
	 * All Janos active routers start with this function. 
	 * We load up native libraries, define the set of
	 * "system" packages, and fire up the root flow.
	 * The arguments to main() should be a class to instantiate
	 * in the root flow and whatever arguments it wants.
	 *
	 * @see FlowClassLoader for info system packages.
	 */
	public static void main(String[] args)
	{
		NATIVE_SETUP.initialize();

		PROFILESTAGE("SysInit");

		Node.println("Janos Java NodeOS v1.1.0");
		
		require(args);
		if (args.length == 0)
		{
			Node.println("ERROR: Must supply a class to load into root flow.");
			Node.shutdown();
		}

		String initClass;
		String[] initArgs;
		String initPackage;
		int pkgend;
		
		initClass = args[0];
		
		// Find the package the initial class is in.
		pkgend = initClass.lastIndexOf('.');
		initPackage = null;
		if (pkgend != -1)
			initPackage = initClass.substring(0, pkgend);
		defineSystemPackages(initPackage);

		initArgs = new String[args.length - 1];
		if (args.length > 1)
			System.arraycopy(args, 1, initArgs, 0, initArgs.length);
		
		PROFILESTAGE("NodeOSInit");

		startRootFlow(initClass, initArgs);

		/* XXX don't do anything if startRootFlow returns... */
	}

	/**
	 * Start the native root flow.  Not really the "root flow" in
	 * Moab case, as we start the root flow in md.c before Kaffe's
	 * main even gets going.
	 */
	private static void startRootFlow(String initClass,
					  String[] initArgs)
	{
		require(initClass);
		require(initArgs);

		Flow rootFlow = new Flow(
			null,
			new Credentials("root cred"),
			new ThreadPoolSpec(3, 32*1024,
				ThreadScheduler.basicScheduler()),
			new MemSpec(),
			new RootFlowTermHandler());
		DEBUGONLY(rootFlow.setName("Root Flow"));

		PROFILESTAGE("ANTSRInit");		

		rootFlow.start(
			new FlowStartSpec(
				initClass,
				initArgs,
				/*CommSpaceHandle*/ null));
	}

	/**
	 * Define the set of "system" packages that are loaded into
	 * the system flow.  This set is basically fixed, except for
	 * the implicit package passed on the command line.
	 */
	private static void defineSystemPackages(String initPackage)
	{
		FlowClassLoader.addSystemPackage("java.lang");
		FlowClassLoader.addSystemPackage("java.io");
		FlowClassLoader.addSystemPackage("java.util");
		FlowClassLoader.addSystemPackage(JANOS_NODEOS_PKG_S);
		FlowClassLoader.addSystemPackage(JANOS_UTIL_PKG_S);
		FlowClassLoader.addSystemPackage(JANOS_IO_PKG_S);
		FlowClassLoader.addSystemPackage(JANOS_ROUTE_PKG_S);
		if (initPackage != null)
			FlowClassLoader.addSystemPackage(initPackage);
	}
}

