/*
 * moab_Debug.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

/**
 * see pj_Debug for doc.
 * 
 * @see Debug
 * @see pj_Debug
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public class moab_Debug
{
	CHECK_NATIVE_NODEOS

#ifdef DEBUG
	/**
	 * Print a file:line location.  Used for asserts and trace
	 * macros.
	 * <p>
	 * Trims the filename down to just the base file name...
	 *
	 * @param file a file name
	 * @param line a line number
	 * @allocation String.substring() may allocate
	 */
	private static void printLocation(String file, int line)
	{
		int off = file.lastIndexOf('/');
		if (off > 0)
			file = file.substring(off+1);
		print(file);
		print(":");
		printInt(line);
	}

	/**
	 * If an assert fails, this routine is called to print the
	 * file and line number information, along with the string
	 * representation of the failed expression, and a backtrace
	 * to the failure.
	 *
	 * @param str the string representation of the expression
	 * @param file the name of the file in which the assert failed
	 * @param line the line number on which the assert failed in <code>file</code>
	 */
	private static void assertFailure(String str, String file, int line)
	{
		java.lang.Thread.currentThread().setPriority(10);
		printLocation(file, line);
		print(": Assertion failed: ");
		println(str);
		
		printStackTrace();
		
		// println(" ---- STATS: ----");
		// // dumpStats();
		// println(" ---- :STATS ----");
		
		exit();
	}
	
	/**
	 * Check that the boolean expression is true.  Calls assertFailure()
	 * if not.
	 *
	 * @param expr the expression to check
	 * @param exprStr the string representation of the expression
	 * @param file the name of the file the assert is in
	 * @param line the line number of the file the assert is in
	 * @see #assertFailure
	 */
	public static void assertCheck(boolean expr, String exprStr, String file, int line) 
	{
		if (!expr) 
		{
			assertFailure(exprStr, file, line);
		}
	}

	/**
	 * Check that the given object reference is non-null. Calls assertFailure() if the reference
	 * is null.
	 *
	 * @param expr the expression to check
	 * @param exprStr the string representation of the expression
	 * @param file the name of the file the assert is in
	 * @param line the line number of the file the assert is in
	 * @see #assertFailure
	 */
	public static void assertCheck(Object ref, String exprStr, String file, int line) 
	{
		if (ref == null) 
		{
			assertFailure(exprStr, file, line);
		}
	}

	/**
	 * Print a trace message.
	 *
	 * @param file the name of the file the TRACE is in
	 * @param line the line number of the file the TRACE is in
	 * @param msg the trace message to print
	 */
	public static void pTrace(String file, int line, String msg) 
	{
		java.lang.Thread.currentThread().setPriority(10); // XXX
		printLocation(file, line);
		print(": ");
		println(msg);
		java.lang.Thread.currentThread().setPriority(5); // XXX
	}

	/**
	 * Assert that the given object is locked (synchronized).
	 *
	 * @param obj the object to check for lock status.
	 */
	public static void checkLocked(Object obj, String exprStr, String file, int line)
	{
		// Cannot check this in Pure Java
		return;
	}
	
	/**
	 * A native print routine that doesn't require the basic Java classes.
	 * Prints a trailing newline.
	 *
	 * @param str The string to print
	 */
	public static void println(String str)
	{
		java.lang.System.err.println(str);
	}
	

	/**
	 * A native print routine that doesn't require the basic Java classes.
	 * Does not print a trailing newline.
	 *
	 * @param str The string to print
	 */
	public static void print(String str)
	{
		java.lang.System.err.print(str);
	}

	/**
	 * A native print routine that doesn't require the basic Java classes.
	 * Prints the given integer as a string (in base 10).
	 *
	 * @param i the integer to print
	 */
	public static void printInt(int i)
	{
		java.lang.System.err.print(i);
	}

	/**
	 * A native print routine that doesn't require the basic Java classes.
	 * Prints the given integer as a string (in hex).
	 *
	 * @param i the integer to print
	 */
	public static void printIntHex(int i)
	{
		// XXX not implemented
		java.lang.System.err.print("0xXX");
	}

	/**
	 * A native print routine that doesn't require the basic Java classes.
	 * Prints the given char array as a sequence of characters.
	 *
	 * @param array the char array to print
	 * @param start the first element of the char array to print
	 * @param end the last element of the char array to print
	 */
	public static void printCharArray(char[] array, int start, int end)
	{
		// XXX not implemented
		print("char[] XX");
	}

	/**
	 * A direct exit.  Do not pass Go; do not collect two hundred dollars.
	 *
	 * TODO: Rather dangerous to leave around.
	 */ 
	public static void exit()
	{
		try
		{
			java.lang.System.err.println("Debug.exit() invoked.");
			// XXX by " + Flow.currentFlow().getName());
			java.lang.System.exit(11);
		}
		catch (Throwable th)
		{
			java.lang.System.err.println("Error thrown in exit(): " +th.getMessage());
		}
		java.lang.System.exit(11);
	}
	
	/**
	 * Get a string of the given Buffer.
	 */
	public static String dbgPacketSummary(BufferHandle bufh)
	{
		return dbgPacketSummary(bufh, 0, bufh.getLength());
	}

	/**
	 * Get a string of maxDump bytes from the given Buffer.
	 */
	public static String dbgPacketSummary(BufferHandle bufh, int len)
	{
		return dbgPacketSummary(bufh, 0, len);
	}

	/**
	 * Get a string of a subsequence of the bytes in the given packet.
	 */
	public static String dbgPacketSummary(BufferHandle bufh, int offset, int length)
	{
		NATIVE_BUFFERHANDLE b;
		String r;
		
		b = bufh.getNativeBufH();
		r = dbgPacketSummary(b, offset, length);
		return r;
	}

	public static String dbgPacketSummary(NATIVE_BUFFERHANDLE b)
	{
		return dbgPacketSummary(b, 0, b.getLength());
	}

	public static String dbgPacketSummary(NATIVE_BUFFERHANDLE b, int len)
	{
		return dbgPacketSummary(b, 0, len);
	}

	public static String dbgPacketSummary(NATIVE_BUFFERHANDLE b, int offset, int maxlen)
	{
		char[] sbuf;
		int usedstrlen;

		sbuf = new char[50 + (maxlen*4)];
		usedstrlen = b.dbgPacketSummary(sbuf, offset, maxlen);
		return new String(sbuf, 0, usedstrlen);
	}

	/**
	 * Print the stack trace to the current point of execution in the current
	 * thread.
	 */
	public static void printStackTrace()
	{
		Throwable t = new Throwable();
		java.lang.Thread.currentThread().setPriority(10); // XXX
		t.printStackTrace(java.lang.System.err);
		java.lang.Thread.currentThread().setPriority(5); // XXX
	}
	
	public static InternalError fatalException(Throwable t, String msg)
	{
		java.lang.Thread.currentThread().setPriority(10); // XXX
		t.printStackTrace(java.lang.System.err);
		return dPanic(msg);
	}
	
	/**
	 * A less generic not-implemented print-and-die fucntion.
	 * Prints the provided string and a backtrace before killing the VM.
	 *
	 * @param expl the explaination for this rude behavior.
	 */
	public static InternalError notImplemented(String expl)
	{
		print("NOT IMPLEMENTED: ");
		println(expl);
		printStackTrace();
		exit();
		return new InternalError("exit() returned in Debug.notImplemented()");
	}

	/**
	 * A generic panic() function.  Prints the provided string,
	 * a backtrace, and then kills the VM.
	 * <p>
	 * Whee!  This returns an InternalError so that the caller
	 * might throw the result giving the compiler the impression
	 * that dPanic() will never return.  Hehehe.
	 *
	 * @param str The panic message to use as an excuse for killing the machine.
	 */
	public static InternalError dPanic(String str)
	{
		print("PANIC: ");
		println(str);
		printStackTrace();
		exit();
		return new InternalError("exit() returned in Debug.dPanic");
	}
#endif /*DEBUG*/
}
