/*
 * pj_ANEPLayer.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

/**
 * pj_ANEPStreamLayer is a protocol layer implementation that processes packets
 * that are sent over a stream.  Its really meant to be used as a base class
 * which is extended for a specific communication type (TCP/StdIO/etc...).
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE abstract class pj_ANEPStreamLayer
	extends pj_ProtocolLayer
{
	PACKAGE pj_ANEPStreamLayer()
	{
	}

	public String toString()
	{
		return "pj_ANEPStreamLayer[]";
	}

	/**
	 * StreamReader is an inner class used to read from an InputStream and
	 * produce PacketBuffer's.  The stream is expected to contain only
	 * ANEP packets since the length of the packet encoded in the ANEP
	 * header is used to delimit the beginning and end of packets.
	 */
	protected static abstract class StreamReader
		implements Runnable
	{
		private final InputStream is;
		private int users = 1;
		private final pj_BufferHandle pb;

		protected StreamReader(InputStream is, pj_BufferHandle pb)
		{
			this.is = is;
			this.pb = pb;
		}

		protected void acquire()
		{
			this.users++;
		}

		protected boolean release()
		{
			this.users--;
			return this.users == 0;
		}

		/**
		 * Handle a packet parsed out of the stream
		 */
		protected abstract boolean handlePacket(pj_BufferHandle pb);

		/**
		 * Helper function for always reading the specified amount
		 * of data.
		 */
		private void read(byte data[], int offset, int len)
			throws IOException
		{
			int remaining = len, bytesRead;

			while ( remaining != 0 )
			{
				bytesRead = this.is.read(data,
							 offset,
							 remaining);
				switch ( bytesRead )
				{
				case -1:
				case 0:
					throw new IOException();
				default:
					remaining -= bytesRead;
					offset += bytesRead;
					break;
				}
			}
		}

		private void readANEPPacketInto(pj_BufferHandle bh)
			throws IOException
		{
			int anepLength;
			byte[] buf;
			

			require(bh);
			require(bh.getOffset() == 0);
			require(bh.getByteArray());

			buf = bh.getByteArray();
			/* Read the ANEP header */
			assert(bh.getLength() > 8);
			this.read(buf, 0, 8);

			anepLength = (((buf[6] & 0xFF) << 8)
				      | (buf[7] & 0xFF));

			/* Read the rest of the packet */
			assert(bh.getLength() >= anepLength);
			this.read(buf, 8, anepLength - 8);

			bh.setLength(anepLength);

			ensure(bh.getByteArray() == buf);
		}

		public void run()
		{
			require(this.pb);

			try
			{
				/* Keep going while someone cares */
				while ( this.users != 0 )
				{
					readANEPPacketInto(this.pb);
					this.handlePacket(this.pb);
				}
			}
			catch (IOException e)
			{
			}

			try
			{
				this.is.close();
			}
			catch (IOException e)
			{
			}
		}
	}

	/**
	 * StreamWriter is a helper class used to write ANEP packes to an
	 * OutputStream.
	 */
	protected static class StreamWriter
	{
		private OutputStream os;
		private int users = 1;

		protected StreamWriter(OutputStream os)
		{
			this.os = os;
		}

		protected void acquire()
		{
			this.users++;
		}

		protected boolean release()
		{
			this.users--;
			return this.users == 0;
		}

		protected synchronized void write(pj_BufferHandle pb)
		{
			try
			{
				this.os.write(pb.getByteArray(),
					      pb.getOffset(),
					      pb.getLength());
			}
			catch (IOException e)
			{
			}
		}
	}

	PACKAGE int getPacketLength()
	{
		return 16384; // XXX XXX 
	}

}
