/*
 * pj_BufferHandle.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

/**
 * pj_BufferHandle does not need to do array bounds checks or
 * to synchronize access.  (Done by the wrapper layer.)
 * <p>
 * A pj_BufferHandle can be on one of three queues: a pj_InChannel's
 * free buffer queue, a pj_OutChannel's outgoing queue, or a
 * pj_ThreadPool's dispatched BufferQ (waiting to be handled).
 * <p>
 * XXX I belive all of the synchronized statements in here are redundant
 * with protections necessairly provided by the caller.  I think.
 * <p>
 * XXX cannot synchronize on the pj_BufferHandle object, as it is externally
 * visible and is shared with the dispatcher.
 *
 * @see BufferHandle
 * @see pj_InChannel
 * @see pj_OutChannel
 * @see pj_ThreadPool
 */
PACKAGE class pj_BufferHandle
{
	private static final int MAX_PACKET_LEN = 16 * 1024; // XX XXX XX XX

	/**
	 * The actual data buffer.  Should be allocated in
	 * fixed size chunks.
	 */
	private byte[] data;

	/**
	 * Length of the valid region of the data.  
	 */
	private int length;

	/**
	 * Current offset into the buffer
	 */
	private int offset;

	// --- Break following into a separate (superclass) WorkItem?

	/**
	 * For the free buffer queue on InChannels and OutChannels.
	 *
	 * @see InChannel
	 * @see OutChannel
	 */
	private pj_BufferHandle next;

	/**
	 * For the free buffer queue on InChannels and OutChannels.
	 *
	 * @see InChannel
	 * @see OutChannel
	 */
	private pj_BufferHandle prev;

	/**
	 * The deliver func associated with incoming packets that
	 * are queued up in a ThreadPool.
	 */
	private DeliverFunc deliverFunc;


	/**
	 * The payload start associated with incoming packets
	 * that are queud up in a ThreadPool.
	 *
	 * @see pj_InChannel#handle()
	 */
	private int payloadStart;


	PACKAGE static int getMaxPacketLen()
	{
		return MAX_PACKET_LEN;
	}

	protected pj_BufferHandle(pj_BufferHandle other, int offset, int length)
	{
		require(offset >= 0);
		require(length >= 0);
		require(offset <= MAX_PACKET_LEN);
		require(length <= MAX_PACKET_LEN);
		require(offset+length <= MAX_PACKET_LEN);

		if (other != null)
			this.cloneHandle(other, offset, length);
		else
		{
			this.data = new byte[MAX_PACKET_LEN];
			this.offset = offset;
			this.length = length;
		}
		
		ensure(this.data != null);
		ensure(this.offset >= 0);
		ensure(this.length >= 0);
	}

	PACKAGE int getLength()
	{
		return this.length;
	}

	PACKAGE int getOffset()
	{
		return this.offset;
	}

	/**
	 * Set the state associated with incoming buffers.
	 *
	 * @param f the deliver func associated with the InChannel
	 * this BufferHandle was dispatched out of
	 * @param payloadStart an integer associated with the buffer.
	 * internally used to mark the first byte after the filter in
	 * this buffer, but may be anything (from InChannelHandle.push()).
	 */
	PACKAGE void setIncomingState(DeliverFunc f, int payloadStart)
	{
		require(f);
		// Make no restrictions on payloadStart: require(payloadStart >= 0);
		require(this.deliverFunc == null);

		this.deliverFunc = f;
		this.payloadStart = payloadStart;
	}

	// DEBUGONLY
	PACKAGE void clearIncomingState()
	{
		this.deliverFunc = null;
		this.payloadStart = 0;
	}

	PACKAGE DeliverFunc getDeliverFunc()
	{
		return this.deliverFunc;
	}

	PACKAGE int getPayloadStart()
	{
		return this.payloadStart;
	}

	/**
	 * For pj_* layer internal use only.
	 */
	PACKAGE void setLength(int len)
	{
		require(len >= 0);
		require(len <= this.data.length);

		synchronized(this)
		{
			this.length = len;
		}
	}
	

	PACKAGE byte[] getByteArray()
	{
		return this.data;
	}

	/**
	 * Reset this BufferHandle, dropping the buffer
	 * it points to and setting offset and length to zero.
	 */
	PACKAGE void reset()
	{
		synchronized(this)
		{
			this.data = null;
			this.offset = 0;
			this.length = 0;
		}
	}

	PACKAGE void newBuffer(int offset, int length)
	{
		require(offset >= 0);
		require(length >= 0);
		require(length <= MAX_PACKET_LEN);
		
		synchronized(this)
		{
			// just drop this.data on the floor

			this.data = new byte[MAX_PACKET_LEN];
			this.offset = offset;
			this.length = length;
		}
	}

	PACKAGE void cloneHandle(pj_BufferHandle src, int fromOffset, int fromLength)
	{
		byte[] srcData;
		int srcOffset;
		int srcLength;

		/* BufferHandle did all of the bounds checks */

		require(fromOffset >= 0);
		require(fromLength >= 0);

		require(src.notOnAQueue());
		require(this.notOnAQueue());

		require(src.deliverFunc == null);
		require(src.payloadStart == 0);
		require(this.deliverFunc == null);
		require(this.payloadStart == 0);


		// Copy consistent information out of the src buffer
		synchronized(src) // synch with reset()
		{
			srcData = src.data;
			srcOffset = src.offset;
			srcLength = src.length;
		}

		assert(srcData.length == MAX_PACKET_LEN);
		assert((srcOffset + fromOffset + fromLength) <= srcLength);

		synchronized(this) // synch with reset()
		{
			// Just drop this.data (gotta love gc!)  

			this.data = srcData;
			this.offset = srcOffset + fromOffset;
			this.length = fromLength;
		}
	}

	/**
	 * Swap the contents of this pj_BufferHandle with the given
	 * src pj_BufferHandle.  Only the data, offset and length
	 * fields are copied (the other must be irrelevant).
	 * Only the specified sub-sequence of the src buffer is
	 * "copied" into this buffer handle.
	 *
	 * @param src the buffer to swap with this
	 * @param fromOffset the offset (in addition to src.offset) to set in this
	 * @param fromLength the new length of this buffer
	 */
	PACKAGE void swapWith(pj_BufferHandle src, int fromOffset, int fromLength)
	{
		byte[] srcData, thisData;
		int srcOffset, thisOffset;
		int srcLength, thisLength;

		/*
		 * Technically, we probably don't need to synchronize
		 * these reads and writes out of the buffers, as
		 * swapWith() is only called in a context where
		 * buffers are owned entirely by the system, and
		 * only visible to a single thread.
		 */

		require(fromOffset >= 0);
		require(fromLength >= 0);

		require(src.notOnAQueue());
		require(this.notOnAQueue());

		require(src.deliverFunc == null);
		require(src.payloadStart == 0);
		require(this.deliverFunc == null);
		require(this.payloadStart == 0);

		// Copy consistent information out of the src buffer
		synchronized(src) // synch with reset()
		{
			srcData = src.data;
			srcOffset = src.offset;
			srcLength = src.length;
		}

		assert(srcData.length == MAX_PACKET_LEN);
		assert(srcOffset >= 0);
		assert((srcOffset + fromOffset + fromLength) <= srcLength);

		// Copy consistent info out of this buffer, and push
		// src data into this.

		synchronized(this) // synch with reset()
		{
			thisData = this.data;
			thisOffset = this.offset;
			thisLength = this.length;

			this.data = srcData;
			this.offset = srcOffset + fromOffset;
			this.length = fromLength;
		}

		assert(thisData.length == MAX_PACKET_LEN);
		assert(thisOffset >= 0);
		
		// Copy consistent information into src
		synchronized(src)
		{
			src.data = thisData;
			src.offset = thisOffset;
			src.length = thisLength;
		}

	}

	PACKAGE byte getByte(int index)
	{
		require(index >= 0);
		require(index < length);
		return data[offset + index];
	}

	PACKAGE void setByte(int index, byte val)
	{
		require(index >= 0);
		require(index < length);
		data[offset + index] = val;
	}

	PACKAGE int getInt(int index)
	{
		require(index >= 0);
		require(index + 4 <= length);
		index += offset;
		byte[] buf = data;
		return (((buf[index++] & 0xFF) << 24) |
			((buf[index++] & 0xFF) << 16) |
			((buf[index++] & 0xFF) <<  8) |
			((buf[index]   & 0xFF)));
	}

	PACKAGE void setInt(int index, int val)
	{
		require(index >= 0);
		require(index + 4 <= length);

		index += offset;
		byte[] buf = data;
		data[index++] = (byte) (val >> 24);
		data[index++] = (byte) (val >> 16);
		data[index++] = (byte) (val >> 8);
		data[index] = (byte) (val);
	}

	PACKAGE void getBytes(int srcPosition, byte[] dest,
			      int destPosition, int destLength)
	{
		require(srcPosition >= 0);
		require(srcPosition + destLength <= length);

		System.arraycopy(data, offset + srcPosition,
				 dest, destPosition, destLength);
	}

	PACKAGE void setBytes(int destPosition, byte[] src,
			      int srcPosition, int srcLength)
	{
		require(destPosition >= 0);
		require(destPosition + srcLength <= length);

		System.arraycopy(src, srcPosition,
				 data, offset + destPosition, srcLength);
	}

	
	/**
	 * Debugging only.  Returns true if this buffer is
	 * not on a queue (i.e, next == prev == null).
	 */
	PACKAGE boolean notOnAQueue()
	{
		// Wholly on or, wholly off.
		assert(((next == null) && (prev == null))
		       || ((next != null) && (prev != null)));
		// Singleton list, or not
		assert(((next == this) && (prev == this))
		       || ((next != this) && (prev != this)));

		return (next == null); // implies (prev == null)
	}

	/**
	 * Debugging only.  Returns true if this buffer is on
	 * a queue.
	 */
	PACKAGE boolean onAQueue()
	{
		// Wholly on or, wholly off.
		assert(((next == null) && (prev == null))
		       || ((next != null) && (prev != null)));
		// Singleton list, or not
		assert(((next == this) && (prev == this))
		       || ((next != this) && (prev != this)));

		return (next != null); // implies (prev != null)
	}

	PACKAGE boolean isSingletonQueue()
	{
		if (this.next == this)
		{
			assert(this.prev == this);
			return true;
		}
		return false;
	}

	/**
	 * Add this packet buffer to the "head" of given queue q.
	 * Q is doubly linked, the new head (this) is returned.
	 * <p>
	 * If queue is:  A <> B <> C <> A, and this is X.  New
	 * queue will be X <> A <> B <> C <> X.
	 *
	 * @param q pointer to the first element of the list (or null).
	 * @return the new head of the list (always this).
	 */
	PACKAGE pj_BufferHandle addToQueueHead(pj_BufferHandle q)
	{
		require(this.notOnAQueue());

		if (q == null)
		{
			this.next = this;
			this.prev = this;
		}
		else
		{
			// Queue best not be pointing to this!
			assert(q != this);
			assert(q.next != this);
			assert(q.prev != this);
		
			// Queue best be well formed!
			assert(q.prev != null);
			require(q.onAQueue());
			require(q.prev.onAQueue());
			require(q.next.onAQueue());

			this.next = q;
			this.prev = q.prev;

			// fixup current first and last
			q.prev.next = this;
			q.prev = this;
		}
		
		ensure(this.onAQueue());

		// return new head of q -- this
		return this;
	}

	/**
	 * Add this pj_BufferHandle to tail of the given queue.
	 * Returns the new queue *head* (if it changes).
	 * <p>
	 * XXX Note that this is identical to addToQueueHead() except
	 * for the final return statement.
	 */
	PACKAGE pj_BufferHandle addToQueueTail(pj_BufferHandle q)
	{
		require(this.notOnAQueue());

		if (q == null)
		{
			this.next = this;
			this.prev = this;
			return this;
		}
		else
		{
			// Queue best not be pointing to this!
			assert(q != this);
			assert(q.next != this);
			assert(q.prev != this);
		
			// Queue best be well formed!
			assert(q.prev != null);
			require(q.onAQueue());
			require(q.prev.onAQueue());
			require(q.next.onAQueue());

			this.next = q;
			this.prev = q.prev;

			// fixup current first and last
			q.prev.next = this;
			q.prev = this;
		}

		ensure(this.onAQueue());
		return q;
	}
	
	PACKAGE pj_BufferHandle prev()
	{
		return this.prev;
	}
	
	/**
	 * Remove this pb from the queue it is on.  It must be on a
	 * queue.  Return the following item on the list (thus if this
	 * is head of list, then it returns a new head).
	 *
	 * @return the new <emph>head</emph> of the queue.
	 */
	PACKAGE pj_BufferHandle removeFromQueue()
	{
		final pj_BufferHandle newHead;

		require(this.onAQueue());

		if (this.next == this)
		{
			// singleton list.
			newHead = null;
		}
		else
		{
			newHead = this.next;
			
			// not a singleton
			this.next.prev = this.prev;
			this.prev.next = this.next;

			// Leave the list in good shape
			ensure(newHead.onAQueue());
		}
		
		this.next = null;
		this.prev = null;

		ensure(this.notOnAQueue());
		return newHead;
	}

	PACKAGE void dbgPacketSummary(StringBuffer sbuf, int offset, int len)
	{
		for(int i = offset; i < this.length; i++)
		{
			if(i >= len)
			{
				sbuf.append("...");
				break;
			}

			// I want two hex digits per byte.
			if((i % 16) == 0)
				sbuf.append("  ");
			byte val = data[i];
			byte hi = (byte)((int)(val&0xFF) >> 4);
			sbuf.append(Integer.toHexString(hi&0xF));
			sbuf.append(Integer.toHexString(val&0xF));
			sbuf.append(" ");
		}
	}
	
	public String toString()
	{
		return "pj_BufferHandle[data=" + this.data 
			+ "; off=" + offset
			+ "; len=" + length
			+ "]";
	}
}
