/*
 * pj_DemultiplexKey.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector;

/**
 * A key for matching incoming packets.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class pj_DemultiplexKey
{
	/**
	 * A segment of a demultiplex key.
	 */
	private static abstract class KeySegment
	{
		/**
		 * Offset from the previous segment
		 */
		public final int offset;

		public KeySegment(int offset)
		{
			this.offset = offset;
		}

		/**
		 * Attempt to match this segment in the given payload.
		 */
		public abstract int match(byte payload[],
					  int payloadStart,
					  int payloadLength);
	}

	/**
	 * A segment that matches a series of bytes.
	 */
	protected static class ByteMask
		extends KeySegment
	{
		/**
		 * Number of bytes to match
		 */
		public final int length;
		/**
		 * The actual bytes to match
		 */
		public final byte sequence[];
		/**
		 * Bitmask of which bytes to match
		 */
		public final boolean mask[];

		public ByteMask(int offset, int length,
				byte sequence[], boolean mask[])
		{
			super(offset);

			require(length <= sequence.length);
			require((mask == null)
				|| (sequence.length <= mask.length));

			this.length = length;
			this.sequence = sequence;
			this.mask = mask;
		}
		
		public int match(byte payload[],
				 int payloadStart,
				 int payloadLength)
		{
			int retval;

			/* Check for underrun */
			if( this.length <= (payloadLength - payloadStart) )
			{
				int lpc;

				/* Match the bytes */
				retval = payloadStart + this.length;
				for( lpc = 0;
				     (lpc < this.length) && (retval != -1);
				     lpc++ )
				{
					if( ((mask == null) || mask[lpc]) &&
					    (sequence[lpc] !=
					     payload[payloadStart + lpc]) )
					{
						retval = -1;
					}
				}
			}
			else
			{
				retval = -1;
			}
			return retval;
		}

		public String toString()
		{
			StringBuffer r = new StringBuffer(120);
			
			r.append("pj_DemultiplexKey.ByteMask[offset=")
				.append(this.offset)
				.append("; length=")
				.append(this.length)
				.append("; sequence='");
			
			for(int i = 0; i < length; i++)
			{
				if ((mask == null)
				    || (mask[i]))
					r.append(sequence[i] + " ");
				else
					r.append("X ");
			}
			r.append("'")
				.append("; mask=")
				.append(this.mask)
				.append("]");

			return r.toString();
		}
	}

	/**
	 * A key segment that matches an ANEP header
	 *
	 * 0               *   1           *       2       *           3
	 * 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
	 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	 * |   Version     |     Flags     |            Type ID            |
	 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	 * |      ANEP Header Length       |     ANEP Packet Length        |
	 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	 * ~                            Options                            ~
	 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	 * |                               .                               |
	 * ~                            Payload                            ~
	 * |                               .                               |
	 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	 */
	private static class ANEPHeader
		extends KeySegment
	{
		/**
		 * A byte mask for matching the first part of the header
		 */
		private final ByteMask headerMask;

		/**
		 * Bitmask for matching the first part of the header, we don't
		 * want to match the flags field.
		 */
		private static final boolean mask[] = {true, false, true, true};
		/**
		 * Similar to above, but we don't match the version either
		 */
		private static final boolean allMask[] = {false, false, true, true};

		/**
		 * Construct an ANEPHeader that matches a specific version and
		 * protocol ID.
		 */
		public ANEPHeader(int offset, byte version,
				  boolean allVersions, short protocolID)
		{
			super(offset);

			byte[] sequence;
			sequence = new byte[4];
			sequence[0] = version;
			sequence[2] = (byte)((protocolID >> 8) & 0xFF);
			sequence[3] = (byte)((protocolID) & 0xFF);
			this.headerMask = new ByteMask(0, 4, sequence,
						       allVersions ?
						       allMask : mask);
		}

		public int match(byte payload[],
				 int payloadStart,
				 int payloadLength)
		{
			int retval;

			if( ((retval = this.headerMask.match(payload,
							     payloadStart,
							     payloadLength))
			     > 0) &&
			    (payloadLength >= (retval + 4)) )
			{
				int headerLen =
					(((payload[payloadStart + 4] & 0xFF) << 8)
					 | (payload[payloadStart + 5] & 0xFF))
					* 4;

				if( headerLen < 8 ) // minimum allowed value.
				{
					TRACE(Trace.NODEOS_DISPATCH, "ANEPHeader Malformed packet: headerlen="+headerLen+ " < 8.");
					retval = -1;
				}
				else if( headerLen > payloadLength )
				{
					TRACE(Trace.NODEOS_DISPATCH, "ANEPHeader Malformed packet: headerlen="+headerLen+ "; payloadLength=" +payloadLength+ ".");
					retval = -1;
				}
				else
				{
					retval = payloadStart + headerLen;
				}
			}
			else
			{
				retval = -1;
			}
			return retval;
		}

		public String toString()
		{
			return "pj_DemultiplexKey.ANEPHeader[offset="
				+ this.offset
				+ "; headerMask="
				+ this.headerMask
				+ "]";
		}
	}

	/**
	 * The set of segments that make up this key.  The are matched based
	 * on their ordering.
	 */
	private final Vector segments = new Vector();

	/**
	 * The approximate length of data that this key will match.
	 */
	private int totalLength;

	PACKAGE pj_DemultiplexKey(pj_Flow flow)
	{
		require(flow);
		require(flow == pj_Flow.native_currentFlow());
	}

	/**
	 * Add a segment that matches the given sequence of bytes.
	 *
	 * @param offset The offset from the previous segment to start matching
	 * @param length The length of the bytes to match
	 * @param sequence The sequence of bytes to match
	 * @param mask If non-null, its a bit mask that indicates which bytes
	 * in `sequence' should be matched.
	 */
	PACKAGE void match(int offset, int length,
			  byte sequence[], boolean mask[])
	{
		this.segments.addElement(new ByteMask(offset, length,
						      sequence, mask));
		this.totalLength += offset + length;
	}

	/**
	 * Add a segment to match an IPv4 header
	 */
	PACKAGE void matchIPv4(int offset, int source, int dest, int protocol)
	{
		/* XXX ... */
		throw new InternalError("IPv4 header matching isn't done yet.");
	}

	/**
	 * Add a segment to match a UDP header, needs to be preceded by an
	 * IPv4 segment
	 */
	PACKAGE void matchUDP(int offset, int srcPort, int destPort)
	{
		/* XXX ... */
		throw new InternalError("UDP header matching isn't done yet.");
	}

	/**
	 * Add a segment to match a TCP header, needs to be preceded by an
	 * IPv4 segment
	 */
	PACKAGE void matchTCP(int offset, int srcPort, int destPort)
	{
		/* XXX ... */
		throw new InternalError("TCP header matching isn't done yet.");
	}
	
	/**
	 * Add a segment that matches an ANEP header
	 *
	 * @param offset The offset from the previous segment
	 * @param version The ANEP version number to match
	 * @param allVersions Match any version of ANEP
	 * @param protocolID The protocolID to match
	 */
	PACKAGE void matchANEP(int offset, byte version,
			      boolean allVersions, short protocolID)
	{
		this.segments.addElement(new ANEPHeader(offset,
							version,
							allVersions,
							protocolID));
		this.totalLength += 8;
	}

	/**
	 * Get the approximate length of the bytes that the can matches on.
	 */
	PACKAGE int length()
	{
		return this.totalLength;
	}

	/**
	 * Test a packet to see if this key matches.
	 *
	 * @param payload The packet data
	 * @param payloadStart The place to start matching in `payload'
	 * @param payloadLength The length of data in payload, including bytes
	 * up to payloadStart.
	 *
	 * @return 0 if there was an error otherwise the index of the first byte after the
	 * end of the key.
	 */
	PACKAGE int matches(byte payload[],
			    int payloadStart,
			    int payloadLength)
	{
		int lpc, size, currPayload = payloadStart, retval = 0;
		KeySegment ks;

		/* Walk over the key segments */
		size = this.segments.size();
		for( lpc = 0; (lpc < size) && (currPayload >= 0); lpc++ )
		{
			ks = (KeySegment)this.segments.elementAt(lpc);
			/* Skip to the segments offset */
			currPayload += ks.offset;
			/* Try to match this segment to the current payload */
			currPayload = ks.match(payload,
					       currPayload,
					       payloadLength);
		}
		if( currPayload >= 0 )
			retval = currPayload;
		return retval;
	}

	public String toString()
	{
		String retval = "pj_DemultiplexKey[totalLength="
			+ this.totalLength
			+ "; segments={";
		int lpc, size;

		size = this.segments.size();
		for( lpc = 0; lpc < size; lpc++ )
		{
			retval += this.segments.elementAt(lpc);
		}
		retval += "}]";
		return retval;
	}
}
