/*
 * pj_Flow.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector; // XXX

/**
 * Java implementation of a NodeOS Flow.
 * A flow is the closest approximation to a process in Janos.
 * 
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class pj_Flow
{
	/**
	 * The credentials provided when this pj_Flow was created.  This
	 * is the bucket from which the pj_Flow gets to extract resources
	 * and access Node facilities.
	 */
	private final pj_Credentials cred;

	/**
	 * The threadPool for this flow.
	 */
	private final pj_ThreadPool threadPool;

	/**
	 * The Memory limit for this flow.
	 */
	private final pj_MemSpec memSpec;

	/**
	 * Channels created in this pj_Flow.  This list only exists
	 * so that the channels can be cleanly destroyed when the
	 * flow is terminated.
	 */
	private final Vector channels = new Vector(5);

	/**
	 * The parent flow of this flow.  Used for handling termination
	 * requests.
	 * <p>
	 * XXX this is a cross-process pointer and must be treated
	 * specially.
	 */
	private final pj_Flow parentFlow;

	/**
	 * A per-flow Object that user code can set and get.
	 * (for associating things with the flow).
	 */
	private Object obj;

	/**
	 * A per-flow pointer to the termination handler for
	 * this flow.  Invoked when the flow is terminated.
	 */
	private final Runnable terminationHandler;

	/**
	 * XXX: need to examine the global locking structure for
	 * deadlock, etc
	 * <p>
	 * XXX: Probably should move to locks on each element of the
	 * state and not a global state lock.
	 */
	PACKAGE final Object lock = new Object();

	/**
	 * Initalize a new Flow.  See newFlow().
	 * <p>
	 * XXX implicitly starts threads running (blocked) on the thread pool.  Gak.
	 */
	PACKAGE pj_Flow(pj_Flow parentFlow,
			pj_Credentials cred,
			pj_ThreadPoolSpec threadPoolSpec,
			pj_MemSpec memSpec,
			Runnable termHandler)
	{
		this.parentFlow = parentFlow;
		this.cred = cred;
		this.memSpec = memSpec;
		this.terminationHandler = termHandler;

		this.threadPool = new pj_ThreadPool(this,
						    threadPoolSpec);
	}

	/**
	 * Create an explicit thread.  This is a new thread which is
	 * not subject to whatever concurrency limit this Flow's
	 * ThreadPool has.  But, it will be scheduled by the
	 * ThreadPool's scheduler.
	 * 
	 * @allocations: This function will allocate a new Thread object.
	 * @param r the Runnable to start the target thread on.
	 * @param name the String name of the new thread.
	 */
	PACKAGE void threadStart(java.lang.Runnable r, String name)
	{
		this.threadPool.threadStart(r, name);
	}

	private boolean destroyed = false;

	PACKAGE synchronized void destroy()
	{
		// For debugging, make sure this isn't invoked more than once
		assert(!destroyed);
		this.destroyed = true;

		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);

		// XXX really should do this in a separate thread!
		if (terminationHandler != null)
			terminationHandler.run();

		for (int i = 0; i < channels.size(); i++)
		{
			// Yes, this is stupid.  It is just temporary.
			// XXX make all the pj_channels extend
			// abstract class pj_AnyChannel

			Object c = channels.elementAt(i);
			if (c instanceof pj_InChannel)
			{
				((pj_InChannel)c).destroy();
			}
			else if (c instanceof pj_OutChannel)
			{
				((pj_OutChannel)c).destroy();
			}
			else if (c instanceof pj_CutThroughChannel)
			{
				((pj_CutThroughChannel)c).destroy();
			}
			else
			{
				panic("Non-channel object in channel list.");
			}
		}

		threadPool.destroy();
	}
	
	PACKAGE void addChannel(pj_InChannel c)
	{
		channels.addElement(c);
	}

	PACKAGE void addChannel(pj_OutChannel c)
	{
		channels.addElement(c);
	}

	PACKAGE void addChannel(pj_CutThroughChannel c)
	{
		channels.addElement(c);
	}

	PACKAGE Object getObject()
	{
		return obj;
	}

	PACKAGE void setObject(Object obj)
	{
		this.obj = obj;
	}

	PACKAGE pj_ThreadPool getThreadPool()
	{
		return threadPool;
	}

	public String toString()
	{
		if (false) // longwinded
			return "Flow["
				DEBUGONLY(+ "name='" +name)
				+ "; parent=" + ((parentFlow==null)?"no":"yes")
				+ "; cred=" + cred
				+ "; obj=" + obj
				+ "]";
		else // short and sweet
			return "Flow[" 
				DEBUGONLY(+ "name='" +name)
				+ "; parent=" + ((parentFlow==null)?"n":"y")
				+ "]";
	}

	/**
	 * Get the current flow.  The current Flow is the
	 * flow associated with the current Thread.
	 *
	 * @see Thread#getFlow()
	 */
	PACKAGE static pj_Flow native_currentFlow()
	{
		return pj_Thread.self().getFlow();
	}

#ifdef DEBUG
	// always non-null.
	private String name = "John Doe Flow";
	
	PACKAGE void setName(String name)
	{
		if (name != null)
			this.name = name;
	}

	PACKAGE String getName()
	{
		return name;
	}
#endif /*DEBUG*/	
}

