/*
 * pj_InChannel.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector; // XXX
import java.io.IOException;

/**
 * Pure Java implementation of a NodeOS InChannel.
 *
 * @see InChannel
 * @see OutChannel
 * @see pj_OutChannel
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class pj_InChannel
{
	/**
	 * The flow this receive channel is associated with.
	 *
	 * XXX what is this used for?
	 */
	PACKAGE final pj_Flow flow;

	/**
	 * The demultiplex key associated with this channel.
	 */
	PACKAGE final pj_DemultiplexKey classifier;

	/**
	 * Need the handler so the pj_InChannel can remove itself
	 * when it is destroyed.
	 * <p>
	 * XXX is this necessary?
	 */
	private pj_InChannelDispatcher handler;
	
	/**
	 * The callback for delivery.
	 */
	private final DeliverFunc deliverFunc;

	/**
	 * Queue of free packet buffers for this pj_InChannel.
	 * Queue is circular, doubly-linked list.  freeBufferQ points
	 * to the "head", freeBufferQ.prev points to the tail.
	 */
	private pj_BufferHandle freeBufferQ;

	private final pj_ThreadPool threadPool;

	/**
	 * Lock for protecting the free packet buffer queue.
	 */
	private final Object freeBufferQLock = new Object(); // XXX

	PACKAGE pj_InChannel(pj_Flow flow,
			     pj_DemultiplexKey classifier,
			     String protocolSpec,
			     String addrSpec,
			     DeliverFunc deliverFunc)
		throws IOException,
		PacketAddressFormatException,
		ProtocolSpecFormatException
	{
		pj_ProtocolLayer pl;
		pj_ProtocolSpec ps;
		String layername;
		
		this.flow = flow;
		this.threadPool = flow.getThreadPool(); // XXX steal Flow's pool for now
		this.classifier = classifier;
		this.deliverFunc = deliverFunc;

		/* Find the matching protocol layer */
		ps = new pj_ProtocolSpec(protocolSpec);
		layername = ps.getComponent(ps.componentCount() - 1);
		pl = pj_ProtocolLayer.findLayer(layername);
		if( pl != null )
		{
			pj_ProtocolLayer.PacketAddressMask source;
			pj_InChannelDispatcher icd;

			/* Get the source address mask */
			source = pl.createAddressMask();
			source.parseAddressMask(addrSpec);
			icd = pj_InChannelDispatcher.getDispatcher(pl, source);
			this.handler = icd;
			icd.addChannel(this);
		}
		flow.addChannel(this);
	}
	
	/**
	 * Dispatch an incoming packet for this in channel.
	 * <p>
	 * If the channel has available buffers then its
	 * 
	 * <p>
	 * This is operating in "interrupt context", that is this is
	 * all system time up to this point.
	 * 
	 * @param systembh the incoming buffer
	 * @param packetStart the "start" of the packet as seen by the receiver.
	 * @param payloadStart the start of the payload (i.e., skipping matched header bytes).
	 * @return returns true if the swap happened.
	 */
	PACKAGE boolean handle(pj_BufferHandle systembh,
			       int packetStart,
			       int payloadStart)
	{
		pj_BufferHandle freebh;

		require(systembh);
		require(packetStart >= 0);
		// Make no restrictions on payload start: require(payloadStart >= 0);
		
		freebh = native_remBuffer();
		
		if (freebh != null)
		{
			// Swap the buffers between freebh and systembh
			// BufferHandles do *not* migrate around the system.
			freebh.swapWith(systembh, packetStart,
					 systembh.getLength() - packetStart);
			freebh.setIncomingState(this.deliverFunc, payloadStart);
			this.threadPool.addWork(freebh);
			return true;
		}

		return false;
	}

	/**
	 * Add a buffer to the set of buffers available for buffering
	 * incoming packets to this pj_InChannel.
	 * <p>
	 * This is run as a LIFO queue to maximize locality.
	 */
	PACKAGE void native_addBuffer(pj_BufferHandle bufh)
	{
		synchronized(this.freeBufferQLock)
		{
			// Push on the head
			freeBufferQ = bufh.addToQueueHead(freeBufferQ);			
		}
	}

	/**
	 * Remove a buffer from the queue of free buffers
	 * on this pj_InChannel.  Returns null if there are
	 * no free buffers on this pj_InChannel.
	 * <p>
	 * This is run as a LIFO queue to maximize locality.
	 */
	PACKAGE pj_BufferHandle native_remBuffer()
	{
		pj_BufferHandle retval = null;

		synchronized(this.freeBufferQLock)
		{
			// Pop the head
			if (freeBufferQ != null)
			{
				retval = freeBufferQ;
				freeBufferQ = freeBufferQ.removeFromQueue();
			}
		}
		return retval;
	}
	
	public String toString()
	{
		return "pj_InChannel[flow="
			+ flow
			+ "; handler="
			+ handler
			+ "; classifier="
			+ classifier
			+ "; threadPool="
			+ threadPool
			+ "; deliverFunc="
			+ deliverFunc
			+ "]";
	}

	PACKAGE void destroy()
	{
		// XXX
		this.handler.remChannel(this);
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);
	}
	
	protected void finalize()
	{
		this.handler.remChannel(this);
	}
}
