/*
 * pj_OutChannel.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.io.IOException;

import java.net.DatagramSocket;
import java.net.DatagramPacket;

IMPORT_DEBUG

/**
 * Java wrapper for an OutChannel.
 *
 * @see CutThroughChannel
 * @see InChannel
 * @see OutChannel
 */
PACKAGE class pj_OutChannel
	extends pj_ProtocolLayer.PacketGenerator
{
	/**
	 * Queue of buffers waiting to be sent.  Only used
	 * for asynchronous sends (which actually aren't implemented yet!)
	 */
	private pj_BufferHandle bufferQ;

	/**
	 * Lock for the bufferQ
	 */
	private final Object bufferQLock = new Object();

	PACKAGE final pj_Flow flow;
	PACKAGE pj_ProtocolLayer protocolLayer;

	PACKAGE pj_OutChannel(pj_Flow flow,
			     String protocolSpec,
			     String addressSpec)
		throws IOException,
		       PacketAddressFormatException,
		       ProtocolSpecFormatException
	{
		pj_ProtocolLayer pl;
		pj_ProtocolSpec ps;
		
		require(flow);
		require(protocolSpec);
		require(addressSpec);

		this.flow = flow;

		/* Find the matching protocol layer */
		ps = new pj_ProtocolSpec(protocolSpec);
		if( (pl = pj_ProtocolLayer.
		     findLayer(ps.getComponent(ps.componentCount() - 1)))
		    != null )
		{
			AddressSpec as;

			/*
			 * Get an address object corresponding to the
			 * destination
			 */
			as = pl.createAddress(addressSpec);
			this.destination = as;
			pl.attachGenerator(this);
			this.protocolLayer = pl;
		}
		else
		{
			throw new IOException("protocolSpec="+protocolSpec+
					      " addressSpec="+addressSpec+
					      " ps="+ps);
		}

		flow.addChannel(this);
	}

	/**
	 * Packet generator function defined by ProtocolLayer.PacketGenerator.
	 */
	PACKAGE pj_BufferHandle generate()
	{
		pj_BufferHandle retval;
		
		synchronized(this.bufferQLock)
		{
			// Pull head of list
			retval = this.bufferQ;

			if (retval != null) 
			{
				// Remove head from list
				bufferQ = retval.removeFromQueue();
			}
		}
		return retval;
	}

	/**
	 * Queue a packet for sending whenever convenient.
	 *
	 * XXX needs callback function
	 */
	PACKAGE void queueAsyncPacket(pj_BufferHandle pb)
	{
		require(pb);

		synchronized(this.bufferQLock)
		{
			bufferQ = pb.addToQueueTail(bufferQ);
		}
		protocolLayer.generatorReady(this); 
	}

	/**
	 * Send a single buffer.  Synchronous (blocks until
	 * send is "done".
	 */
	PACKAGE void send(pj_BufferHandle bh)
	{
		require(bh);

		TRACE(Trace.NODEOS_OUTGOING, "sending: destination=" +this+ " data=" +Debug.dbgPacketSummary(bh));

		protocolLayer.synchronousSend(this, bh);
	}

	/**
	 * Send a single buffer.  Synchronous.  Puts the
	 * buffer on the specified inChan after the send.
	 */
	PACKAGE void sendAndRecycle(pj_BufferHandle bh, pj_InChannel inChan)
	{
		require(bh);
		require(inChan);

		TRACE(Trace.NODEOS_OUTGOING, "sending(w/recycle): destination=" +this+ " data=" +Debug.dbgPacketSummary(bh));
		send(bh);

		inChan.native_addBuffer(bh);
	}
	
	/**
	 * Queue a packet for sending on this OutChannel.  callback
	 * will be invoked when the call is complete.
	 * <p>
	 * XXX doesn't work yet.
	 * <p>
	 * XXX need to reconcile with synchronous sends.
	 */
	PACKAGE void sendAsynch(pj_BufferHandle bh, DeliverFunc callback)
	{
		require(bh);

		TRACE(Trace.NODEOS_OUTGOING, "sending(async): destination=" +this+ " data=" +Debug.dbgPacketSummary(bh));

		queueAsyncPacket(bh);

		NOTIMPLEMENTED("sendAsynch()");
	}


	/**
	 * Send an array of buffer handles.
	 */
	PACKAGE void sendv(pj_BufferHandle[] bhl)
	{
		require(bhl); // XXX
		require(bhl.length > 0); // XXX

		if (bhl.length == 1) 
		{
			this.send(bhl[0]);
		}
		else
		{
			NOTIMPLEMENTED("sendv() of a >1 length array of buffers");
		}
	}

	PACKAGE void destroy()
	{
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);
		this.protocolLayer.detachGenerator(this);
	}

	public String toString()
	{
		return "pj_OutChannel:"
			+ super.toString()
			+ "[flow="
			+ this.flow
			+ "; protocolLayer="
			+ this.protocolLayer
			+ "]";
	}
}
