/*
 * pj_ProtocolLayer.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.util.Hashtable;
import java.io.IOException;

/**
 * pj_ProtocolLayer is supposed to act as an abstraction for simulating
 * transport layers.  Its made up of several interfaces and classes for
 * moving packets through the layer to their destination.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE abstract class pj_ProtocolLayer
{
	/**
	 * The maximum packet length for all the layers.  Since packets are
	 * exchanged with users of the layers we have to make packets big
	 * enough to accomodate all the different protocol layers.
	 */
	private static int maxPacketLength = 0;

	/**
	 * Return the current value of maxPacketLength.
	 */
	PACKAGE static int getMaxPacketLength()
	{
		return pj_ProtocolLayer.maxPacketLength;
	}

	/**
	 * The table of all registered protocol layers.
	 */
	private static Hashtable layers = new Hashtable();

	/**
	 * Install a new protocol layer, basically, this just makes the layer
	 * PACKAGEally available.
	 */
	PACKAGE static void installLayer(pj_ProtocolLayer pl)
	{
		pj_ProtocolLayer.layers.put(pl.getID(), pl);
		pj_ProtocolLayer.maxPacketLength =
			Math.max(pj_ProtocolLayer.maxPacketLength,
				 pl.getPacketLength());
	}

	/**
	 * Find an installed protocol layer with the given name.
	 */
	PACKAGE static pj_ProtocolLayer findLayer(String id)
	{
		return (pj_ProtocolLayer)pj_ProtocolLayer.layers.get(id);
	}

	/**
	 * The global packet recycler.
	 */
	private static PacketRecycler recycler;
	
	PACKAGE static void attachRecycler(PacketRecycler pr)
	{
		pj_ProtocolLayer.recycler = pr;
	}

	PACKAGE static void recycle(pj_BufferHandle pb)
	{
		pj_ProtocolLayer.recycler.recycle(pb);
	}

	PACKAGE static void detachRecycler(PacketRecycler pr)
	{
		pj_ProtocolLayer.recycler = null;
	}

	/**
	 * Return the string identifier for this protocol layer.
	 */
	PACKAGE abstract String getID();
	/**
	 * Get the maximum packet length for this protocol layer.  This
	 * is pretty much for internal use only.
	 */
	PACKAGE abstract int getPacketLength();
	/**
	 * Create an address object that is compatible with this protocol
	 * layer.
	 */
	PACKAGE abstract AddressSpec createAddress(String addr)
		throws PacketAddressFormatException;
	/**
	 * Create an address mask object that is compatible with this
	 * protocol layer.
	 */
	PACKAGE abstract PacketAddressMask createAddressMask();
	/**
	 * Attach a packet handler to the layer.  The source mask for the
	 * handler determines what packets will be dispatched to the handler.
	 */
	PACKAGE abstract void attachHandler(PacketHandler ph)
		throws IOException;
	/**
	 * Detach a packet handler from the layer.
	 */
	PACKAGE abstract void detachHandler(PacketHandler ph);
	/**
	 * Attach a packet generator to the layer.  The destination of the
	 * packet is determined by the destination address object in the
	 * packet generator object.
	 */
	PACKAGE abstract void attachGenerator(PacketGenerator pg)
		throws IOException;

	/**
	 * Signal that a generator has asynchronous packets and is
	 * ready to be serviced.
	 */
	PACKAGE abstract void generatorReady(PacketGenerator pg);

	/**
	 * Synchronous send a packet, blocking until the send
	 * completes.
	 */
	PACKAGE abstract void synchronousSend(PacketGenerator pg, pj_BufferHandle bh);

	/**
	 * Detach a packet generator from the protocol layer.
	 */
	PACKAGE abstract void detachGenerator(PacketGenerator pg);

	/**
	 * An interface that represents a mask that can be applied to
	 * incoming packets to determine matching packets.
	 */
	PACKAGE interface PacketAddressMask
	{
		/**
		 * Identical to PacketAddress.parseAddress().
		 */
		void parseAddressMask(String addr)
			throws PacketAddressFormatException;
	}

	/**
	 * PacketHandler is an abstract class that is subclassed by the
	 * users of protocol layers to handle incoming packets.
	 */
	PACKAGE static abstract class PacketHandler
	{
		/**
		 * The source mask determines what incoming packets will be
		 * sent to the handle method.
		 */
		PACKAGE PacketAddressMask source;

		/**
		 * The method that will handle incoming packets.  Returns
		 * true if the packet was handled, false if not.
		 */
		PACKAGE abstract boolean handle(pj_BufferHandle pb);

		public String toString()
		{
			return "PacketHandler[source="
				+ this.source
				+ "]";
		}
	}

	/**
	 * PacketGenerator is an abstract class that is subclassed by the
	 * users of protocol layers to handle outgoing packets.
	 */
	PACKAGE static abstract class PacketGenerator
	{
		/**
		 * Function called by the protocol layer to get more
		 * asynchronous packets to send.
		 */
		PACKAGE abstract pj_BufferHandle generate();

		/**
		 * The destination of any generated packets.
		 */
		PACKAGE AddressSpec destination;
		
		public String toString()
		{
			return "PacketGenerator[destination="
				+ this.destination
				+ "]";
		}
	}

	PACKAGE interface PacketRecycler
	{
		void recycle(pj_BufferHandle pb);
	}
}
