/*
 * pj_TCPStreamLayer.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.util.Hashtable;
import java.util.Vector;
import java.io.IOException;
import java.net.Socket;
import java.net.ServerSocket;

/**
 * pj_TCPStreamLayer is a protocol layer implementation that sends
 * ANEP packets over TCP.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class pj_TCPStreamLayer
	extends pj_ANEPStreamLayer
{
	PACKAGE pj_TCPStreamLayer()
	{
	}

	PACKAGE String getID()
	{
		return "tcp";
	}

	PACKAGE AddressSpec createAddress(String addr)
		throws PacketAddressFormatException
	{
		return new AddressSpecPortAddr(addr);
	}

	PACKAGE PacketAddressMask createAddressMask()
	{
		return new pj_PortAddressMask();
	}

	/**
	 * TCPStreamAcceptor is used to accept connections on a tcp port
	 * and then start a reader thread for the new connection.
	 */
	PACKAGE static class TCPStreamAcceptor
		implements Runnable
	{
		/**
		 * The list of masks that match this source port
		 */
		PACKAGE final pj_PacketHandlerList masks;

		/**
		 * The socket to listen to
		 */
		PACKAGE ServerSocket acceptor;

		/**
		 * Indicates whether we're alive and should continue listening
		 * for more connections
		 */
		PACKAGE boolean dying = false;

		/**
		 * The set of readers that we've spawned for each new
		 * connection
		 */
		PACKAGE final Vector readers;
		
		PACKAGE TCPStreamAcceptor(int port)
			throws IOException
		{
			this.masks = new pj_PacketHandlerList();
			this.readers = new Vector();
			this.acceptor = new ServerSocket(port, 4);
		}

		public void run()
		{
			try
			{
				while ( !dying )
				{
					TCPStreamReader sr;
					Socket socket;

					/*
					 * Accept a connection and start the
					 * reader
					 */
					socket = this.acceptor.accept();
					sr = new TCPStreamReader(socket);
					/* Keep track of all readers */
					this.readers.addElement(sr);
					sr.masks = this.masks;
					new java.lang.Thread(sr).start();
				}
			}
			catch (IOException e)
			{
			}
			try
			{
				int lpc;

				/*
				 * Cleanup ourselves and connections we've
				 * spawned
				 */
				this.acceptor.close();
				for ( lpc = 0;
				      lpc < this.readers.size();
				      lpc++ )
				{
					((TCPStreamReader)this.readers.
					 elementAt(lpc)).release();
				}
			}
			catch (IOException e)
			{
			}
		}
	}

	/**
	 * StreamReader subclass which works over TCP streams and dispatches
	 * packets through an IPHandlerList.
	 */
	PACKAGE static class TCPStreamReader
		extends pj_ANEPStreamLayer.StreamReader
	{
		PACKAGE pj_PacketHandlerList masks; // XXX who sets?  when?
		private Socket socket;

		PACKAGE TCPStreamReader(Socket socket)
			throws IOException
		{
			// XXX wrong layer to create BufferHandle()
			super(socket.getInputStream(), BufferHandle.create());
			this.socket = socket;
		}
		
		protected boolean handlePacket(pj_BufferHandle pb)
		{
			PacketHandler ph;

			ph = this.masks.matchMask(this.socket);
			if (ph != null)
			{
				return ph.handle(pb);
			}
			return false;
		}
	}

	/**
	 * StreamWriter subclass which transmits packets over a TCP socket.
	 */
	PACKAGE static class TCPStreamWriter
		extends pj_ANEPStreamLayer.StreamWriter
	{
		PACKAGE Socket socket;

		PACKAGE TCPStreamWriter(Socket socket)
			throws IOException
		{
			super (socket.getOutputStream());
			this.socket = socket;
		}

		PACKAGE void destroy()
		{
			try
			{
				this.socket.close();
			}
			catch (IOException e)
			{
			}
		}
	}

	/**
	 * Used to map ports to acceptors.
	 */
	private final Hashtable streamAcceptors = new Hashtable();

	/**
	 * Used to map destinations to already opened sockets.
	 */
	private final Hashtable streamWriters = new Hashtable();
	
	PACKAGE void attachHandler(PacketHandler ph)
		throws IOException
	{
		int port = ((pj_PortAddressMask)ph.source).destPort;
		Integer i = new Integer(port);
		TCPStreamAcceptor sa;

		synchronized (this.streamAcceptors)
		{
			/*
			 * Check if we already have an acceptor for this port,
			 * otherwise create a new one.
			 */
			sa = (TCPStreamAcceptor)this.streamAcceptors.get(i);
			if (sa == null)
			{
				sa = new TCPStreamAcceptor(port);
				new java.lang.Thread(sa).start();
				this.streamAcceptors.put(i, sa);
			}
			sa.masks.addHandler(ph);
		}
	}

	PACKAGE void detachHandler(PacketHandler ph)
	{
		int port = ((pj_PortAddressMask)ph.source).destPort;
		Integer i = new Integer(port);
		TCPStreamAcceptor sa;

		synchronized (this.streamAcceptors)
		{
			/*
			 * Remove the packet handler for this acceptor, if
			 * there are no more handlers than kill the acceptor
			 */
			sa = (TCPStreamAcceptor)this.streamAcceptors.get(i);
			assert(sa);
			if ( sa.masks.remHandler(ph) )
			{
				sa.dying = true;
				this.streamAcceptors.remove(sa);
			}
		}
	}
	
	PACKAGE void attachGenerator(PacketGenerator pg)
	{
		TCPStreamWriter sw;

		synchronized (this.streamWriters)
		{
			sw = (TCPStreamWriter)streamWriters.
				get(pg.destination);

			/* Create a new writer for this destination */
			if ( sw == null )
			{
				AddressSpecPortAddr ia;
				
				ia = (AddressSpecPortAddr)pg.destination;
				try
				{
					sw = new TCPStreamWriter(
						new Socket(ia.destAddr(),
							   ia.destPort()));
					this.streamWriters.put(ia, sw);
				}
				catch (IOException e)
				{
				}
			}
			else
			{
				sw.acquire();
			}
		}
	}

	/**
	 * Synchronous send a packet, blocking until the send
	 * completes.
	 */
	PACKAGE void synchronousSend(PacketGenerator pg,
				     pj_BufferHandle bh)
	{
		TCPStreamWriter sw;
		
		require(pg);
		require(bh);

		synchronized (this.streamWriters)
		{
			/* Get the stream writer */
			sw = (TCPStreamWriter)streamWriters.
				get(pg.destination);

			if ( sw == null )
			{
				this.attachGenerator(pg);
				sw = (TCPStreamWriter)streamWriters.
					get(pg.destination);
			}

			/* Do the write */
			if ( sw != null )
			{
				sw.write(bh);
			}
		}
	}
	
	/**
	 * Indicate that the generator is ready for an
	 * async send.
	 * <p>
	 * Just does a synchronous send inline, though.
	 */
	PACKAGE void generatorReady(PacketGenerator pg)
	{
		TCPStreamWriter sw;
		pj_BufferHandle bh;

		bh = pg.generate();
		synchronousSend(pg, bh);
	}

	PACKAGE void detachGenerator(PacketGenerator pg)
	{
		TCPStreamWriter sw;

		synchronized (this.streamWriters)
		{
			/*
			 * Remove the stream writer if there are no more users
			 */
			sw = (TCPStreamWriter)streamWriters.
				get(pg.destination);
			if ( (sw != null) && sw.release() )
			{
				this.streamWriters.remove(sw);
				sw.destroy();
			}
		}
	}
}
