/*
 * pj_Thread.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

/**
 * A "Native" Thread.
 * <p>
 * A package-private wrapper for the "native" thread
 * layer.
 * <p>
 * This is a simple wrapper for java.lang.Thread.  Its
 * mostly a separate class so I can cleanly subclass it
 * without exposing all of the standard java.lang.Thread
 * methods.
 * <p>
 * PART OF THE PURE JAVA NODEOS IMPLEMENTATION.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class pj_Thread
{
	/**
	 * Per-thread state.
	 * <p>
	 * This is part of the NodeOS spec.  Probably doesn't need to
	 * be part of the Java-level interface (users of this should
	 * just subclass Janos Thread).
	 */
	private Object obj;
	
	/**
	 * The ThreadPool this thread is a part of.
	 */
	private final pj_ThreadPool threadPool;

	private final ThreadWrapper jlThread;

	/**
	 * Hide java.lang.Thread in a member class so that subclasses
	 * of pj_Thread don't inherit the public java.lang.Thread
	 * methods.
	 */
	private class ThreadWrapper
		extends java.lang.Thread
	{
		/**
		 * The Runnable to use when thread is started.
		 */
		private final java.lang.Runnable r;

		private ThreadWrapper(Runnable r, String name)
		{
			super(name);

			require(r);
			this.r = r;
			this.setDaemon(false);
		}

		public void run()
		{
			try
			{
				assert(r != null);
				r.run();
			}
			catch(Throwable t)
			{
				System.out.println("Exception in thread " +this+ " (runnable = " +r+ ")");
				System.out.println("Thread blew up with:" +t.getMessage());
				t.printStackTrace();
			}
		}

		/**
		 * Return container class.  Used to convert from
		 * java.lang.Thread.currentThread() to a pj_Thread().
		 */
		private pj_Thread pj_Thread()
		{
			return pj_Thread.this;
		}
	}

	/**
	 * InfoHook class that reports on all of the native threads
	 * in the system.
	 */
	private static class pj_ThreadInfo
		extends InfoHook
	{
		private pj_ThreadInfo()
		{
			super("threads");
		}
		
		public String[] getInfo()
		{
			String[] retval;
			java.lang.Thread[] allThreads;
			ThreadGroup r, p;

			// find root ThreadGroup
			r = pj_Thread.currentThread().jlThread.getThreadGroup();
			p = r;
			while (p != null)
			{
				r = p;
				p = r.getParent();
			}
			assert(r != null);

			allThreads = new java.lang.Thread[r.activeCount() + 50];
			int ct = r.enumerate(allThreads, true);
			
			retval = new String[ct];
			for (int i = 0; i < ct; i++)
			{
				retval[i] = allThreads[i].toString()
					+ " name="
					+ allThreads[i].getName();
			}

			return retval;
		}
	}

	static
	{
		Status.addInfoHook(new pj_ThreadInfo());
	}

	/**
	 * Create a new native thread.  This constructor is
	 * used to only create "explicit" threads.
	 * <p>
	 * The given Runnable will be used when the
	 * thread is started (instead of the run method on this).
	 *
	 * @allocations: A new ThreadWrapper
	 */
	PACKAGE pj_Thread(pj_ThreadPool threadPool, Runnable runnable, String name)
	{
		require(threadPool);
		require(runnable);
		
		this.jlThread = new ThreadWrapper(runnable, name);
		this.threadPool = threadPool;

		DEBUGONLY(this.jlThread.setPriority(5););

		TRACE(Trace.NODEOS_THREADS, " Created new thread: " +this);
	}

	PACKAGE static void yieldTo(pj_Thread th)
	{
		java.lang.Thread.yield();
	}

	PACKAGE static void sleep(long millis, int nanos)
		throws InterruptedException
	{
		java.lang.Thread.sleep(millis, nanos);
	}

	PACKAGE static Object getObject()
	{
		return self().obj;
	}

	PACKAGE static Object setObject(Object obj)
	{
		pj_Thread self = self();
		Object old;

		// XXX publically visible lock, but only hoses user
		synchronized(self) 
		{
			old = self.obj;
			self.obj = obj;
		}
		return old;
	}


	/**
	 * Package-private run method that all *implicit* threads
	 * start in.
	 */
	PACKAGE void start()
	{
		this.jlThread.start();
	}

	PACKAGE pj_Flow getFlow()
	{
		return this.threadPool.native_getFlow();
	}

	/**
	 * Return the Thread object associated with the currently
	 * executing thread of control.
	 *
	 * @see currentThread()
	 * @return a Thread object associated with the current thread.
	 */
	PACKAGE static pj_Thread self() 
	{
		return currentThread();
	}

	/**
	 *
	 * @see #self()
	 */
	PACKAGE static pj_Thread currentThread()
	{
		java.lang.Thread jth;
		jth = java.lang.Thread.currentThread();
#ifdef DEBUG
		if (!(jth instanceof ThreadWrapper))
		{
			Node.printStackTrace(null);
			Node.fatalError("ERROR: current thread is not a pj_Thread.ThreadWrapper!");
		}
#endif
		return ((ThreadWrapper)jth).pj_Thread();
	}

	public String toString()
	{
		String name = this.jlThread.getName();
		return "pj_Thread[n=" +name+ "]";
	}
}
