/*
 * pj_UDPLayer.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.io.IOException;
import java.util.Hashtable;
import java.util.Vector;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.SocketException;

/**
 * UDPLayer is a protocol layer implementation that runs over UDP.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class pj_UDPLayer
	extends pj_ProtocolLayer
{
	/**
	 * Datgram wrapper for sending packets.  The stupid empty
	 * buffer is replaced with the pj_BufferHandle's buffer right
	 * before the send.
	 */
	private final DatagramPacket dp = new DatagramPacket(new byte[0], 0);

	/**
	 * Output UDP socket for sending packets.
	 */
	private final DatagramSocket ds;
	
	PACKAGE pj_UDPLayer()
		throws SocketException
	{
		this.ds = new DatagramSocket();
	}

	public String toString()
	{
		return "UDPLayer[ds=" +ds+ "]";
	}

	/**
	 * The table of sockets that we have open for reading, organized by
	 * port number.
	 */
	private final Hashtable sockets = new Hashtable();

	/**
	 * Class that takes care of reading sockets and calling any associated
	 * packet handlers.
	 */
	private static class SocketReader
		implements Runnable
	{
		/**
		 * Vector of packet handlers attached to this socket.
		 */
		PACKAGE final pj_PacketHandlerList masks;
		PACKAGE final DatagramSocket ds;
		PACKAGE boolean dying = false;
		
		/**
		 * Initial packet buffer that first packet is recevied into.
		 */
		private final pj_BufferHandle pb;

		PACKAGE SocketReader(int port, pj_BufferHandle pb)
			throws SocketException
		{
			this.masks = new pj_PacketHandlerList();
			this.ds = new DatagramSocket(port);
			this.pb = pb;
		}

		public void run()
		{
			final DatagramPacket dp;
			PacketHandler ph;

			require(this.pb);
			require(this.pb.getByteArray().length > 0);

			// Put the initial buffer in DatagramPacket
			dp = new DatagramPacket(pb.getByteArray(), pb.getByteArray().length);

			while (!dying)
			{
				try
				{
					this.ds.receive(dp);

					// Set effective length of pb to received data
					pb.setLength(dp.getLength());
					TRACE(Trace.NODEOS_INCOMING, "in!");
					TRACE(Trace.NODEOS_INCOMING, "received: " + pj_Debug.dbgPacketSummary(pb));
					synchronized(this.masks)
					{
						ph = this.masks.matchMask(dp);
					}
					if( ph != null )
					{
						// Swap up for new packet buffer
						ph.handle(pb);

						// pb may or may not point to a new buffer
						dp.setData(pb.getByteArray());
						dp.setLength(pb.getByteArray().length);
					}
				}
				catch(Exception e)
				{
					Node.println("exception!! " + e);
					e.printStackTrace();
				}
			}
			TRACE(Trace.NODEOS_INCOMING, "recv port (on " +this.ds.getLocalPort()+ "): exiting");
			this.ds.close();
		}
	}

	PACKAGE String getID()
	{
		return "udp";
	}

	PACKAGE int getPacketLength()
	{
		return 16384;
	}

	PACKAGE AddressSpec createAddress(String addr)
		throws PacketAddressFormatException
	{
		return new AddressSpecPortAddr(addr);
	}

	PACKAGE PacketAddressMask createAddressMask()
	{
		return new pj_PortAddressMask();
	}

	PACKAGE void attachHandler(PacketHandler ph)
		throws IOException
	{
		int port = ((pj_PortAddressMask)ph.source).destPort;
		Integer i = new Integer(port);
		SocketReader sr;

		synchronized(this.sockets)
		{
			if( (sr = (SocketReader)this.sockets.
			     get(i)) == null )
			{
				java.lang.Thread srThread;
				
				// XXX wrong layer to create BufferHandle()
				sr = new SocketReader(port, BufferHandle.create());
				srThread = new java.lang.Thread(sr);
				srThread.setName("SocketReader for pj_UDPLayer "
						 + "port "
						 + port);
				srThread.start();
				this.sockets.put(i, sr);
			}
			sr.masks.addHandler(ph);
		}
	}

	PACKAGE void detachHandler(PacketHandler ph)
	{
		int port = ((pj_PortAddressMask)ph.source).destPort;
		Integer i = new Integer(port);
		SocketReader sr;

		try
		{
			synchronized(this.sockets)
			{
				sr = (SocketReader)this.sockets.get(i);
				if ((sr != null)
				    && (sr.masks.remHandler(ph)))
				{
					sr.dying = true;
					this.sockets.remove(i);

					// Send a packet to break last thread out of its loop
					try
					{
						this.dp.setAddress(sr.ds.
							getLocalAddress());
						this.dp.setPort(port);
						this.dp.setData(new byte[0]);
						this.dp.setLength(0);
						this.ds.send(this.dp);
					}
					catch (IOException e)
					{
					}
					
				}
			}
		}
		catch(Exception e)
		{
		}
	}

	PACKAGE void attachGenerator(PacketGenerator pg)
	{
		/* We don't need to track the generators */
	}

	/**
	 * Synchronous send a packet, blocking until the send
	 * completes.
	 */
	PACKAGE void synchronousSend(PacketGenerator pg,
				     pj_BufferHandle bh)
	{
		AddressSpecPortAddr ua;

		// XXX should clear out the asynchronous queue first!

		// XXX assert the reference count on bh is correct

		/* Get the destination and setup the datagram packet */
		ua = (AddressSpecPortAddr)pg.destination;

		synchronized(this.dp)
		{
			this.dp.setAddress(ua.destAddr());
			this.dp.setPort(ua.destPort());

			// XXX need a synchronized(pb) retrevial of byte[] and length.
			this.dp.setData(bh.getByteArray());
			this.dp.setLength(bh.getLength());

			try
			{
				this.ds.send(this.dp);
			}
			catch(IOException e)
			{
				DEBUGONLY(Debug.println("ioe in UDP layer: " +e.getMessage()));
			}
		}
	}
	
	/**
	 * Signal to UDP layer that asynchronous packets are
	 * available.  We just do a synchronous send.  :)
	 */

	PACKAGE void generatorReady(PacketGenerator pg)
	{
		pj_BufferHandle pb = pg.generate();
		synchronousSend(pg, pb);
	}

	PACKAGE void detachGenerator(PacketGenerator pg)
	{
		/* We don't need to track the generators */
	}
}
