/*
 * LinkedList.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_UTIL_PKG;

/**
 * A simple doubly linked list implementation, it exports the node object
 * directly so adding and removing is fast
 */
public class LinkedList
{
	/**
	 * A node for use in doubly linked lists.
	 */
	public static class Node
	{
		/**
		 * The next node in the list
		 */
		private Node succ;
		
		/**
		 * The previous node in the list
		 */
		private Node pred;
		
		/**
		 * Get the next node in the list
		 */
		public final Node getSucc()
		{
			return this.succ;
		}
		
		/**
		 * Get the previous node in the list
		 */
		public final Node getPred()
		{
			return this.pred;
		}
		
		/**
		 * Remove the node from the list
		 */
		public void remove()
		{
			this.pred.succ = this.succ;
			this.succ.pred = this.pred;
			this.succ = null;
			this.pred = null;
		}
		
		/**
		 * Insert a node after this one
		 */
		public void append(Node node)
		{
			this.succ.pred = node;
			node.succ = this.succ;
			node.pred = this;
			this.succ = node;
		}

		/**
		 * Insert a node before this one
		 */
		public void prepend(Node node)
		{
			this.pred.succ = node;
			node.succ = this;
			node.pred = this.pred;
			this.pred = node;
		}
	}

	/**
	 * Points to a dummy node which points to the head of the list
	 */
	private Node head;

	/**
	 * Points to a dummy node which points to the tail of the list
	 */
	private Node tail;

	/**
	 * Create a linked list
	 */
	public LinkedList()
	{
		this.head = new Node();
		this.tail = new Node();
		this.head.succ = this.tail;
		this.tail.pred = this.head;
	}

	/**
	 * Returns true if the list is empty
	 */
	public final boolean isEmpty()
	{
		boolean retval = false;

		if( this.head.succ == this.tail )
			retval = true;
		return( retval );
	}

	/**
	 * Get the head of the list
	 */
	public final Node getHead()
	{
		return this.head.succ;
	}

	/**
	 * Get the tail of the list
	 */
	public final Node getTail()
	{
		return this.tail.pred;
	}
	
	/**
	 * Add a node to the head of the list
	 */
	public void addHead( Node node )
	{
		node.succ = this.head.succ;
		node.pred = this.head;
		this.head.succ.pred = node;
		this.head.succ = node;
	}

	/**
	 * Add a node to the tail of the list
	 */
	public void addTail( Node node )
	{
		node.succ = this.tail;
		node.pred = this.tail.pred;
		this.tail.pred.succ = node;
		this.tail = node;
	}

	/**
	 * Remove the node at the head of the list, returns null if its empty
	 */
	public Node remHead()
	{
		Node retval = null;

		if( !this.isEmpty() )
		{
			retval = this.head.succ;
			retval.remove();
		}
		return( retval );
	}

	/**
	 * Remove the node at the tail of the list, returns null if its empty
	 */
	public Node remTail()
	{
		Node retval = null;

		if( !this.isEmpty() )
		{
			retval = this.tail.pred;
			retval.remove();
		}
		return( retval );
	}
}
