#
# JavaClass.pm
#
# Copyright (c) 1999, 2000 University of Utah and the Flux Group.
# All rights reserved.
#
# This file is part of Janos.
# 
# Janos is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License, version 2, as
# published by the Free Software Foundation.
# 
# Janos is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details, in the file
# COPYING.  If that file is not present, write to the Free
# Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
# 02111-1307, USA
# 
# The University of Utah and the Flux Group requests users to
# return to flux-dist@cs.utah.edu any improvements that they make.
# 

#
# Functions for reading in and writing out a Java .class file.
# Also does a bit of consistency checking of the file.
#
# The only really nasty thing I've done (because of poor perl skils more
# than anything else) is to make the %class a local() in a number of
# places so that the check routines can see it.
#
# Class structure:  Generally references to hashes. Tables are implemented as arrays.
#
# TODO: 
#	.make a &checkClass() function.
#	.Make CLASSIN and CLASSOUT parameters to read/write functions.
#	.POD documentation
#	.Cannot handle modifying float values.  I can read and decode, but don't
#	have the math to convert back to a binary format (both floats and doubles).
#	.Define attributes the same way bytecodes are defined (and get rid
#	of the almost-identical attribute reading/writing code.

package JavaClass;

use Carp;

###
### Define constants for Java Classes
###

*classMagic = \0xcafebabe;	# The magic header every .class file starts with

## The magic identifiers for entries in the .class Constant Table.
*CONSTANT_Class = \7;
*CONSTANT_FieldRef = \9;
*CONSTANT_MethodRef = \10;
*CONSTANT_InterfaceMethodRef = \11;
*CONSTANT_String = \8;
*CONSTANT_Integer = \3;
*CONSTANT_Float = \4;
*CONSTANT_Long = \5;
*CONSTANT_Double = \6;
*CONSTANT_NameAndType = \12;
*CONSTANT_Utf8 = \1;

## String names associated with each type of Constant Table entry.
%CONSTANTNames = (
    $CONSTANT_Class => "Class",
    $CONSTANT_FieldRef => "Field",
    $CONSTANT_MethodRef => "Method",
    $CONSTANT_InterfaceMethodRef => "Inteface Method",
    $CONSTANT_String => "String",
    $CONSTANT_Float => "Float",
    $CONSTANT_Integer => "Integer",
    $CONSTANT_Double => "Double",
    $CONSTANT_Long => "Long",
    $CONSTANT_NameAndType => "Name&Type",
    $CONSTANT_Utf8 => "Utf8"
    );


## String names for the shorthand used in signatures
$sig{'V'} = 'void';
$sig{'I'} = 'int';
$sig{'J'} = 'long';
$sig{'Z'} = 'boolean';
$sig{'F'} = 'float';
$sig{'D'} = 'double';
$sig{'B'} = 'byte';
$sig{'S'} = 'short';
$sig{'C'} = 'char';

## Access control flags for classes, methods and fields.
*ACC_PUBLIC    = \0x0001;
*ACC_PRIVATE   = \0x0002;
*ACC_PROTECTED = \0x0004;
*ACC_STATIC    = \0x0008;
*ACC_FINAL     = \0x0010;
*ACC_SUPER     = \0x0020;
*ACC_VOLATILE  = \0x0040;
*ACC_TRANSIENT = \0x0080;
*ACC_NATIVE    = \0x0100;
*ACC_INTERFACE = \0x0200;
*ACC_ABSTRACT  = \0x0400;

*ACC_UNKNOWN   = \0xF800;

###
###  Global variables
###

# Control the verbosity of &printClass()
$detailedFields = 0;
$detailedMethods = 0;
$showMethodCode = 1;
$printAsIRead = 1;

###
### Conversion functions
###

## parseJavaSig() takes a single argument, a single Java-internal
## method signature and returns a list ($package, $return, $class,
## $method, @args) where the items have been converted to a more
## source-like format (e.g., english).
sub parseJavaSig() {
  ## Parameters
  my $jsig = shift;

  ## Local variables
  my $class = '';
  my $package = '';
  my $method = '';
  my @args = ();
  my $ret = '';

  ## Temporaries
  my $depth = 0;
  my $repct = 0;
  my $arg = '';

  ### First is the class (all chars until a ".")
  $jsig =~ s/^([^.]*).//;
  $class = $1;
  $class =~ s,/,.,g; # / -> .

  # Peel the package name out of the class name (everything before last ".")
  if ($class =~ m/(.*)\.[^\.]*$/) {
    $package = $1;
  }

  ### Second comes the method name (all chars until a left paren)
  $jsig =~ s/^([^\(]*)\(//;
  $method = $1;

  ### Now the arguments
 SIGPARSE:
  while(1) {
    $repct = $jsig =~ s/^(I|J|Z|F|D|B|S|C|L|\[|\))//;  ## No V types
    die "badly formed signature at '$jsig'" if ($repct == 0);
    $arg = $1;
    
    ## Stop if we hit the end paren
    last SIGPARSE if $arg eq "\)";
    
    if ($arg eq '[') {
      $depth++;
      # continue parsing array type... 
      next SIGPARSE;
    } elsif ($arg eq 'L') {
      $jsig =~ s/^([^;]*);//;
      $arg = $1;
      $arg =~ s,/,.,g;
    } else {
      ## convert single-char identifier to english
      $arg = $sig{$arg};
    }
    
    ## If we hit an array, tack the array depth on the end
    if ($depth > 0) {
      $arg = $arg . "[]" x $depth;
      $depth = 0;
    }

    # Put the arg at the end of the list of args
    push (@args, $arg)
  }

  ### Last is the return type
  $depth = 0;
  $repct = $jsig =~ s/^(I|J|Z|F|D|B|S|C|L|V|\[)//; ## Adds V over argument types
  die "badly formed return type: \'$jsig\'" if ($repct == 0);
  $ret = $1;

  # If its an array, eat the [ and re-set $ret
  if ($ret eq '[') {
    $depth = 1;
    while ($jsig =~ s/\[//) {
      $depth++;
    }
    $jsig =~ s/^(I|J|Z|F|D|B|S|C|L)//; ## No [ or V
    die "badly formed return type: \'$jsig\'" if ($repct == 0);
    $ret = $1;
  } 

  if ($ret eq 'L') {
    $jsig =~ s/^([^;]*);//;
    $ret = $1;
    $ret =~ s,/,.,g;
  } else {
    ## Convert single char identifier to english
    $ret = $sig{"$ret"};
  }

  # Tack the array brackets on
  if ($depth > 0) {
    $ret = $ret . "[]" x $depth;
  }
    
  ### Return the info in an easy-to-use list
  return ($package, $ret, $class, $method, @args);
}

###
### Print functions
###

sub printClass {
    my $r_class = shift;		

    &printConstantPool($r_class);

    ## Print accessflags, class name, superclass.
    &printClassHeader($r_class);

    ## Print direct super interfaces
    &printInterfaces($r_class);

    ## Print fields
    &printFields($r_class);

    ## Print methods
    &printMethods($r_class);

    ## Print attributes
    &printAttributes("Class Attributes", "", $r_class, $r_class->{attributes});
}

sub printClassHeader {
    my $r_class = shift;

    ## Print access flags as a string
    my $flStr = &ACCFlagsToString($r_class->{accessFlags});
    print "$flStr\n";

    ## Print 'this_class'
    my $thisClName = %{$r_class->{constantPool}[$r_class->{thisClass}]}->{nameIndex};
    $thisClName = %{$r_class->{constantPool}[$thisClName]}->{val};
    print "this_class @ $r_class->{thisClass} ($thisClName)\n";

    ## Print 'super_class'
    if ($r_class->{superClass} != 0) {
      my $superClName = %{$r_class->{constantPool}[$r_class->{superClass}]}->{nameIndex};
      $superClName = %{$r_class->{constantPool}[$superClName]}->{val};
      print "super_class @ $r_class->{superClass} ($superClName)\n";
    } else {
      print "No super class\n";
    }
}

sub printMethods {
    my $r_class = shift;

    if ($r_class->{methodCt} == 0) {
	print "No methods.\n";
    } else {
	$i = 0;
	print "Methods:\n";
	while ($i < $r_class->{methodCt}) {
	    my %method = %{$r_class->{methods}[$i]};
    
	    my $accflags = ACCFlagsToString($method{accessFlags});
	    my $name = $r_class->{constantPool}[$method{nameIndex}]->{val};
	    my $desc = $r_class->{constantPool}[$method{descriptorIndex}]->{val};

	    if ($detailedMethods) {
		print "  [$i] ";
		print "$accflags $desc $name\n";
		printf "\t%s; %s; %s\n",
			"desc @ $method{descriptorIndex}",
			"name @ $method{nameIndex}",
			"attrct: $method{attributesCt}";
		&printAttributes("    Method Attributes", "\t", $r_class, $method{attributes});
	    } else {
		print "$accflags $name $desc\n";
	    }

	} continue {
	    $i++;
	}
    }
}

sub printFields {
    my $r_class = shift;

    if ($r_class->{fieldCt} == 0) {
	print "No fields.\n";
    } else {
	$i = 0;
	print "Fields:\n";
	while ($i < $r_class->{fieldCt}) {
	    my %field = %{$r_class->{fields}[$i]};
    
	    my $accflags = ACCFlagsToString($field{accessFlags});
	    my $name = $r_class->{constantPool}[$field{nameIndex}]->{val};
	    my $desc = $r_class->{constantPool}[$field{descriptorIndex}]->{val};

	    if ($detailedFields) {
		print "  [$i] ";
		print "$accflags $desc $name\n";
		printf "\t%s; %s; %s\n",
		       "desc @ $field{descriptorIndex}",
		       "name @ $field{nameIndex}",
		       "attrct: $field{attributesCt}";
		&printAttributes("    Field Attributes", "\t", $r_class, $field{attributes});
	    } else {
		print ("$accflags $desc $name\n");
	    }
	} continue {
	    $i++;
	}
    }
}

sub printInterfaces {
    my $r_class = shift;

    if ($r_class->{interfaceCt} == 0) {
	print "No interfaces.\n";
    } else {
	my $i = 0; 
	print "Interfaces:\n";
	while ($i < $r_class->{interfaceCt}) {
	    my $iClass = $r_class->{interfaces}[$i];
	    my %iClassConst = %{$r_class->{constantPool}[$iClass]};
	    my $iClassName = $iClassConst{nameIndex};
	    my $interfaceName = %{$r_class->{constantPool}[$iClassName]}->{val};
	    print "\t$i] @ $iClass ($interfaceName)\n";
	} continue {
	    $i++;
	}
    }
}

sub ACCFlagsToString {
    my $flags = shift;
    my @flags = ();

    push(@flags, "public") if $flags & $ACC_PUBLIC;
    push(@flags, "private") if $flags & $ACC_PRIVATE;
    push(@flags, "protected") if $flags & $ACC_PROTECTED;
    push(@flags, "static") if $flags & $ACC_STATIC;
    push(@flags, "final") if $flags & $ACC_FINAL;
    push(@flags, "super") if $flags & $ACC_SUPER;
    push(@flags, "volatile") if $flags & $ACC_VOLATILE;
    push(@flags, "transient") if $flags & $ACC_TRANSIENT;
    push(@flags, "interface") if $flags & $ACC_INTERFACE;
    push(@flags, "abstract") if $flags & $ACC_ABSTRACT;
    push(@flags, "native") if $flags & $ACC_NATIVE;
    push(@flags, "UNKNOWN") if $flags & $ACC_UNKNOWN;

    return join(',', @flags);
}

sub printConstantPool {
    my $r_class = shift;

    print("Constant Pool Entries: $r_class->{constantPoolCt}\n");

    $i = 1;
    while ($i < $r_class->{constantPoolCt}) {
	my %cpEntry = %{$r_class->{constantPool}[$i]};

	print "$i] $CONSTANTNames{$cpEntry{tag}}: ";
	$str = constantPoolEntryToStr($r_class, $i);
	print ("$str\n");
	
	if (($cpEntry{tag} eq $CONSTANT_Double)
	    || ($cpEntry{tag} eq $CONSTANT_Long)) {

	    ## Ick.  8-byte entries take two constant pool entries
	    $i++; 
	}

	$i++;
    }
}


sub constantPoolEntryToStr {
    my ($r_class, $cpEntryIndex) = @_;

    my $verboseStyle = 0;

    return "" if ($cpEntryIndex == 0);

    my $r_cpEntry = $r_class->{constantPool}[$cpEntryIndex];

    if ($r_cpEntry->{tag} eq $CONSTANT_Class) {
	my $ni = $r_cpEntry->{nameIndex};
	&checkIndex($r_class, $ni, "Name", $CONSTANT_Utf8);

	my $nm = &constantPoolEntryToStr($r_class, $ni);
	
	if ($verboseStyle) {
	  return ".name @ $ni ($nm);";
	} else {
	  return "$nm";
	}
	
    } elsif (($r_cpEntry->{tag} eq $CONSTANT_FieldRef)
	     || ($r_cpEntry->{tag} eq $CONSTANT_MethodRef)
	     || ($r_cpEntry->{tag} eq $CONSTANT_InterfaceMethodRef)) {
	&checkIndex($r_class, $r_cpEntry->{classIndex}, "Class", $CONSTANT_Class);
	&checkIndex($r_class, $r_cpEntry->{nameTypeIndex}, "Name & Type", $CONSTANT_NameAndType);
	
	my $className = &constantPoolEntryToStr($r_class, $r_cpEntry->{classIndex});
	my $slotName = &constantPoolEntryToStr($r_class, $r_cpEntry->{nameTypeIndex});

	if ($verboseStyle) {
	  return ".class @ $r_cpEntry->{classIndex}; .name&type @ $r_cpEntry->{nameTypeIndex};";
	} else {
	  return "$className.$slotName";
	}

    } elsif ($r_cpEntry->{tag} eq $CONSTANT_String) {
	my $si = $r_cpEntry->{stringIndex};
	&checkIndex($r_class, $si, "String", $CONSTANT_Utf8);
	my $str = $r_class->{constantPool}[$si]->{val};
	
	if ($verboseStyle) {
	  return ".string @ $r_cpEntry->{stringIndex} ($str);";
	} else {
	  return "\"$str\"";
	}

    } elsif ($r_cpEntry->{tag} eq $CONSTANT_NameAndType) {
	my $ni = $r_cpEntry->{nameIndex};
	my $di = $r_cpEntry->{descriptorIndex};
	
	&checkIndex($r_class, $ni, "Name", $CONSTANT_Utf8);
	&checkIndex($r_class, $di, "Descriptor", $CONSTANT_Utf8);
	
	#my $nstr = $r_class->{constantPool}[$ni]->{val};
	#my $dstr = $r_class->{constantPool}[$di]->{val};
	my $nstr = &constantPoolEntryToStr($r_class, $ni);
	my $dstr = &constantPoolEntryToStr($r_class, $di);
	
	if ($verboseStyle) {
	  return ".name @ $ni ($nstr); .descriptor @ $di ($dstr); ";
	} else {
	  return "${nstr}::${dstr}";
	}

    } elsif ($r_cpEntry->{tag} eq $CONSTANT_Integer) {
      if ($verboseStyle) {
	return ".value = $r_cpEntry->{val}";
      } else {
	return "$r_cpEntry->{val}";
      }

    } elsif ($r_cpEntry->{tag} eq $CONSTANT_Utf8) {

      if ($verboseStyle) {
	return ".length=$r_cpEntry->{len}; " .
	  ".val=$r_cpEntry->{val}; ";
      } else {
	return "$r_cpEntry->{val}";
      }

    } elsif ($r_cpEntry->{tag} eq $CONSTANT_Float) {

      if ($verboseStyle) {
	return ".val=$r_cpEntry->{strVal}; ";
      } else {
	return "$r_cpEntry->{strVal}";
      }

    } elsif ($r_cpEntry->{tag} eq $CONSTANT_Double) {
      if ($verboseStyle) {
	return ".val=$r_cpEntry->{strVal}; ";
      } else {
	return "$r_cpEntry->{strVal}";
      }

    } elsif ($r_cpEntry->{tag} eq $CONSTANT_Long) {
      if ($verboseStyle) {
	return ".val=$r_cpEntry->{strVal}; ";
      } else {
	return "$r_cpEntry->{strVal}";
      }
    } else {
	&internalError("Unknown Constant type $r_cpEntry->{tag}!\n");
    }
}

sub printAttributes {
    my ($label, $prefix, $r_class, $r_attrs) = @_;

    return if (!defined($r_attrs));

    my $i = 0;
    
    print ("$label:\n");
    foreach $r_attr (@{$r_attrs}) {
	&printAttribute($prefix, $r_class, $r_attr);
    }
}

sub printAttribute {
    my ($prefix, $r_class, $r_attr) = @_;

    my $name = $r_attr->{name};

    ## Handle some of the attrs special-like
    if ($name eq 'Code') {
	print ("${prefix}$name:\n");
	&printCodeAttribute("${prefix}\t", $r_class, $r_attr);

    } elsif ($name eq 'Synthetic') {
	print ("${prefix}$name.\n");

    } elsif ($name eq 'SourceFile') {
	my $sfname = $r_class->{constantPool}[$r_attr->{sourceFileIndex}]->{val};
	print "${prefix}\t$name: $sfname\n";

    } elsif ($name eq 'InnerClasses') {
	print ("${prefix}\t$name: ($r_attr->{innerClassesCount})\n");
	my $i = 0;
	while ($i < $r_attr->{innerClassesCount}) {
	    my %ic = %{$r_attr->{innerClasses}[$i]};
	    printf "${prefix}\t %3d] inner: <Class %s>\n", $i, &constantPoolEntryToStr($r_class, $ic{innerClassIdx});
	    printf "${prefix}\t      outer: <Class %s>\n", &constantPoolEntryToStr($r_class, $ic{outerClassIdx});
	    printf "${prefix}\t      inner name: <Utf8: %s>\n", &constantPoolEntryToStr($r_class, $ic{innerNameIdx});
	    printf "${prefix}\t      access flags: %s\n", &ACCFlagsToString($ic{innerClassAccessFlags});
	    $i++;
	}

    } else {
	my $key;
	print ("${prefix}\t$name:\n");
	foreach $key (sort(keys %{$r_attr})) {
	    if (($key ne "name") 
		&& ($key ne "nameIndex")) {
		print "${prefix}\t\t$key: $r_attr->{$key}\n";
	    }
	}
    }
}

## print the Code attribute.
## arguments are (string prefix, ref to the class, code attr)
sub printCodeAttribute {
    my ($prefix, $r_class, $r_attr) = @_;
    
    return if (!defined($r_attr));
    
    print "${prefix}maxStack: $r_attr->{maxStack}\n";
    print "${prefix}maxLocals: $r_attr->{maxLocals}\n";
    print "${prefix}codeLength: $r_attr->{codeLength}\n";
    
    &printBytecode("${prefix}", $r_class, $r_attr->{code}, $r_attr->{codeLength});
    
    print "${prefix}exceptionTableLen:  $r_attr->{exceptionTableLen}\n";
    my $i = 0;
    while( $i < $r_attr->{exceptionTableLen} ) {
	my %ent = %{$r_attr->{exceptionTable}[$i]};
	my $extype = '';
	
	if ($ent{catchtypeidx} == 0) {
	    $extype = "<any>";
	} else {
	    my $extypeidx = %{$r_class->{constantPool}[$ent{catchtypeidx}]}->{nameIndex};
	    $extype = %{$r_class->{constantPool}[$extypeidx]}->{val};
	}
	printf "${prefix}\t[$ent{startpc} - $ent{endpc}) -> $ent{handlerpc} ($extype)\n";
	$i++;
    }
    
    &printAttributes("${prefix}", $r_class, $r_attr->{codeAttrs});
}

sub printBytecode() {
    my ($prefix, $r_class, $r_code, $codelen) = @_;
    my $offset = 0;
    
    &internalError("code block is undefined") if (!defined($r_code));

    while ($offset < $codelen) {
	&internalError("expected instr at offset $offset is missing ") if (!defined($r_code->{$offset}));

	my %instr = %{$r_code->{$offset}};
	my $opcode = $instr{opcode};
	my $ilen = $BytecodeInfo{$opcode}->{length};
	my $iname = $BytecodeInfo{$opcode}->{name};
	
	printf "${prefix}%4d] %s\t", $offset, $iname;
	
	### Variable length instructions have length stored in them
	if ($ilen == 0) {
	    &internalError("Instruction at $offset (op=$opcode) has no length defined.\n")
	      if (!defined($instr{length}));
	    $ilen = $instr{length};
	}
	
	foreach $key (keys(%instr)) {
	    if ($key eq 'index') {
		printf "%s", &constantPoolEntryToStr($r_class, $instr{$key});
	    } elsif ($key eq 'branch') {
		$reloff = $instr{$key};
		# XXX EWWWW.  Converting to "signed":
		$reloff = $reloff - 0x10000 if ($reloff >= 0x8000);
		$absip = $reloff + $offset;
		printf "(%+d -> %#u)", $reloff, $absip;
	    } elsif ($key eq 'const') {
		printf "(%#x)", $instr{$key};
	    } elsif ($key eq 'sindex') {
		printf "(lvar %#x)", $instr{$key};
	    } elsif (($key ne 'opcode')
	      && ($key ne 'length')
		&& ($key ne 'name')) {
		print "$key: $instr{$key}"
	    }
	}

	print "\n";
	
	$offset += $ilen;
    }
}

sub readClassHeader {
    my $r_class = shift;

    $r_class->{accessFlags} = &read_u2();
    
    $r_class->{thisClass} = &read_u2();
    &checkIndex($r_class, $r_class->{thisClass}, 
		"this_class", $CONSTANT_Class);
    
    $r_class->{superClass} = &read_u2();
    if ($r_class->{superClass} != 0) {
	&checkIndex($r_class, $r_class->{superClass}, 
		    "super_class", $CONSTANT_Class);
    } else {
	print ("Warning: class has no super class.  Must be java.lang.Object.\n"); 
    }
}

###
### Read Class function
###

sub readClass {
    my $classFile = shift;
    local(%class) = (());

    if ($classFile =~ /\.class$/) {
	open(CLASSIN, $classFile) 
	  || die ("Cannot open $classFile for reading.");
    } else {
	open(CLASSIN, "${classFile}.class")
	  || die ("Cannot open ${classFile}.class for reading.");
    }
    
    ###
    ### Header Magic
    ###
    	
    $class{magic} = read_u4();
    if ($class{magic} != $classMagic) {
    	fatal("Bad class magic '$class{magic}' --expected '$classMagic'.  " .
	      "$classFile is probably not a Java class file.");
    }
    
    ## Read in the major and minor version numbers
    $class{minorVersion} = &read_u2();
    $class{majorVersion} = &read_u2();
    
    #print("Version: $class{majorVersion}.$class{minorVersion}  (expected 45.3)\n") 
    #	if ($class{minorVersion} ne 3) || ($class{majorVersion} ne 45);
    
    ###
    ### Constant Pool
    ###
    $class{constantPoolCt} = &read_u2();
    $class{constantPool} = [];
    
    $i = 1; # constant pool actually starts with entry 1...
    while ($i < $class{constantPoolCt}) {
    	my %cpEntry;
    	$cpEntry{tag} = &read_u1();
    
    	if ($cpEntry{tag} eq $CONSTANT_Class) {
    	    $cpEntry{nameIndex} = &read_u2();
    
    	    &checkIndex(\%class, $cpEntry{nameIndex}, "Name");
    	} elsif ($cpEntry{tag} eq $CONSTANT_FieldRef) {
    	    $cpEntry{classIndex} = &read_u2();
    	    $cpEntry{nameTypeIndex} = &read_u2();
    	    
    	    &checkIndex(\%class, $cpEntry{classIndex}, "Class");
    	    &checkIndex(\%class, $cpEntry{nameTypeIndex}, "Name & Type");
    	} elsif ($cpEntry{tag} eq $CONSTANT_MethodRef) {
    	    $cpEntry{classIndex} = &read_u2();
    	    $cpEntry{nameTypeIndex} = &read_u2();
    
    	    &checkIndex(\%class, $cpEntry{classIndex}, "Class");
    	    &checkIndex(\%class, $cpEntry{nameTypeIndex}, "Name & Type");
    	} elsif ($cpEntry{tag} eq $CONSTANT_InterfaceMethodRef) {
    	    $cpEntry{classIndex} = &read_u2();
    	    $cpEntry{nameTypeIndex} = &read_u2();
    
    	    &checkIndex(\%class, $cpEntry{classIndex}, "Class");
    	    &checkIndex(\%class, $cpEntry{nameTypeIndex}, "Name & Type");
    	} elsif ($cpEntry{tag} eq $CONSTANT_String) {
    	    $cpEntry{stringIndex} = &read_u2();
    
    	    &checkIndex(\%class, $cpEntry{stringIndex}, "String");
    	} elsif ($cpEntry{tag} eq $CONSTANT_NameAndType) {
    	    $cpEntry{nameIndex} = &read_u2();
    	    $cpEntry{descriptorIndex} = &read_u2();
    
    	    &checkIndex(\%class, $cpEntry{nameIndex}, "Name");
    	    &checkIndex(\%class, $cpEntry{descriptorIndex}, "Descriptor");
    	} elsif ($cpEntry{tag} eq $CONSTANT_Integer) {
    	    $cpEntry{val} = &read_u4();
    	} elsif ($cpEntry{tag} eq $CONSTANT_Utf8) {
    	    $cpEntry{len} = &read_u2();
    	    $cpEntry{val} = &read_utf8($cpEntry{len});
    	} elsif ($cpEntry{tag} eq $CONSTANT_Float) {
	    $cpEntry{val} = &read_u4();
	    $cpEntry{strVal} = &parse_float($cpEntry{val});
    	} elsif ($cpEntry{tag} eq $CONSTANT_Double) {
	    $cpEntry{val} = &read_u8();
	    $cpEntry{strVal} = &parse_double($cpEntry{val});
    	} elsif ($cpEntry{tag} eq $CONSTANT_Long) {
	    $cpEntry{val} = &read_u8();
	    $cpEntry{strVal} = &parse_long($cpEntry{val});
    	} else {
    	    &internalError("Unknown Constant type $cpEntry{tag}!\n");
    	}
    
    	$class{constantPool}[$i] = \%cpEntry;

	## Ick.  8-byte entries take two constant pool entries
	$i++ if (($cpEntry{tag} == $CONSTANT_Long)
		 || ($cpEntry{tag} == $CONSTANT_Double));
    } continue {
    	$i++;
    }
    
    #&printConstantPool(\%class) if ($printAsIRead);

    ###
    ### Misc. Class Info
    ###

    &readClassHeader(\%class);
    #&printClassHeader(\%class) if ($printAsIRead);
    
    ###
    ### Direct super-interfaces
    ###
    $class{interfaceCt} = &read_u2();
    $class{interfaces} = [];
    
    $i = 0;
    while ($i < $class{interfaceCt}) {
    	$class{interfaces}[$i] = &read_u2();
    
    	&checkIndex(\%class, $class{interfaces}[$i], "Interface \#$i", $CONSTANT_Class);	
    } continue {
    	$i++;
    }

    #&printInterfaces(\%class) if ($printAsIRead);
    
    ###
    ### Fields
    ###
    $class{fieldCt} = &read_u2();
    $class{fields} = [];
    
    $i = 0;
    while ($i < $class{fieldCt}) {
    	my %field;
    	$field{accessFlags} = &read_u2();
    	$field{nameIndex} = &read_u2();
    	$field{descriptorIndex} = &read_u2();
    	$field{attributesCt} = &read_u2();
    	$field{attributes} = &readAttributes("field", \%class, $field{attributesCt});
    
    	&checkIndex(\%class, $field{nameIndex}, "Field Name", $CONSTANT_Utf8);
    	&checkIndex(\%class, $field{descriptorIndex}, "Field Descriptor", $CONSTANT_Utf8);
    
    	$class{fields}[$i] = \%field;
    } continue {
    	$i++;
    }

    #&printFields(\%class) if ($printAsIRead);

    ###
    ### Methods
    ###
    $class{methodCt} = &read_u2();
    $class{methods} = [];
    
    $i = 0;
    while ($i < $class{methodCt}) {
    	my %method;
    
    	$method{accessFlags} = &read_u2();
    	$method{nameIndex} = &read_u2();
    	$method{descriptorIndex} = &read_u2();
    	$method{attributesCt} = &read_u2();
    	$method{attributes} = &readAttributes("method $i", \%class, $method{attributesCt});
    
    	&checkIndex(\%class, $method{nameIndex}, "Method Name", $CONSTANT_Utf8);
    	&checkIndex(\%class, $method{descriptorIndex}, "Method Descriptor", $CONSTANT_Utf8);

    	$class{methods}[$i] = \%method;
    } continue {
    	$i++;
    }

    #&printMethods(\%class) if ($printAsIRead);

    ###
    ### Class attributes
    ###
    $class{attributesCt} = &read_u2();
    $class{attributes} = &readAttributes("class", \%class, $class{attributesCt});

    #&printAttributes("Class Attributes", "", \%class, $class{attributes}) 
    #  if ($printAsIRead);

    ###
    ### End of .class file
    ###

    return \%class;
}

###
### Write Class function
###

sub writeClass {
    my $r_class = shift;
    my $classFile = shift;

    if ($classFile =~ /\.class$/) {
	open(CLASSOUT, ">$classFile") 
	    || die ("Cannot open $classFile for writing");
    } else {
	open(CLASSOUT, ">$classFile.class") 
	    || die ("Cannot open $classFile.class for writing");
    }

    ###
    ### Header Magic
    ###

    if ($r_class->{magic} != $classMagic) {
	close(CLASSOUT);
    	&fatal("Bad class magic '$r_class->{magic}' --expected '$classMagic'.  Not writing class file.");
    }
    	
    &write_u4($r_class->{magic});
    
    ## Write major/minor version numbers
    &write_u2($r_class->{minorVersion});
    &write_u2($r_class->{majorVersion});
    
    ###
    ### Constant Pool
    ###
    &write_u2($r_class->{constantPoolCt});

    $i = 1; # constant pool actually starts with entry 1...
    while ($i < $r_class->{constantPoolCt}) {
    	my %cpEntry = %{$r_class->{constantPool}[$i]};
    	&write_u1($cpEntry{tag});
    
    	if ($cpEntry{tag} eq $CONSTANT_Class) {
    	    &checkIndex($r_class, $cpEntry{nameIndex}, "Name", $CONSTANT_Utf8);

    	    &write_u2($cpEntry{nameIndex});
    	} elsif (($cpEntry{tag} eq $CONSTANT_FieldRef)
		 || ($cpEntry{tag} eq $CONSTANT_MethodRef)
		 || ($cpEntry{tag} eq $CONSTANT_InterfaceMethodRef)) {
    	    &checkIndex($r_class, $cpEntry{classIndex}, "Class", $CONSTANT_Class);
    	    &checkIndex($r_class, $cpEntry{nameTypeIndex}, "Name & Type", $CONSTANT_NameAndType);

    	    &write_u2($cpEntry{classIndex});
    	    &write_u2($cpEntry{nameTypeIndex});
    	} elsif ($cpEntry{tag} eq $CONSTANT_String) {
    	    &checkIndex($r_class, $cpEntry{stringIndex}, "String", $CONSTANT_Utf8);

    	    &write_u2($cpEntry{stringIndex});
    	} elsif ($cpEntry{tag} eq $CONSTANT_NameAndType) {
    	    &checkIndex($r_class, $cpEntry{nameIndex}, "Name", $CONSTANT_Utf8);
    	    &checkIndex($r_class, $cpEntry{descriptorIndex}, "Descriptor", $CONSTANT_Utf8);

    	    &write_u2($cpEntry{nameIndex});
    	    &write_u2($cpEntry{descriptorIndex});
    	} elsif ($cpEntry{tag} eq $CONSTANT_Integer) {
    	    &write_u4($cpEntry{val});
    	} elsif ($cpEntry{tag} eq $CONSTANT_Utf8) {
    	    &write_u2($cpEntry{len});
    	    &write_utf8($cpEntry{val});
    	} elsif ($cpEntry{tag} eq $CONSTANT_Float) {
	    &write_u4($cpEntry{val});
    	} elsif ($cpEntry{tag} eq $CONSTANT_Double) {
	    &write_u8($cpEntry{val});
	    $i++; ## Ick.  8-byte entries take two constant pool entries
    	} elsif ($cpEntry{tag} eq $CONSTANT_Long) {
	    &write_u8($cpEntry{val});
	    $i++; ## Ick.  8-byte entries take two constant pool entries
    	} else {
    	    &internalError("Unknown Constant type $cpEntry{tag}!\n");
    	}
    } continue {
    	$i++;
    }
    
    ###
    ### Misc. Class Info
    ###
    
    &write_u2($r_class->{accessFlags});
    &write_u2($r_class->{thisClass});
    &write_u2($r_class->{superClass});
    
    ###
    ### Direct super-interfaces
    ###
    &write_u2($r_class->{interfaceCt});
    
    $i = 0;
    while ($i < $r_class->{interfaceCt}) {
    	&write_u2($r_class->{interfaces}[$i]);
    } continue {
    	$i++;
    }
    
    ###
    ### Fields
    ###
    &write_u2($r_class->{fieldCt});
    
    $i = 0;
    while ($i < $r_class->{fieldCt}) {
    	my %field = %{$r_class->{fields}[$i]};
    	&write_u2($field{accessFlags});
    	&write_u2($field{nameIndex});
    	&write_u2($field{descriptorIndex});
    	&write_u2($field{attributesCt});
    	&writeAttributes($field{attributes});
    } continue {
    	$i++;
    }
    
    ###
    ### Methods
    ###
    &write_u2($r_class->{methodCt});
    
    $i = 0;
    while ($i < $r_class->{methodCt}) {
    	my %method = %{$r_class->{methods}[$i]};
    
    	&write_u2($method{accessFlags});
    	&write_u2($method{nameIndex});
	&write_u2($method{descriptorIndex});
    	&write_u2($method{attributesCt});
    	&writeAttributes($method{attributes});
    } continue {
    	$i++;
    }
    
    ###
    ### Class attributes
    ###
    &write_u2($r_class->{attributesCt});
    &writeAttributes($r_class->{attributes});

    ###
    ### End of .class file
    ###
}

###
### Integrity check functions
###

sub assert {
  my ($bool, $msg) = @_;
  
  if (! $bool) {
    &fatal("ERROR: assert failed: $msg.");
  }
}

sub checkIndex {
    my ($r_class, $val, $name, $type) = @_;

    if ((!defined($r_class))
	|| (!defined($r_class->{constantPool}))) {
	&internalError("\$r_class is broken in checkIndex()!");
    }

    if ($val == 0) {
      &fatal("ERROR: Constant pool index is 0 for $name.  ".
	     "(Expecting a valid CONSTANT_$CONSTANTNames{$type} entry.)");
    }
     
    if ($val >= $r_class->{constantPoolCt}) {
	&fatal("ERROR: $name index for current constant is $val, " .
	       "must be less than $r_class->{constantPoolCt}\n");
    } 

    if (defined($type)) {
	my $actualTag = $r_class->{constantPool}[$val]{tag};
	if ($actualTag != $type) {
	    &fatal("ERROR: $name expects a CONSTANT_$CONSTANTNames{$type} " .
		   "entry at $val, but found a " .
		   "CONSTANT_$CONSTANTNames{$actualTag} entry\n");
	}
    }
}

###
### Read primitives
###

sub read_u8 {
    my $long = 0;

    (read(CLASSIN, $long, 8) == 8) || confess("premature eof in read_u8()\n");

    # my ($b1, $b2, $b3, $b4, $b5, $b6, $b7, $b8) = unpack("CCCCCCCC", $long);
    # return (($b1 << 56) + ($b2 << 48) + ($b3 << 40) + ($b4 << 32)
    #	    + ($b5 << 24) + ($b6 << 16) + ($b7 << 8) + $b4);

    return join(".", unpack("CCCCCCCC", $long));
}

sub read_u4 {
    my $long = 0;
    (read(CLASSIN, $long, 4) == 4) || confess("premature eof in read_u4()\n");
    return unpack("N", $long);
}

sub read_u2 {
    my $short = 0;
    (read(CLASSIN, $short, 2) == 2) || confess("premature eof in read_u2()\n");
    return unpack('n', $short);
}

sub read_u1 {
    my $byte = 0;
    (read(CLASSIN, $byte, 1) == 1) || confess("premature eof in read_u1()\n");
    return unpack("C", $byte);
}

sub read_n {
    my $byteCt = shift;
    my $foo = '';
    (read(CLASSIN, $foo, $byteCt) == $byteCt) || confess("premature eof in read_n($byteCt)\n");

    return $foo;
}

sub parse_long {
  my $silly = shift;
  my @u8_bits = split(/\./, $silly);

  my $hi = ($u8_bits[0] << 24) + ($u8_bits[1] << 16) + ($u8_bits[2] << 8) + $u8_bits[3];
  my $low = ($u8_bits[4] << 24) + ($u8_bits[5] << 16) + ($u8_bits[6] << 8) + $u8_bits[7];
   
  return sprintf("0x%lx", $low) if ($hi == 0);

  return "0x" . sprintf("%lx", $hi) . sprintf("%8.8lx", $low) 
}

sub parse_float {
    my $intVal = shift;
    
    return "+INF" if ($intVal == 0x7f800000);
    return "-INF" if ($intVal == 0xff800000);
    if ((($intVal >= 0x7f800001) && ($intVal <= 0x7fffffff))
	|| (($intVal >= 0xff800001) && ($intVal <= 0xffffffff))) {
	return "NaN";
    }

    ## Otherwise, convert to a floating point number
    $sign     = (($intVal >> 31) == 0) ?  1 : -1;
    $exponent = (($intVal >> 23) & 0xFF);
    $mantissa = ($exponent == 0) ? ($intVal & 0x7fffff) << 1 : ($intVal & 0x7fffff) | 0x800000;

    return $sign * $mantissa * 2 ** ($exponent - 150);
}

sub parse_double {
    my $intVal = shift;

    if (0) {
      return "+INF" if ($intVal == 0x7f800000);
      return "-INF" if ($intVal == 0xff800000);
      if ((($intVal >= 0x7f800001) && ($intVal <= 0x7fffffff))
	  || (($intVal >= 0xff800001) && ($intVal <= 0xffffffff))) {
	return "NaN";
      }
      
      ## Otherwise, convert to a floating point number
      $sign     = (($intVal >> 31) == 0) ?  1 : -1;
      $exponent = (($intVal >> 23) & 0xFF);
      $mantissa = ($exponent == 0) ? ($intVal & 0x7fffff) << 1 : ($intVal & 0x7fffff) | 0x800000;
      
      return $sign * $mantissa * 2 ** ($exponent - 150);
    }
    return "<fix me>"
}

sub read_utf8 {
    my $byteCt = shift;
    my $utf = '';
    (read(CLASSIN, $utf, $byteCt) == $byteCt) || confess("premature eof in read_utf8($byteCt)\n");

    my $str = unpack("A$byteCt", $utf);

    return $str;
}

## readAttributes(class ref, number of attributes) -> attr array reference
sub readAttributes {
    my ($where, $r_class, $ct) = @_;
    
    my @attrs = [];

    return undef if ($ct < 1);

    my $i = 0;
    while ($i < $ct) {
	my %attribute = ();
	$attribute{nameIndex} = &read_u2();
	$attribute{len} = &read_u4();
	
	&checkIndex($r_class, $attribute{nameIndex}, "name of $where attribute $i", $CONSTANT_Utf8);
	$attribute{name} = %{$r_class->{constantPool}[$attribute{nameIndex}]}->{val};

	## Read the various kinds of attributes that are understood
	if ($attribute{name} eq "Code") {
	    readCodeAttribute("$where code", $r_class, \%attribute);
	} elsif ($attribute{name} eq "SourceFile") {
	    readSourceFileAttribute($r_class, \%attribute);
	} elsif ($attribute{name} eq "ConstantValue") {
	    readConstantValueAttribute($r_class, \%attribute);
	} elsif ($attribute{name} eq "Exceptions") {
	    readExceptionsAttribute($r_class, \%attribute);
	} elsif ($attribute{name} eq "LocalVariableTable") {
	    readLocalVariableTableAttribute($r_class, \%attribute);
	} elsif ($attribute{name} eq "InnerClasses") {
	    readInnerClassesAttribute($r_class, \%attribute);
	} elsif ($attribute{name} eq "Synthetic") {
	    readSyntheticAttribute($r_class, \%attribute);
	} elsif ($attribute{name} eq "Deprecated") {
	    readDeprecatedAttribute($r_class, \%attribute);
	} elsif ($attribute{name} eq "LineNumberTable") {
	    readLineNumberTableAttribute($r_class, \%attribute);
	} else {
	    readUnknownAttribute($r_class, \%attribute);
	}

	$attrs[$i] = \%attribute;
    } continue {
	$i++;
    }

    return \@attrs;
}

###
### Attribute reading functions
###

sub readCodeAttribute {
    my ($where, $r_class, $r_attr) = @_;
    
    $r_attr->{maxStack} = &read_u2();
    $r_attr->{maxLocals} = &read_u2();
    $r_attr->{codeLength} = &read_u4();
    $r_attr->{code} = &readBytecode($r_class, $r_attr->{codeLength});
    $r_attr->{exceptionTableLen} = &read_u2(); 
    
    $r_attr->{exceptionTable} = [];
    my $i = 0;
    while ($i < $r_attr->{exceptionTableLen}) {
	my %excTableEntry;
	
	$excTableEntry{startpc} = &read_u2();
	if ($excTableEntry{startpc} > $r_attr->{codeLength}) { 
	    &fatal("ERROR: $where malformed exception table entry " .
		   "(startpc=$excTableEntry{startpc}, code is $r_attr->{codeLength})."); 
	}
	
	$excTableEntry{endpc} = &read_u2();
	if ($excTableEntry{endpc} > $r_attr->{codeLength}) { 
	    &fatal("ERROR: malformed exception table entry (endpc)."); 
	}
	
	$excTableEntry{handlerpc} = &read_u2();
	if ($excTableEntry{handlerpc} > $r_attr->{codeLength}) { 
	    &fatal("ERROR: malformed exception table entry (handlerpc)."); 
	}
	
	$excTableEntry{catchtypeidx} = &read_u2();
	## A zero index is used for 'finally' clauses.
	if ($excTableEntry{catchtypeidx} != 0) {
	    &checkIndex($r_class, $excTableEntry{catchtypeidx}, 
			"Exception Class", $CONSTANT_Class);
	}
	
	$r_attr->{exceptionTable}[$i] = \%excTableEntry;
	$i++;
    }
    
    $r_attr->{codeAttrCount} = &read_u2();
    $r_attr->{codeAttrs} = &readAttributes("$where", $r_class, $r_attr->{codeAttrCount});
}

sub readBytecode {
    my ($r_class, $codelen) = @_;
    my %codeblock = ();
    my $offset = 0;

    while ( $offset < $codelen ) {
	my %instr = ();
	my $ilen;

	$instr{opcode} = &read_u1();
	$ilen = $BytecodeInfo{$instr{opcode}}->{length};

	&internalError("Unknown opcode $instr{opcode}.") if (!defined($BytecodeInfo{$instr{opcode}}));
	&internalError("Opcode $instr{opcode} is messed up.") if (!defined($ilen));

	#print "reading code $BytecodeInfo{$instr{opcode}}->{name} ($ilen bytes)\n";

	if ($ilen < 1) {
	    ## A variable length bytecode instruction.  Three options:
	    ## lookupswitch, tableswitch, or wide.

	    ## tableswitch is variable length
	    if ($instr{opcode} == 170) {
		$instr{padsize} = 3 - (($offset) % 4);
		$instr{padbytes} = &read_n($instr{padsize});
		$instr{default} = &read_u4();
		$instr{low} = &read_u4();
		$instr{high} = &read_u4();
		$instr{jumptablesize} = ($instr{high} - $instr{low} + 1) * 4;
		$instr{jumptable} = [];
		my $i = 0;
		# printf ("off: %d -> pad: %d, (%#08x,%#08x,%#08x) %d,  %d\n", 
		# 	    $offset, $instr{padsize},
		# 	    $instr{default},$instr{low},$instr{high},,
		# 	    $instr{jumptablesize});
		while ($i < $instr{jumptablesize}) {
		    $instr{jumptable}[$i] = &read_u4();
		    $i += 4;  # jumptablesize is in bytes
		}
		## Size is 1 (opcode) + padsize + default/high/low + jumptable
		$ilen = 1 + $instr{padsize} + 12 + $instr{jumptablesize};

	    } else {
		&internalError("Need to write instr parser for op=$instr{opcode}");
	    }

	    $instr{length} = $ilen;
	} elsif ($ilen > 1) {
	    my $argfmt = $BytecodeInfo{$instr{opcode}}->{argfmt};

	    &internalError("No argfmt defined for op=$instr{opcode}") 
	      if (!defined($argfmt));

	    my $bytesread = 1; # 1 for the opcode
	    foreach $arg (split(/:/, $argfmt)) {
		my ($name, $size) = split(/\//, $arg);
		$size = 1 if (!defined($size));
		if ($size == 1) {
		    $instr{$name} = &read_u1();
		} elsif ($size == 2) {
		    $instr{$name} = &read_u2();
		} elsif ($size == 4) {
		    $instr{$name} = &read_u4();
		} else {
		    &internalError("Bad size ($size) in format string $argfmt.");
		}
		$bytesread += $size;
	    }
	    &assert($bytesread == $ilen, "Supposed to read $ilen, but read $bytesread bytes in bytecode.");
	}

	#my $xname = $BytecodeInfo{$instr{opcode}}->{name};
	#print "read $xname\n";
	$codeblock{$offset} = \%instr;

	$offset += $ilen;
    }

    return \%codeblock;
}


sub readSourceFileAttribute {
    my ($r_class, $r_attr) = @_;
    
    &assert($r_attr->{len} == 2, "Source File attribute is the wrong length.");
    
    $r_attr->{sourceFileIndex} = &read_u2();
    &checkIndex($r_class, $r_attr->{sourceFileIndex}, "Source File Name Index", $CONSTANT_Utf8);
}

sub readConstantValueAttribute {
    my ($r_class, $r_attr) = @_;
    
    &assert($r_attr->{len} == 2, "Constant Value Attribute is the wrong length.");
    
    $r_attr->{constantValueIndex} = &read_u2();
    # XXX validity of index depends on type of field.
}


sub readExceptionsAttribute {
    my ($r_class, $r_attr) = @_;
    
    $r_attr->{exceptionCount} = &read_u2();
    
    &assert($r_attr->{len} == (2 + ($r_attr->{exceptionCount} * 2)), "ExceptionsAttribute is the wrong length.");
    
    my $i = 0;
    $r_attr->{exceptionIndexTable} = [];
    while ($i < $r_attr->{exceptionCount}) {
	$r_attr->{exceptionIndexTable}[$i] = &read_u2();
	
	&checkIndex($r_class, $r_attr->{exceptionIndexTable}[$i], "Exception table Class", $CONSTANT_Class);
	
	$i++;
    }
}

sub readLocalVariableTableAttribute {
    my ($r_class, $r_attr) = @_;

    $r_attr->{localVariableTablelen} = &read_u2();
    
    &assert($r_attr->{len} == (2 + ($r_attr->{localVariableTablelen} * 10)), 
	    "LocalVariableTable attribute is the wrong length.");
    
    my $i = 0;
    $r_attr->{localVariableTable} = [];
    while ( $i < $r_attr->{localVariableTablelen} ) {
	my %lventry = ();
	$lventry{start_pc} = &read_u2();
	## XXX check with associated code block
	$lventry{length} = &read_u2();
	## XXX check with associated code block

	$lventry{nameIndex} = &read_u2();
	&checkIndex($r_class, $lventry{nameIndex}, "Local Varible Name index", $CONSTANT_Utf8);
	$lventry{descriptorIndex} = &read_u2();
	&checkIndex($r_class, $lventry{nameIndex}, "Local Varible Name index", $CONSTANT_Utf8);

	$lventry{index} = &read_u2();
	$r_attr->{localVariableTable}[$i] = \%lventry;
	$i++;
    }
}

sub readInnerClassesAttribute {
    my ($r_class, $r_attr) = @_;

    $r_attr->{innerClassesCount} = &read_u2();
    
    &assert($r_attr->{len} == (2 + ($r_attr->{innerClassesCount} * 8)), 
	    "InnerClasses attribute is the wrong length.");
    
    my $i = 0;
    $r_attr->{innerClasses} = [];
    while ( $i < $r_attr->{innerClassesCount} ) {
	my %entry = ();
	$entry{innerClassIdx} = &read_u2();
	&checkIndex($r_class, $entry{innerClassIdx}, 
		    "InnerClass Class index", $CONSTANT_Class);

	$entry{outerClassIdx} = &read_u2();
	if ($entry{outerClassIdx} != 0) { # non-member inner classes have 0 here
	    &checkIndex($r_class, $entry{innerClassIdx}, 
			"InnerClass Outer Class index", $CONSTANT_Class);
	}
	$entry{innerNameIdx} = &read_u2();
	if ($entry{outerClassIdx} != 0) { # anonymous classes don't have names
	    &checkIndex($r_class, $entry{innerNameIdx}, 
			"InnerClass Name index", $CONSTANT_Utf8);
	}
	$entry{innerClassAccessFlags} = &read_u2();

	$r_attr->{innerClasses}[$i] = \%entry;
	$i++;
    }
}

sub readSyntheticAttribute {
    my ($r_class, $r_attr) = @_;

    &assert($r_attr->{len} == 0, "Synthetic attribute has non-zero length.");
}

sub readDeprecatedAttribute {
    my ($r_class, $r_attr) = @_;

    &assert($r_attr->{len} == 0, "Deprecated attribute has non-zero length.");
}

sub readLineNumberTableAttribute {
    my ($r_class, $r_attr) = @_;
    
    $r_attr->{lineNumberTableLen} = &read_u2();
    
    my $expectedLen = 2 + ($r_attr->{lineNumberTableLen} * 4);
    &assert($r_attr->{len} eq $expectedLen, 
	    "LineNumberTable attribute is the wrong length: got $r_attr->{len} expected $expectedLen");
    
    my $i = 0;
    $r_attr->{lineNumberTable} = [];
    while ( $i < $r_attr->{lineNumberTableLen} ) {
	my %lineNumberEntry = ();
	$lineNumberEntry{startPC} = &read_u2();
	## XXX check with associated code block
	$lineNumberEntry{lineNumber} = &read_u2();
	$r_attr->{lineNumberTable}[$i] = \%lineNumberEntry;
	$i++;
    }
}

sub readUnknownAttribute {
  my ($r_class, $r_attr) = @_;

  $r_attr->{rawAttr} = &read_n($r_attr->{len});
}


###
### Write primitives
###

sub write_u4 {
    my $val = shift;
    print CLASSOUT pack("N", $val);
}

sub write_u2 {
    my $val = shift;
    &fatalError("u2 too big") if ($val > 0x0000ffff);
    defined($val) || confess("ack val not defined");
    print CLASSOUT pack("n", $val);
}

sub write_u1 {
    my $byte = shift;
    &fatalError("u1 too big") if ($byte > 0x00ff);
    print CLASSOUT pack("C", $byte);
}

sub write_n {
    my ($length, $val) = @_; 
    if (length($val) != $length) {
	&internalError("Cannot write out part of a string.");
    }
    print CLASSOUT $val;	# XXX assumes $val."length" is 'n'
}

sub write_utf8 {
    my $utf8 = shift;
    print CLASSOUT $utf8
}

sub writeAttributes {
    my $r_attrs = shift;

    return if (!defined $r_attrs);

    my $i = 0;
    foreach $r_attr (@{$r_attrs}) {
	my %attribute = %{$r_attr};
	&write_u2($attribute{nameIndex});
	&write_u4($attribute{len});

	if ($attribute{name} eq "Code") {
	    writeCodeAttribute(\%attribute);
	} elsif ($attribute{name} eq "SourceFile") {
	    writeSourceFileAttribute(\%attribute);
	} elsif ($attribute{name} eq "ConstantValue") {
	    writeConstantValueAttribute(\%attribute);
	} elsif ($attribute{name} eq "Exceptions") {
	    writeExceptionsAttribute(\%attribute);
	} elsif ($attribute{name} eq "LocalVariableTable") {
	    writeLocalVariableTableAttribute(\%attribute);
	} elsif ($attribute{name} eq "InnerClasses") {
	    writeInnerClassesAttribute(\%attribute);
	} elsif ($attribute{name} eq "Synthetic") {
	    writeSyntheticAttribute(\%attribute);
	} elsif ($attribute{name} eq "Deprecated") {
	    writeDeprecatedAttribute(\%attribute);
	} elsif ($attribute{name} eq "LineNumberTable") {
	    writeLineNumberTableAttribute(\%attribute);
	} else {
	    writeUnknownAttribute(\%attribute);
	}
    }
}

###
### Attribute writing functions
###

sub writeCodeAttribute {
    my ($r_attr) = @_;
    
    &write_u2($r_attr->{maxStack});
    &write_u2($r_attr->{maxLocals});
    &write_u4($r_attr->{codeLength});
    &writeBytecode($r_attr->{code}, $r_attr->{codeLength});
    &write_u2($r_attr->{exceptionTableLen});
    
    my $i = 0;
    while ($i < $r_attr->{exceptionTableLen}) {
	my %excTableEntry = %{$r_attr->{exceptionTable}[$i]};
	&write_u2($excTableEntry{startpc});
	&write_u2($excTableEntry{endpc});
	&write_u2($excTableEntry{handlerpc});
	&write_u2($excTableEntry{catchtypeidx});
	$i++;
    }
    
    &write_u2($r_attr->{codeAttrCount});
    &writeAttributes($r_attr->{codeAttrs});
}

sub writeBytecode {
    my ($r_code, $codelen) = @_;
    my $offset = 0;
    while ( $offset < $codelen ) {
	my %instr = %{$r_code->{$offset}};
	my $ilen = $BytecodeInfo{$instr{opcode}}->{length};

	&internalError("Unknown opcode $instr{opcode}.") if (!defined($BytecodeInfo{$instr{opcode}}));
	&internalError("Opcode $instr{opcode} is messed up.") if (!defined($ilen));
	
	## Write the opcode out
	&write_u1($instr{opcode});

	if ($ilen < 1) {
	    &internalError("Cannot yet handle writing of variable length instructions.");
	} elsif ($ilen > 1) {
	    my $argfmt = $BytecodeInfo{$instr{opcode}}->{argfmt};

	    &internalError("No argfmt defined for op=$instr{opcode}") 
	      if (!defined($argfmt));

	    my $byteswritten = 1; # 1 for the opcode
	    foreach $arg (split(/:/, $argfmt)) {
		my ($name, $size) = split(/\//, $arg);
		$size = 1 if (!defined($size));
		if ($size == 1) {
		    &write_u1($instr{$name});
		} elsif ($size == 2) {
		    &write_u2($instr{$name});
		} elsif ($size == 4) {
		    &write_u4($instr{$name});
		} else {
		    &internalError("Bad size ($size) in format string $argfmt.");
		}
		$byteswritten += $size;
	    }
	    &assert($byteswritten == $ilen,
		    "Supposed to write ${ilen} bytes, but wrote ${byteswritten} bytes.");
	}

	$offset += $ilen;
    }
}

sub writeSourceFileAttribute {
    my ($r_attr) = @_;
    &write_u2($r_attr->{sourceFileIndex});
}

sub writeConstantValueAttribute {
    my ($r_attr) = @_;
    &write_u2($r_attr->{constantValueIndex});
}

sub writeExceptionsAttribute {
    my ($r_attr) = @_;
    
    &write_u2($r_attr->{exceptionCount});
    
    my $i = 0;
    while ($i < $r_attr->{exceptionCount}) {
	&write_u2($r_attr->{exceptionIndexTable}[$i]);
	$i++;
    }
}

sub writeLocalVariableTableAttribute {
    my ($r_attr) = @_;

    &write_u2($r_attr->{localVariableTablelen});
    
    my $i = 0;
    
    while ( $i < $r_attr->{localVariableTablelen} ) {
	my %lventry = %{$r_attr->{localVariableTable}[$i]};
	&write_u2($lventry{start_pc});
	&write_u2($lventry{length});
	&write_u2($lventry{nameIndex});
	&write_u2($lventry{descriptorIndex});
	&write_u2($lventry{index});
	$i++;
    }
}

sub writeInnerClassesAttribute {
    my ($r_attr) = @_;

    &write_u2($r_attr->{innerClassesCount});
    
    my $i = 0;
    while ( $i < $r_attr->{localVariableTablelen} ) {
	my %entry = %{$r_attr->{innerClasses}[$i]};
	&write_u2($entry{innerClassIdx});
	&write_u2($entry{outerClassIdx});
	&write_u2($entry{innerNameIdx});
	&write_u2($entry{innerClassAccessFlags});
	$i++;
    }
}

sub writeSyntheticAttribute {
    my ($r_attr) = @_;
    &assert($r_attr->{len} == 0, "Synthetic attribute has non-zero length.");
}

sub writeDeprecatedAttribute {
    my ($r_attr) = @_;
    &assert($r_attr->{len} == 0, "Deprecated attribute has non-zero length.");
}

sub writeLineNumberTableAttribute {
    my ($r_attr) = @_;
    
    &write_u2($r_attr->{lineNumberTableLen});

    my $i = 0;
    
    while ( $i < $r_attr->{lineNumberTableLen} ) {
	my %lineNumberEntry = %{$r_attr->{lineNumberTable}[$i]};
	&write_u2($lineNumberEntry{startPC});
	&write_u2($lineNumberEntry{lineNumber});
	$i++;
    }
}

sub writeUnknownAttribute {
  my ($r_attr) = @_;
  &write_n($r_attr->{len}, $r_attr->{rawAttr});
}


### 
### Grab functions for parsing raw blocks of data (e.g, code).
###


sub grab_u1 {
    my ($offset, $string) = @_;
    my $val = 0;
    ($val) = unpack("C", substr($string, $offset, 1));
    return $val;
}

sub grab_u4 {
    my ($offset, $string) = @_;
    my $val = 0;
    ($val) = unpack("N", substr($string, $offset, 4));
    return $val;
}

sub grab_u2 {
    my ($offset, $string) = @_;
    my $val = 0;
    ($val) = unpack("n", substr($string, $offset, 2));
    return $val;
}

sub grab_n {
    my ($offset, $string, $len) = @_;
    return substr($string, $offset, $len)
}

###
### Fatal error handler
###

sub internalError {
    print STDERR "ERROR: ";
    print STDERR @_;
    print STDERR "\n";

    confess("Here's where we went astray:");

    exit 31;
}

sub fatal {
    print STDERR @_;
    print STDERR "\n";
    exit 11;
}

###
### Attribute information.
###
### Indexed by attribute name.
###
### XXX Currently unused

%AttributeInfo = 
  (
   'SourceFile'    => { fmt => 'sourcefile_cpindex/2' },
   'ConstantValue' => { fmt => 'constantval_cpindex/2' },
   'Code'          => { reader => 'readBytecode', 
			writer => 'writeBytecode'},
   'Exceptions'    => { fmt => 'exceptionCount/2,exception_cpindex[exceptionCount]'},
  );

###
### Bytecode information.
###
### Table is indexed by opcode.  Each opcode is a hash of properties
### including:
###    name : the mnemonic name of the opcode
###    length : the length of the bytecode (if not variable)
###    argfmt* : the format of the inline arguments to the bytecode
###

### argfmt is cool.  It is a series of ':' separate fields.  One for
### each logical argument in the bytecode.  Each argument is 1 byte
### long by default.  Longer arguments have a /x appended where 'x' is
### the length (2 or 4).  The name of the logical argument is 'index'
### if the argument is an index in the constant pool, 'sindex' for stack
### indicies, 'branch' for bytecode offsets, 'const' for a constant value,
### A constant value (e.g, '0') implies the argument must be that constant.
### 'atype' is magic for the newarray bytecode.

###
### Table is sorted by mnemonic, except for the various "odd" bytecodes
### which are all at the end of the table.
###
###  * denotes an optional field

%BytecodeInfo = (
   50 => { name => 'aaload',		length => 1 },
   83 => { name => 'aastore',		length => 1 },
    1 => { name => 'aconst_null',	length => 1 },
   25 => { name => 'aload',		length => 2,	'argfmt' => 'sindex' },
   42 => { name => 'aload_0',		length => 1 },
   43 => { name => 'aload_1',		length => 1 },
   44 => { name => 'aload_2',		length => 1 },
   45 => { name => 'aload_3',		length => 1 },
  189 => { name => 'anewarray',		length => 3,	'argfmt' => 'index/2' },
  176 => { name => 'areturn',		length => 1 },
  190 => { name => 'arraylength',	length => 1 },
   58 => { name => 'astore',		length => 2,	'argfmt' => 'sindex' },
   75 => { name => 'astore_0',		length => 1 },
   76 => { name => 'astore_1',		length => 1 },
   77 => { name => 'astore_2',		length => 1 },
   78 => { name => 'astore_3',		length => 1 },
  191 => { name => 'athrow',		length => 1 },
   51 => { name => 'baload',		length => 1 },
   84 => { name => 'bastore',		length => 1 },
   16 => { name => 'bipush',		length => 2,	'argfmt' => 'const' },
   52 => { name => 'caload',		length => 1 },
   85 => { name => 'castore',		length => 1 },
  192 => { name => 'checkcast',		length => 3,	'argfmt' => 'index/2' },
  144 => { name => 'd2f',		length => 1 },
  142 => { name => 'd2i',		length => 1 },
  143 => { name => 'd2l',		length => 1 },
   99 => { name => 'dadd',		length => 1 },
   49 => { name => 'daload',		length => 1 },
   82 => { name => 'dastore',		length => 1 },
  152 => { name => 'dcmpg',		length => 1 },
  151 => { name => 'dcmpl',		length => 1 },
   14 => { name => 'dconst_0',		length => 1 },
   15 => { name => 'dconst_1',		length => 1 },
  111 => { name => 'ddiv',		length => 1 },
   24 => { name => 'dload',		length => 2,	'argfmt' => 'sindex' },
   38 => { name => 'dload_0',		length => 1 },
   39 => { name => 'dload_1',		length => 1 },
   40 => { name => 'dload_2',		length => 1 },
   41 => { name => 'dload_3',		length => 1 },
  107 => { name => 'dmul',		length => 1 },
  119 => { name => 'dneg',		length => 1 },
  115 => { name => 'drem',		length => 1 },
  175 => { name => 'dreturn',		length => 1 },
   57 => { name => 'dstore',		length => 2,	'argfmt' => 'sindex' },
   71 => { name => 'dstore_0',		length => 1 },
   72 => { name => 'dstore_1',		length => 1 },
   73 => { name => 'dstore_2',		length => 1 },
   74 => { name => 'dstore_3',		length => 1 },
  103 => { name => 'dsub',		length => 1 },
   89 => { name => 'dup',		length => 1 },
   92 => { name => 'dup2',		length => 1 },
   93 => { name => 'dup2_x1',		length => 1 },
   94 => { name => 'dup2_x2',		length => 1 },
   90 => { name => 'dup_x1',		length => 1 },
   91 => { name => 'dup_x2',		length => 1 },
  141 => { name => 'f2d',		length => 1 },
  139 => { name => 'f2i',		length => 1 },
  140 => { name => 'f2l',		length => 1 },
   98 => { name => 'fadd',		length => 1 },
   48 => { name => 'faload',		length => 1 },
   81 => { name => 'fastore',		length => 1 },
  150 => { name => 'fcmpg',		length => 1 },
  149 => { name => 'fcmpl',		length => 1 },
   11 => { name => 'fconst_0',		length => 1 },
   12 => { name => 'fconst_1',		length => 1 },
   13 => { name => 'fconst_2',		length => 1 },
  110 => { name => 'fdiv',		length => 1 },
   23 => { name => 'fload',		length => 2,	'argfmt' => 'sindex' },
   34 => { name => 'fload_0',		length => 1 },
   35 => { name => 'fload_1',		length => 1 },
   36 => { name => 'fload_2',		length => 1 },
   37 => { name => 'fload_3',		length => 1 },
  106 => { name => 'fmul',		length => 1 },
  118 => { name => 'fneg',		length => 1 },
  114 => { name => 'frem',		length => 1 },
  174 => { name => 'freturn',		length => 1 },
   56 => { name => 'fstore',		length => 2,	'argfmt' => 'sindex' },
   67 => { name => 'fstore_0',		length => 1 },
   68 => { name => 'fstore_1',		length => 1 },
   69 => { name => 'fstore_2',		length => 1 },
   70 => { name => 'fstore_3',		length => 1 },
  102 => { name => 'fsub',		length => 1 },
  180 => { name => 'getfield',		length => 3,	'argfmt' => 'index/2' },
  178 => { name => 'getstatic',		length => 3,	'argfmt' => 'index/2' },
  167 => { name => 'goto',		length => 3,	'argfmt' => 'branch/2' },
  200 => { name => 'goto_w',		length => 5,	'argfmt' => 'branch/4' },
  135 => { name => 'i2d',		length => 1 },
  134 => { name => 'i2f',		length => 1 },
  133 => { name => 'i2l',		length => 1 },
   96 => { name => 'iadd',		length => 1 },
   46 => { name => 'iaload',		length => 1 },
  126 => { name => 'iand',		length => 1 },
   79 => { name => 'iastore',		length => 1 },
    2 => { name => 'iconst_m1',		length => 1 }, ## not well documented; pushes -1
    3 => { name => 'iconst_0',		length => 1 },
    4 => { name => 'iconst_1',		length => 1 },
    5 => { name => 'iconst_2',		length => 1 },
    6 => { name => 'iconst_3',		length => 1 },
    7 => { name => 'iconst_4',		length => 1 },
    8 => { name => 'iconst_5',		length => 1 },
  108 => { name => 'idiv',		length => 1 },
  165 => { name => 'if_acmpeq',		length => 3,	'argfmt' => 'branch/2' },
  166 => { name => 'if_acmpne',		length => 3,	'argfmt' => 'branch/2' },
  159 => { name => 'if_icmpeq',		length => 3,	'argfmt' => 'branch/2' },
  162 => { name => 'if_icmpge',		length => 3,	'argfmt' => 'branch/2' },
  163 => { name => 'if_icmpgt',		length => 3,	'argfmt' => 'branch/2' },
  164 => { name => 'if_icmple',		length => 3,	'argfmt' => 'branch/2' },
  161 => { name => 'if_icmplt',		length => 3,	'argfmt' => 'branch/2' },
  160 => { name => 'if_icmpne',		length => 3,	'argfmt' => 'branch/2' },
  153 => { name => 'ifeq',		length => 3,	'argfmt' => 'branch/2' },
  156 => { name => 'ifge',		length => 3,	'argfmt' => 'branch/2' },
  157 => { name => 'ifgt',		length => 3,	'argfmt' => 'branch/2' },
  158 => { name => 'ifle',		length => 3,	'argfmt' => 'branch/2' },
  155 => { name => 'iflt',		length => 3,	'argfmt' => 'branch/2' },
  154 => { name => 'ifne',		length => 3,	'argfmt' => 'branch/2' },
  199 => { name => 'ifnonnull',		length => 3,	'argfmt' => 'branch/2' },
  198 => { name => 'ifnull',		length => 3,	'argfmt' => 'branch/2' },
  132 => { name => 'iinc',		length => 3,	'argfmt' => 'sindex:const' },
   21 => { name => 'iload',		length => 2,	'argfmt' => 'sindex' },
   26 => { name => 'iload_0',		length => 1 },
   27 => { name => 'iload_1',		length => 1 },
   28 => { name => 'iload_2',		length => 1 },
   29 => { name => 'iload_3',		length => 1 },
  104 => { name => 'imul',		length => 1 },
  116 => { name => 'ineg',		length => 1 },
  193 => { name => 'instanceof',	length => 3,	'argfmt' => 'index/2' },
  185 => { name => 'invokeinterface',	length => 5,	'argfmt' => 'index/2:count:0' },
  183 => { name => 'invokespecial',	length => 3,	'argfmt' => 'index/2' },
  184 => { name => 'invokestatic',	length => 3,	'argfmt' => 'index/2' },
  182 => { name => 'invokevirtual',	length => 3,	'argfmt' => 'index/2' },
  128 => { name => 'ior',		length => 1 },
  112 => { name => 'irem',		length => 1 },
  172 => { name => 'ireturn',		length => 1 },
  120 => { name => 'ishl',		length => 1 },
  122 => { name => 'ishr',		length => 1 },
   54 => { name => 'istore',		length => 2,	'argfmt' => 'sindex' },
   59 => { name => 'istore_0',		length => 1 },
   60 => { name => 'istore_1',		length => 1 },
   61 => { name => 'istore_2',		length => 1 },
   62 => { name => 'istore_3',		length => 1 },
  100 => { name => 'isub',		length => 1 },
  124 => { name => 'iushr',		length => 1 },
  130 => { name => 'ixor',		length => 1 },
  168 => { name => 'jsr',		length => 3,	'argfmt' => 'branch/2' },
  201 => { name => 'jsr_w',		length => 5,	'argfmt' => 'branch/4' },
  138 => { name => 'l2d',		length => 1 },
  137 => { name => 'l2f',		length => 1 },
  136 => { name => 'l2i',		length => 1 },
   97 => { name => 'ladd',		length => 1 },
   47 => { name => 'laload',		length => 1 },
  127 => { name => 'land',		length => 1 },
   80 => { name => 'lastore',		length => 1 },
  148 => { name => 'lcmp',		length => 1 },
    9 => { name => 'lconst_0',		length => 1 },
   10 => { name => 'lconst_1',		length => 1 },
   18 => { name => 'ldc',		length => 2,	'argfmt' => 'index' },
   19 => { name => 'ldc_w',		length => 3,	'argfmt' => 'index/2' },
   20 => { name => 'ldc2_w',		length => 3,	'argfmt' => 'index/2' },
  109 => { name => 'ldiv',		length => 1 },
   22 => { name => 'lload',		length => 2,	'argfmt' => 'sindex' },
   30 => { name => 'lload_0',		length => 1 },
   31 => { name => 'lload_1',		length => 1 },
   32 => { name => 'lload_2',		length => 1 },
   33 => { name => 'lload_3',		length => 1 },
  105 => { name => 'lmul',		length => 1 },
  117 => { name => 'lneg',		length => 1 },
  129 => { name => 'lor',		length => 1 },
  113 => { name => 'lrem',		length => 1 },
  173 => { name => 'lreturn',		length => 1 },
  121 => { name => 'lshl',		length => 1 },
  123 => { name => 'lshr',		length => 1 },
   55 => { name => 'lstore',		length => 2,	'argfmt' => 'sindex' },
   63 => { name => 'lstore_0',		length => 1 },
   64 => { name => 'lstore_1',		length => 1 },
   65 => { name => 'lstore_2',		length => 1 },
   66 => { name => 'lstore_3',		length => 1 },
  101 => { name => 'lsub',		length => 1 },
  125 => { name => 'lushr',		length => 1 },
  131 => { name => 'lxor',		length => 1 },
  194 => { name => 'monitorenter',	length => 1 },
  195 => { name => 'monitorexit',	length => 1 },
  197 => { name => 'multianewarray',	length => 4,	'argfmt' => 'index/2:dimensions' },
  187 => { name => 'new',		length => 3,	'argfmt' => 'index/2' },
  188 => { name => 'newarray',		length => 2,	'argfmt' => 'atype' },
    0 => { name => 'nop',		length => 1 },
   87 => { name => 'pop',		length => 1 },
   88 => { name => 'pop2',		length => 1 },
  181 => { name => 'putfield',		length => 3,	'argfmt' => 'index/2' },
  179 => { name => 'putstatic',		length => 3,	'argfmt' => 'index/2' },
  169 => { name => 'ret',		length => 2,	'argfmt' => 'sindex' },
  177 => { name => 'return',		length => 1 },
   53 => { name => 'saload',		length => 1 },
   86 => { name => 'sastore',		length => 1 },
   17 => { name => 'sipush',		length => 3,	'argfmt' => 'const/2' },
   95 => { name => 'swap',		length => 1 },

		 ## These are the only entries which should have zero length (implies variable length)
  171 => { name => 'lookupswitch',	length => 0 },
  170 => { name => 'tableswitch',	length => 0,	'argfmt' => 'padbytes/x,default/4,low/4,high/4,jumptable' },

  196 => { name => 'wide',		length => 0 }, # very special

  145 => { name => 'int2byte',		length => 1,	special => 'undocumented' }, # XXX ???
  146 => { name => 'int2char',		length => 1,	special => 'undocumented' }, # XXX ???
  147 => { name => 'int2short',		length => 1,	special => 'undocumented' }, # XXX ???
  254 => { name => 'impdep1',		length => 1,	special => 'impdep bytecode' },
  255 => { name => 'impdep2',		length => 1,	special => 'impdep bytecode' },
  202 => { name => 'breakpoint',	length => 1,	special => 'runtime only',}, # XXX
  186 => { name => 'UNKNOWN(186)',	length => 1,	special => 'undocumented' }, # undefined by spec

);

1;

# eof
