##
## createnetwork.py
##
## Copyright (c) 2000 University of Utah and the Flux Group.
## All rights reserved.
##
## This file is part of Janos.
## 
## Janos is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License, version 2, as
## published by the Free Software Foundation.
## 
## Janos is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details, in the file
## COPYING.  If that file is not present, write to the Free
## Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
## 02111-1307, USA
## 
## The University of Utah and the Flux Group requests users to
## return to flux-dist@cs.utah.edu any improvements that they make.
##

import sys
import types
import os
import getpass; # For getuser(), for the BASE_PORT mapping...
import string

####
#### Public interfaces:
#### 

##
# Initialize the network creator.  This function *must* be invoked
# before any other createnetwork function is invoked.  If janosvm is
# provided, a full path should be used so that there is no ambiguity
# based on the context of the remote machine.
#
# @param debug defaults to None, non-None turns on "debug"
# @param rsh the remote shell command; defaults to 'ssh -n -x'
# @param janosvm the JanosVM to invoke; default to 'janosvm'
#
def initialize(debug=None,
	       rsh="ssh -n -x",
	       janosvm=None):

    if janosvm == None:
	janosvm = os.environ["JANOSVM"]

    ## Map users to different port ranges.
    portMap = {
	"unknown"   : 25445,
	"tullmann"  : 25450,
	"tullmann2" : 25455,
	"tullmann3" : 25460,
	"stack"     : 25465,
	"stack2"    : 25470,
	"stack3"    : 25475,
	"eeide"     : 25480,
	"eeide2"    : 25485,
	"eeide3"    : 25490,
	"simister"  : 25495,
	"simister2" : 25500,
	"simister3" : 25505,
	"aclement"  : 25510,
        "aclement2" : 25515,
        "aclement3" : 25520
	}
    Configuration.initialized = 1
    Configuration.debug = debug
    Configuration.rshCommand = rsh
    Configuration.janosvm = janosvm

    user = getpass.getuser()
    if portMap.has_key(user):
	Configuration.BASE_PORT = portMap[user]
    else:
	_WARNING("I don't know who you are \"%s\".  Please add yourself to portMap in createnetwork.py" % user)
	Configuration.BASE_PORT = portMap["unknown"]

##
# Define a host.  A host is a machine that one or more
# nodes may run on.
#
# XXX Add connection info (ssh/rsh), add type info (linux/bsd)
# Make paths per-host?
#
# @param name The real name of the host (e.g., 'alta.cs.utah.edu')
# @return the host object
#
def defineHost(name):
    _assertInitialized()
    return name

##
# Define a Node.  An active node has an active address and it runs
# on a particular host.
#
# @param name the pretty name of the node (for messages, mostly)
# @param host the Host the Node will be run on
# @param activeAddr the "active address" of the node
# @return the host Node object
#
def defineNode(name, 
	       host, 
	       activeAddr,
               application=None):
    _assertInitialized()
    node = Node(name, host, activeAddr, application)
    return node

##
# Define a subnet.  A subnet is a set of hosts that are fully connected.
# (imagine they share the same ethernet segment, for example).  Hosts
# may be on more than one subnet.  Such hosts are implicitly routers/bridges.
#
# @param name the pretty name of the subnet
# @param nodes a tuple of hosts
# @return the subnet object
#
def defineSubnet(name,
		 nodes):
    _assertInitialized()
    return Subnet(name,nodes)

##
# Define a link.  A link is a degenerate subnet with two hosts.  Hosts
# may be on more than one link/subnet.
#
# @param name the pretty name of the link
# @param node1 one of the nodes on the link
# @param node2 the other the node on the link
# @return the link object (actually a subnet object)
#
def defineLink(name,
	       node1,
	       node2):
    _assertInitialized()
    return defineSubnet(name,(node1,node2))

##
# Define a network.  A network is a collection of link and/or subnets.
# They may or may not be disjoint subnets (currently no checks are
# done).
# <p>
# When a network is created it is also printed to stdout.
#
# @param name the pretty name for the network
# @param subnets a tuple of subnet/link objects
# @return the network object
#
def defineNetwork(name="ANetwork",
		  subnets=None):
    _assertInitialized()
    n = Network(name, subnets)
    print n
    return n

##
# Generate network start/stop/query scripts for the
# given network in the given directory.
#
# @param network the network to generate scripts for
# @param dir the directory to stick the scripts in
# @param envmagic optional magic to insert in startup scripts
#
def generateNetworkScripts(network, dir, envMagic=None):
    _assertInitialized()
    # Don't double parenthesis later
    if dir[-1]=='/':
        dir=dir[:-1]
    return network._generateScripts(dir, envMagic)

###
### Private classes.  Use public interface.
###

class Configuration:
    ## See initialize() function, above 
    initialized = None
    DEBUG = None
    rshCommand = None
    janosvm = None
    BASE_PORT = None


##
# A network.  A network has a name and gets
# a tuple of subnets.  
#
# It parses the subnet tuple and adds each node on each subnet
# to the list of nodes in the network.  It checks for duplicate
# named nodes and for nodes with duplicate activeAddrs
#
class Network:
    def __init__(self, name, subnets):
	self.name = name
	self.subnets = subnets
	self.nodes = {}
	self.activeAddrMap = {}

	if self.subnets != None:
	    for sn in self.subnets:
		if sn != None:
		    for node in sn._getNodes():
			if node != None:
			    nname = node._getName()
			    naddr = node._getActiveAddr()
			    if (self.nodes.has_key(nname)
				and self.nodes[nname] != node):
				_ERROR("Node name %s is not unique" % node._getName())
			    if (self.activeAddrMap.has_key(naddr)
				and self.activeAddrMap[naddr] != node):
				_ERROR("Node %s phys addr (%s) is not unique" % 
				      (node._getName(), node._getPhysAddr()))

			    self.activeAddrMap[naddr] = node
			    self.nodes[nname] = node

	if len(self.nodes) < 2:
	    _ERROR("Two few connected nodes to have a network!")

				
    def __old_repr__(self):
	if self.subnets == None: 
	    return "%s: empty network" % self.name
	else:
	    result = "%s: " % self.name
	    for s in self.subnets:
		result = result + ("[%s]\n" % s)
	    return result

    def __repr__(self):
	if self.nodes == None:
	    result = "%s: empty network" % self.name
	else:
	    result = "Network %s:\n" % self.name
	    for nname in self.nodes.keys():
		result = result + ("  %s\n" % self.nodes[nname])
	return result

    def _message(self, msg):
	print(msg)

    def _openFile(self, fname):
	return TextFile(fname, Configuration.debug) ## XXX

    def _createDir(self, name):
	try:
	    os.mkdir(name)
	except:
	    pass

    def _fileHeader(self, msg):
	return "# Generated for Network %s\n# -- %s\n" % (self.name, msg)

    def _genAllnodesFile(self, etcDir):
	f = self._openFile(etcDir + "/allnodes")
	f.write(self._fileHeader("All Nodes"))
	for nname in self.nodes.keys():
	    n = self.nodes[nname]
	    f.write("%s\n" % n._activeInfo())
	f.write("# eof\n")
	f.close()
	return f.name()

    def _genGenericStartNode(self, etcDir, binDir, janosvm, envMagic):
	f = self._openFile(binDir + "/startlocalnode")

	f.write("#!/bin/sh\n")
	f.write(self._fileHeader("startlocalnode"))
	f.write("""
        ACTIVEADDR=$1
        shift
        HOSTPORT=$1
        shift
	""");
        
	f.write("""
        JVM="%s"
        ANTSR="edu.utah.janos.nodeos.Main edu.utah.janos.antsr.core.NodeConfigure"
        PIDFILE="%s/pid.${HOSTPORT}"
	""" % (janosvm, etcDir))
        
	if envMagic != None:
	    f.comment("Add users envMagic")
	    f.write(envMagic)
	    f.write(" ")

	f.write("""
        # Assume PID file doesn't exist.  Check done in 'startnetwork'

        $JVM -Djanos.debugFlags=NONE $ANTSR $ACTIVEADDR $HOSTPORT $@ &
        PID=$!
        
        /bin/rm -f ${PIDFILE}
        /bin/echo $PID > ${PIDFILE}
        
	# Wait for child to die, then clean up its pid file.
        wait
        
        /bin/rm -f ${PIDFILE}
        
        #eof
        """)
	f.makeExecutable()
	f.close()
	return f.name()

    def _genGenericStopNode(self, etcDir, binDir):
	f = self._openFile(binDir + "/stoplocalnode")
	f.write("#!/bin/sh\n")
	f.write(self._fileHeader("stoplocalnode"))
	f.write("""
        HOSTPORT=$1
        shift
        
        PIDFILE=%s/pid.${HOSTPORT}
        
        if test -f ${PIDFILE}; then
            # XXX kill, then check, then kill -9 (recurse?)
            kill `cat ${PIDFILE}`
            # startup script will clean up file
            exit 0
        else
	    echo "No PID file for ${HOSTPORT}."
            exit 1
        fi
        
        #eof
        """ % etcDir)
	f.makeExecutable()
	f.close()
	return f.name()

    def _genGenericNodeStatus(self, etcDir, binDir):
	f = self._openFile(binDir + "/localnodestatus")
	f.line("#!/bin/sh")
	f.line(self._fileHeader("localnodestatus"))
	f.write("""
        HOSTPORT=$1
        shift
        NODENAME=$1
        shift
        
        ###
        
        RC=0
        PID=-1
        PIDFILE="%s/pid.${HOSTPORT}"
        PREFIX="${NODENAME} (${HOSTPORT}):: "
        
        if test -f ${PIDFILE}; then
            PID=`cat ${PIDFILE}`
            #echo "${PREFIX} registered PID ${PID}."
            CMD=`ps -p ${PID} -c| tail -1 | awk '{ print $5 }'`
            if test "${CMD}xx" = "Kaffexx" -o "${CMD}xx" = "javaxx"; then
        	#echo "${PREFIX} Kaffe is ${PID}."
            else
        	echo "${PREFIX} PID is actually ${CMD}?!."
        	RC=1
            fi	
        else
            echo "${PREFIX} No PID registered."
            RC=1
        fi
        
        PORT="`echo $HOSTPORT | sed -e 's/^.*://'`"
        PORTSTAT="`/usr/local/sbin/lsof -i UDP:${PORT} -F`"
        
        if test -z "${PORTSTAT}"; then
            echo "${PREFIX} Nothing is listening on ${PORT}."
            RC=1
        else
            PSTAT_PID="`echo $PORTSTAT | perl -ne 'print $1 if (/\s*p(\d*)\s*/);'`"
            PSTAT_CMD="`echo $PORTSTAT | perl -ne 'print $1 if (/\s*c(\w*)\s*/);'`"
        
            #echo "PSTAT_PID = ${PSTAT_PID}"
            #echo "PSTAT_CMD = ${PSTAT_CMD}"
        
            BAD=0
            if test "${PSTAT_PID}" -ne "${PID}"; then
        	echo "${PREFIX} Unexpected PID (${PSTAT_PID}) listening on udp ${PORT}!"
        	RC=1
        	BAD=1
            fi
            if test "${PSTAT_CMD}xx" != "Kaffexx"; then
        	echo "${PREFIX} Unexpected process (${PSTAT_CMD}) is listening on udp ${PORT}!"
        	RC=1
        	BAD=1
            fi
        
            if test ${BAD} -ne 1; then
        	#echo "${PREFIX} ${PSTAT_CMD} is listening on udp ${PORT}."
            fi
        fi
        
        if test ${RC} -eq 0; then
            echo "${PREFIX} ok (pid is ${PID})."
        fi
        
        exit $RC
        
        #eof
        """ % etcDir)
	f.makeExecutable()
	f.close()
	return f.name()

    def _genRouteFile(self, etcDir, node):
	self._createDir("%s/%s" % (etcDir, node._getPhysAddr()))
	f = self._openFile("%s/%s/routeTable" % (etcDir, node._getPhysAddr()))
	f.write(self._fileHeader("Route Table for %s (%s)" % (node._getName(), node._getPhysAddr())))
	node._writeNeighborTable(f)
	f.write("# eof\n")
	f.close()
	return f.name()

    def _genAppScript(self, etcDir, node):
        if node.application:
            f = self._openFile("%s/%s" % (etcDir, node._getAppScriptName()))
            node._writeAppScript(f)
            f.close()
            return f.name()
        return None

    def _genStartNetworkNodes(self, etcDir, binDir, f, sln, nodes):
	for nname in nodes.keys():
	    n = nodes[nname]
            routeFile = "%s/%s/routeTable" % (etcDir, n._getPhysAddr())
            f.comment("Start %s" % nname)
            f.line('echo "Starting %s on %s"' % (nname, n._getPhysAddr()))
            if n.application:
                f.line("%s %s %s %s %s -consoleport %d -routes %s -msgprefix '%s' -application %s/%s &" %
                       (Configuration.rshCommand,
                        n._getHost(), 
                        sln,
                        n._getActiveAddr(), 
                        n._getPhysAddr(), 
                        n._getConsolePort(),
                        routeFile,
                        nname,
                        etcDir,
                        n._getAppScriptName()))
            else:
                f.line("%s %s %s %s %s -consoleport %d -routes %s -msgprefix '%s' &" %
                       (Configuration.rshCommand,
                        n._getHost(), 
                        sln,
                        n._getActiveAddr(), 
                        n._getPhysAddr(), 
                        n._getConsolePort(),
                        routeFile,
                        nname))
            f.blank()

    def _genStartNetwork(self, etcDir, binDir):
	## XXX Put in bin?
	f = self._openFile("%s/startnetwork" % binDir)
	f.line("#!/bin/sh")
	f.write(self._fileHeader("Start Network %s" % self.name))
	f.blank()
	f.comment("Make sure none of the PID files exists.  Bail if any do.")
	for nname in self.nodes.keys():
	    n = self.nodes[nname]
	    pidFile="%s/pid.%s" % (etcDir, n._getPhysAddr())
	    f.line("if test -f %s; then" % pidFile)
	    f.line('    echo "PID file %s already exists.  Please clean up (try stopnetwork).  No nodes started."' % pidFile)
	    f.line('    exit 1')
	    f.line('fi')

	f.blank()
	f.comment("Fire script up in an xterm")
	f.write("""
        if test $# -eq 0; then
            xterm -name janosnet -aw -b 0 -sb -sl 1000 -title 'Janos Net %s' -e $0 anArgument &
            exit 0;
        fi
        """ % self.name);

	sln = "%s/startlocalnode" % binDir
	f.blank()
	f.line('echo "Spawning nodes.  Will wait for all to complete."')
        f.line('trap "./stopnetwork;exit 0" SIGINT')
	f.blank()

        # Filter out the application-running nodes
        appholder={}
        # Create a deep copy of self.nodes
        nonappnodes={}
        for nname in self.nodes.keys():
            nonappnodes[nname]=self.nodes[nname]
        # Filter
        for nname in nonappnodes.keys():
            if nonappnodes[nname].application:
                appholder[nname]=nonappnodes[nname]
                del nonappnodes[nname]
        if len(appholder):
            f.comment("Start the non-application nodes")
        else:
            f.comment("Start the nodes")
        self._genStartNetworkNodes(etcDir,binDir,f,sln,nonappnodes)
        if len(appholder):
            f.comment("Wait for the non-app nodes to start")
            f.line('echo "Waiting 15 seconds for non-application nodes to start"')
            f.line('sleep 15')
            f.comment("Start the application nodes")
            self._genStartNetworkNodes(etcDir,binDir,f,sln,appholder)
        f.comment("wait for all the sshs to complete.")
        f.line("trap ./stopnetwork SIGINT")
        f.line("wait")        
	f.line('echo "All nodes have died."')
	f.line('echo "Press ENTER to exit shell."')
	f.line('read ENTER')
	f.comment("eof")
	f.makeExecutable()
	f.close()
	return f.name()

    def _genStopNetwork(self, etcDir, binDir):
	## XXX Put in bin?
	f = self._openFile("%s/stopnetwork" % binDir)
	f.line("#!/bin/sh")
	f.write(self._fileHeader("Stop Network %s" % self.name))
	f.blank()
	sln = "%s/stoplocalnode" % binDir
	f.blank()
	for nname in self.nodes.keys():
	    n = self.nodes[nname]
	    f.comment("Stop %s" % nname)
	    f.line('echo "Spawning stop of %s (%s)"' % (nname, n._getPhysAddr()))
	    f.line("%s %s %s %s &" %
		    (Configuration.rshCommand,
		     n._getHost(), 
		     sln,
		     n._getPhysAddr()))
	    f.blank()
	f.comment("wait for all the sshs to complete.")
	f.line('echo "Waiting for all stop commands to complete."')
	f.line("wait")
	f.comment("eof")
	f.makeExecutable()
	f.close()
	return f.name()

    def _genNetStatus(self, etcDir, binDir):
	## XXX Put in bin?
	f = self._openFile("%s/networkstatus" % binDir)
	f.line("#!/bin/sh")
	f.write(self._fileHeader("Status of %s Network" % self.name))
	f.blank()
	lns = "%s/localnodestatus" % binDir
	f.blank()
	for nname in self.nodes.keys():
	    n = self.nodes[nname]
	    f.comment("Status for %s" % nname)
	    #f.line('echo "Checking status of %s (%s)"' % (nname, n._getPhysAddr()))
	    f.line("%s %s %s %s %s &" %
		    (Configuration.rshCommand,
		     n._getHost(), 
		     lns,
		     n._getPhysAddr(),
		     n._getName()))
	    f.blank()

	f.comment("wait for checks to complete.")
	f.line("wait")
	f.comment("eof")
	f.makeExecutable()
	f.close()
	return f.name()

    def _generateScripts(self, dir, envMagic):
	etcDir = dir + "/etc"
	binDir = dir + "/bin"
	#libDir = dir + "/lib"

	self._createDir(dir);
	self._createDir(etcDir)
	self._createDir(binDir)
	#self._createDir(libDir)

	self._message("Generating start/stop scripts.")
	fname = self._genGenericStartNode(etcDir, 
					  binDir, 
					  Configuration.janosvm,
					  envMagic) 
	fname = self._genGenericStopNode(etcDir, binDir)
	fname = self._genGenericNodeStatus(etcDir, binDir)

	self._message("Generating routing tables.")
	for nname in self.nodes.keys():
	    self._genRouteFile(etcDir, self.nodes[nname])

        self._message("Generating application scripts.")
        for nname in self.nodes.keys():
            self._genAppScript(etcDir, self.nodes[nname])
	
	self._message("Generating network startup/shutdown.")
	self._genAllnodesFile(etcDir)
	fname = self._genStartNetwork(etcDir, binDir)
	self._message("   %s" % fname)
	fname = self._genStopNetwork(etcDir, binDir)
	self._message("   %s" % fname)

	self._message("Generating network status checker.")
	self._genNetStatus(etcDir, binDir)

class Subnet:
    def __init__(self, name, nodes):
	self.name = name
	self.nodes = nodes

	for n in self.nodes:
	    n._addToSubnet(self)

    def _getName(self):
	return self.name

    def _getNodes(self):
	return self.nodes

    def __repr__(self):
	if self.nodes == None:
	    result = "%s: empty" % self.name
	else:
	    result = "%s: " % self.name
	    for n in self.nodes:
		result = result + ("[%s] " % n)
	return result

class Node:
    nodePortMap = {};

    def __init__(self, name, host, activeAddr, application):
	self.name = name
	self.host = host
	self.activeAddr = activeAddr
        self.application = application
	self.interfaceCt = 0
	self.subnets = [];

	if not Node.nodePortMap.has_key(host):
	    Node.nodePortMap[host] = Configuration.BASE_PORT

	self.port = Node.nodePortMap[host]
	Node.nodePortMap[host] = Node.nodePortMap[host] + 1

	self.consolePort = self.port + 50

    def _addToSubnet(self, subnet):
	self.interfaceCt = self.interfaceCt + 1
	self.subnets.append(subnet)

    def _getPhysAddr(self):
	return "%s:%d" % (self.host, self.port)

    def _getActiveAddr(self):
	return self.activeAddr

    def _getConsolePort(self):
	return self.consolePort

    def _getHost(self):
	return self.host

    def _getAppScriptName(self):
        # Take the class name (w/o package) and add on the port (avoid
        # collisions if running two of the same app)
        if self.application:
            classname = self.application[:string.find(self.application," ")];
            return classname[string.rfind(classname,".")+1:]+str(self.port);
        return None

    def _writeNeighborTable(self, f):
	neighbors = {}
	for sn in self.subnets:
	    for node in sn._getNodes():
		## Don't have to worry about dups, checked by Network
		if node != self:
		    neighbors[node._getName()] = node
	    
	f.write("# Only includes immediate neighbors\n")
	f.write("# \n")
	f.write("# <source>\t<destination>\t<next>    \t<nextphys>\t# <next name>\n")
	for nname in neighbors.keys():
	    n = neighbors[nname]
	    f.write("%s\t%s\t%s\t%s\t# to %s\n" % (self._getActiveAddr(),
						   n._getActiveAddr(),
						   n._getActiveAddr(),
						   n._getPhysAddr(),
						   n._getName()))
	pass

    def _writeAppScript(self, f):        
        f.line(self.application)
        return self.application
        
    def _getName(self):
	return self.name

    def _activeInfo(self):
	result = "%s # %s (%s:%s)" % (self.activeAddr, self.name, self.host, self.port)
	return result

    def __repr__(self):
        if self.application:
            result = "Node: %s/%s (on %s:%d) running: %s " % (self.name, self.activeAddr, self.host, self.port, self.application)
        else:
            result = "Node: %s/%s (on %s:%d) " % (self.name, self.activeAddr, self.host, self.port)            

	if self.subnets == []: 
	    result = result + "NOT ON ANY SUBNET/LINK"
	else:
	    result = result + "Subnets: "
	    for s in self.subnets:
		result = result + ("%s " % s._getName())
	return result

class TextFile:
    def __init__(self, fname, debug=None):
	self.f = None
	self.debug = debug

	if self.debug != None:
	    print ">>>> %s >>>>" % fname
	    self.f = sys.stdout
	else:
	    self.f = open(fname, "w")

    def name(self):
	return self.f.name

    def makeExecutable(self):
	if self.debug == None:
	    os.chmod(self.f.name, 0775)

    def blank(self):
	self.write("\n")

    def line(self, msg):
	self.write(msg + "\n")

    def setVar(self, var, val):
	self.write("%s=%s\n" % (var, val))

    def comment(self, com):
	self.line("# " + com)

    def write(self, msg):
	self.f.write(msg)

    def close(self):
	if self.debug != None:
	    print "<<<<\n"
	else:
	    self.f.close()
    
def _assertInitialized():
    if Configuration.initialized == 0:
	_ERROR("Please call createnetwork.initialize() first")


def _WARNING(msg):
    print "WARNING %s" % msg


def _ERROR(msg):
    print "ERROR %s" % msg
    raise RunTimeError, msg

# eof

